/**
 * @file scheduler.c
 * @brief The FAViM scheduler
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <engine/scheduler.h>
#include <vmstore/context.h>
#include <vmstore/vm.h>
#include <vmstore/concurrency.h>
#include <operationstore/library_manager.h>
#include <operationstore/library.h>
#include <util/favim_queue.h>
#include <capsulestore/capsule.h>

//! The state of the scheduler
typedef struct {
    func_cb_ptr get_library;
    func_cb_ptr get_opcode;
    func_cb_ptr get_library_pid;
    func_cb_ptr concurrency_halt;
    func_cb_ptr get_needed_libs;
	func_cb_ptr libs_present;
	sos_pid_t pid;
    FavimQueue runQueue;
    FavimContext* runningContext;
    struct {
//        uint8_t inErrorState  : 1;
//        uint8_t errorFlipFlop : 1;
        uint8_t taskRunning   : 1;
        uint8_t halted        : 1;
    } __attribute__((packed)) flags;
} scheduler_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t submit(func_cb_ptr, FavimContext*);
static void run_task();
static void reboot(func_cb_ptr);
static void opdone(func_cb_ptr, FavimContext*);
static int8_t execute(FavimContext*);
static inline int8_t computeInstruction(FavimContext*);


//! The module header of the scheduler
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_SCHEDULER_PID,
	.state_size     = sizeof(scheduler_state_t),
    .num_timers     = 0,
	.num_sub_func   = 6,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "CCv1", FAVIM_VM_PID, FAVIM_VM_GET_LIBRARY_FID},
		{NULL, "CCv1", FAVIM_VM_PID, FAVIM_VM_GET_OPCODE_FID},
		{NULL, "CCv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_GET_LIBRARY_PID_FID},
		{NULL, "vav1", FAVIM_CONC_PID, FAVIM_CONC_HALT_FID},
		{NULL, "YCv1",FAVIM_VM_PID, FAVIM_VM_GET_NEEDED_LIBS_FID},
		{NULL, "Cyv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{submit, "cav1", FAVIM_SCHEDULER_PID, FAVIM_SCHEDULER_SUBMIT_FID},
		{opdone, "vav1", FAVIM_SCHEDULER_PID, FAVIM_SCHEDULER_OPDONE_FID},
		{reboot, "vvv0", FAVIM_SCHEDULER_PID, FAVIM_SCHEDULER_REBOOT_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    scheduler_state_t *s = (scheduler_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("FAVIM: scheduler initialization\n");
			s->pid = msg->did;
//			s->flags.inErrorState = FALSE;
			s->flags.halted = FALSE;
			s->flags.taskRunning = FALSE;
//			s->flags.errorFlipFlop = 0;
//			s->libraries = 1;	//Basic library is already there.
			s->runningContext = NULL;
//			ker_timer_init(s->pid, ERROR_TIMER, TIMER_REPEAT);
			reboot(0);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_SCHEDULER_HALT:
        {
			DEBUG("FAVIM: scheduler halted.\n");
			s->flags.halted = TRUE;
        		return SOS_OK;
        }
        
        case FAVIM_CMD_SCHEDULER_RESUME:
        {
			s->flags.halted = FALSE;
			DEBUG("FAVIM: scheduler resumes....\n");
			if (!s->flags.taskRunning) {
			    s->flags.taskRunning = TRUE;
			    DEBUG("FAVIM SCHEDULER: run_task called...\n");
			    run_task();
			}
        		return SOS_OK;
        }

	    case FAVIM_CMD_SCHEDULER_RUN_TASK:
	    {
			run_task();
			return SOS_OK;
	    }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Submit a context for execution
static int8_t submit(func_cb_ptr proto, FavimContext* context)
{
    DEBUG("FAVIM SCHEDULER: Context %i submitted to run.\n", (int)context->which);
    context->state = FAVIM_CONTEXT_STATE_READY;
    return execute(context);
}


static void opdone(func_cb_ptr proto, FavimContext* caller)
{
	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(FAVIM_SCHEDULER_PID);

	DEBUG("FAVIM SCHEDULER: opdone starts running...\n");
	
	if(caller->state == FAVIM_CONTEXT_STATE_RUN){
		caller->pc++;
		caller->num_executed++;
	}
	
	if(s->runningContext->state == FAVIM_CONTEXT_STATE_HALT) {
		// The caller has executed the halt instruction, signal to concurrency
		DEBUG("FAVIM SCHEDULER: The running context has executed halt, signaling to concurrency\n");
		SOS_CALL(s->concurrency_halt, favim_concurrency_halt_proto, caller);
		s->runningContext = NULL;
	}
	run_task();
}	



//! Schedule
static void run_task()
{
	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(FAVIM_SCHEDULER_PID);
	
	DEBUG("FAVIM SCHEDULER: run_task starts running...\n");
	
	if (s->flags.halted == TRUE) {
		DEBUG("FAVIM SCHEDULER: Halted, don't run.\n");
		s->flags.taskRunning = FALSE;
		return;
	}

	if (s->runningContext != NULL && s->runningContext->num_executed >= FAVIM_CPU_SLICE) { // Slice has expired
		DEBUG("FAVIM SCHEDULER: Slice for context %i expired, re-enqueue.\n", (int)s->runningContext->which);
		s->runningContext->state = FAVIM_CONTEXT_STATE_READY;
		queue_enqueue(s->runningContext, &(s->runQueue), s->runningContext);

		s->runningContext = NULL;
	}
	
	
	if (s->runningContext == NULL /*&& !s->flags.inErrorState*/ && !queue_empty(&s->runQueue) ) 
	{
		DEBUG("FAVIM SCHEDULER: Need a new context.\n");
		s->runningContext = queue_dequeue(NULL, &s->runQueue );
		s->runningContext->num_executed = 0;
		s->runningContext->state = FAVIM_CONTEXT_STATE_RUN;
	}
	
	// TODO really, really think if this could cause a race condition
	if (s->runningContext != NULL) {
		DEBUG("FAVIM SCHEDULER: Running context was not NULL.. So some task is running already..\n");
		
		if (s->runningContext->state != FAVIM_CONTEXT_STATE_RUN) {
			DEBUG("FAVIM SCHEDULER: No task is running even though runningContext was not NULL..\n");
			s->runningContext = NULL;
			if (post_short(s->pid, s->pid, FAVIM_CMD_SCHEDULER_RUN_TASK, 0, 0, 0) != SOS_OK) {
				s->flags.taskRunning = FALSE;
			}
		} else {
			computeInstruction(s->runningContext);
		}
	} else {
		DEBUG("FAVIM SCHEDULER: Running context was NULL\n");
		s->flags.taskRunning = FALSE;
	}
	DEBUG("FAVIM SCHEDULER: run_task call over\n");
}

//! Signal the reboot of the VM
static void reboot(func_cb_ptr proto)
{
	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(FAVIM_SCHEDULER_PID);
//	MateHandlerID id;

	DEBUG("FAVIM SCHEDULER: rebooting.\n");
	s->runningContext = NULL;

	queue_init( &s->runQueue );

//	SOS_CALL(s->synch_reset, func_vv_t);

//	for (id = 0; id < MATE_HANDLER_NUM; id++) {
//		SOS_CALL(s->analysis_clearAnalysis, func_vy_t, id);
//	}
//	DEBUG("MATE ENGINE: VM: Analyzing lock sets.\n");
//	for (id = 0; id < MATE_HANDLER_NUM; id++) {
//		SOS_CALL(s->analysis_analyzeVars, func_vy_t, id);
//	}
//	s->flags.inErrorState = FALSE;
	s->flags.halted = FALSE;

//	DEBUG("MATE ENGINE: VM: Signaling reboot to contexts and libraries.\n");    
//	SOS_CALL(s->timer0_rebooted, func_vv_t);
//	SOS_CALL(s->reboot_rebooted, func_vv_t);
//	SOS_CALL(s->basiclib_rebooted, func_vv_t);

	//post_short(M_TIMER0_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
	//post_short(M_BASIC_LIB, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
	//post_short(M_REBOOT_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
	//post_short(M_CLIFF_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
	//post_short(M_PUSHBUTTON_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
//	ker_led(LED_RED_OFF);
//	ker_led(LED_GREEN_OFF);
//	ker_led(LED_YELLOW_OFF);
}

//! Execute a context
static int8_t execute(FavimContext* context) {
    scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(FAVIM_SCHEDULER_PID);

    if (context->state != FAVIM_CONTEXT_STATE_READY) {
      DEBUG("FAVIM SCHEDULER: Failed to submit context %i: not in READY state.\n", (int)context->which);
      return -EINVAL;
    }
    
    //Reset program counter en execution counter
    context->pc = 0;
    context->num_executed = 0;
    
    queue_enqueue(context, &s->runQueue, context);

    if (!s->flags.taskRunning) {
        s->flags.taskRunning = TRUE;
        DEBUG("FAVIM SCHEDULER: Executing context.. calling run_task.\n");
		run_task();
    }
    return SOS_OK;
}

static inline int8_t computeInstruction(FavimContext* context) {
    scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(FAVIM_SCHEDULER_PID);
    FavimLibraryId lib_id;
    uint8_t lib_mod;
    uint8_t opcode;
    

    FavimOpCode instr = context->capsule->data[context->pc];
    if (context->state != FAVIM_CONTEXT_STATE_RUN) {
      DEBUG("FAVIM SCHEDULER: (%hhi) Tried to execute instruction in non-run state: %hhi\n", context->which, context->state);
      return -EINVAL;
    }
    
    DEBUG("FAVIM SCHEDULER: Inside compute_instruction \n");
	//! If the context just starts executing: check if libraries are present
    if (context->pc == 0) {
    		FavimLibraryBitmask libs = SOS_CALL(s->get_needed_libs,favim_vm_get_needed_libs_proto,instr);
		if (!SOS_CALL(s->libs_present,favim_lib_manager_libs_present_proto,libs)) {
			DEBUG("FAVIM SCHEDULER: Library missing for context %d. Halting execution.\n", context->which);
			SOS_CALL(s->concurrency_halt, favim_concurrency_halt_proto, context);
			s->runningContext = NULL;
			run_task();
			//TODO error handling
			return -EINVAL;
		}
		context->pc++;
		instr = context->capsule->data[context->pc];
    }
	
	//! Get the library from the instruction
    lib_id = SOS_CALL(s->get_library,favim_vm_get_library_proto,instr);
    lib_mod = SOS_CALL(s->get_library_pid, favim_lib_manager_get_library_pid_proto,lib_id);
    //! Get the opcode from the instruction
    opcode = SOS_CALL(s->get_opcode,favim_vm_get_opcode_proto,instr);

	//! Execute the instruction
    DEBUG("FAVIM SCHEDULER: executing instruction %i in library %i\n", opcode, lib_mod);

	FavimExecutionMsg * msg = my_malloc(sizeof(FavimExecutionMsg),s->pid);
	msg->context = context;
	msg->opcode = opcode;
    post_long(lib_mod, s->pid, FAVIM_CMD_LIB_EXECUTE, sizeof(FavimExecutionMsg), msg, SOS_MSG_RELEASE);
    
    return SOS_OK;
}


#ifndef _MODULE_
mod_header_ptr scheduler_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

