/**
 * @file vm.c
 * @brief A FAViM virtual machine
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <vmstore/vm.h>
#include <operationstore/library_manager.h>
#include <engine/event_manager.h>
#include <vmstore/concurrency.h>
#include <capsulestore/capsule_manager.h>

#define HASH_MASK 0x07

//! The state of a FAViM virtual machine
typedef struct {
	func_cb_ptr concurrency_submit;
	func_cb_ptr context_new;
	func_cb_ptr cap_man_add_capsule;
	func_cb_ptr cap_man_remove_capsule;
	func_cb_ptr concurrency_halt;
	func_cb_ptr context_reset;
	func_cb_ptr event_man_signal_halt;
	sos_pid_t pid;
	FavimVM vm;
	//TODO is this enough
	uint8_t nb_contexts;
} vm_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static FavimLibraryId get_library(func_cb_ptr,FavimOpCode);
static FavimLibraryId get_opcode(func_cb_ptr,FavimOpCode);
static FavimLibraryBitmask get_needed_libs(func_cb_ptr, FavimOpCode);
static int8_t lock_global(func_cb_ptr, FavimContext*, FavimVariableId);
static int8_t unlock_global(func_cb_ptr, FavimContext*, FavimVariableId);
static BOOL is_locked(func_cb_ptr, FavimVariableId);
static void halt(func_cb_ptr);
static FavimContextId* get_present_contexts(func_cb_ptr);

static FavimContext* get_context_from_id(FavimEventId);
static void add_context(FavimContext *);
static void remove_context(FavimContext *);
static void remove_context_by_id(FavimContextId);
static inline void insert_he(FavimContextHashEntry *);
static int8_t get_nb_contexts(func_cb_ptr);
static void insert_sorted(FavimContextId *, FavimContextId, uint8_t);
static inline void release_context_mem(FavimContext *);

//! The module header of the VM module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VM_PID,
	.state_size     = sizeof(vm_state_t),
    .num_timers     = 0,
	.num_sub_func   = 7,
	.num_prov_func  = 9,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "cav1", FAVIM_CONC_PID, FAVIM_CONC_SUBMIT_FID},
		{NULL, "aaC2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_NEW_FID},
		{NULL, "cya2", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_ADD_CAPSULE_FID},
		{NULL, "cyv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_REMOVE_CAPSULE_FID},
		{NULL, "vav1", FAVIM_CONC_PID, FAVIM_CONC_HALT_FID},
		{NULL, "vav1", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_RESET_FID},
		{NULL, "vvv0", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_SIGNAL_HALT_FID},
		{get_library, "CCv1",FAVIM_VM_PID,FAVIM_VM_GET_LIBRARY_FID},
		{get_opcode, "CCv1",FAVIM_VM_PID,FAVIM_VM_GET_OPCODE_FID},
		{get_needed_libs, "YCv1",FAVIM_VM_PID, FAVIM_VM_GET_NEEDED_LIBS_FID},
		{lock_global, "caC2", FAVIM_VM_PID, FAVIM_VM_LOCK_GLOBAL_FID},
		{unlock_global, "caC2", FAVIM_VM_PID, FAVIM_VM_UNLOCK_GLOBAL_FID},
		{is_locked, "CCv1", FAVIM_VM_PID, FAVIM_VM_IS_LOCKED_FID},
		{halt, "vvv0", FAVIM_VM_PID, FAVIM_VM_HALT_FID},
		{get_present_contexts, "avv0", FAVIM_VM_PID, FAVIM_VM_GET_PRESENT_CONTEXTS_FID},
		{get_nb_contexts, "Cvv0", FAVIM_VM_PID, FAVIM_VM_GET_NB_CONTEXTS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    vm_state_t *s = (vm_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			DEBUG("FAVIM: VM INITIALIZATION\n");
			s->pid = msg->did;
			s->nb_contexts=0;
			for(i=0; i < 8; ++i){
				s->vm.contexts[i] = NULL;
			}
			
			// Initialise the VM
			DEBUG("FAVIM VM: fake initializing vm\n");
			// initialise to a VM with 2 contexts
			for(i=0; i < FAVIM_GLOBAL_COUNT; ++i){
				s->vm.globals[i] = NULL;
				s->vm.locks[i].holder = NULL;
			}
			
			FavimContext * temp;
			DEBUG("FAVIM VM: context 0\n");
			// initialise the contexts
			temp = SOS_CALL(s->context_new, favim_context_new_proto, &s->vm, REBOOT_CONTEXT_ID);
			add_context(temp);
			
			DEBUG("FAVIM VM: context 1\n");
			temp = SOS_CALL(s->context_new, favim_context_new_proto, &s->vm, TIMER0_CONTEXT_ID);
			add_context(temp);
			
			DEBUG("FAVIM VM: sending reboot message to self\n");
			post_short(FAVIM_VM_PID,FAVIM_VM_PID,FAVIM_CMD_VM_REBOOT,0,0,0);
			
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_VM_ADD_CONTEXT:
        {
	        DEBUG("FAVIM VM: ADD CONTEXT STARTED\n");
	        	//Stop the VM
	        	halt(0);
	        	//Unmarshall the parameter
	        	MsgParam *rcvd = (MsgParam *)msg->data;		
	        FavimContextId id = (FavimContextId)rcvd->byte;
	        //Make the new context
	        FavimContext* con = SOS_CALL(s->context_new, favim_context_new_proto, &s->vm, id);
	        //Add it to the data structures
	        add_context(con);
	        //Reboot the VM
	        	post_short(FAVIM_VM_PID,FAVIM_VM_PID,FAVIM_CMD_VM_REBOOT,0,0,0);
        	
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_REMOVE_CONTEXT:
        {
	        DEBUG("FAVIM VM: REMOVE CONTEXT STARTED\n");
	        	//Stop the VM
	        	halt(0);
	        	//Unmarshall the parameter
	        	MsgParam *rcvd = (MsgParam *)msg->data;		
	        FavimContextId id = (FavimContextId)rcvd->byte;
	        //Remove the context
	        remove_context_by_id(id);
	        //Reboot the VM
	        	post_short(FAVIM_VM_PID,FAVIM_VM_PID,FAVIM_CMD_VM_REBOOT,0,0,0);
	       
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_EVENT_OCCURED:
        {
	        	// Lookup the context associated with the event
	        	MsgParam *rcvd = (MsgParam *)msg->data;		
	        	FavimEventId id = (FavimEventId)rcvd->byte;
	        	FavimContext* context = get_context_from_id(id);
	        	// Submit the context for execution
	        	SOS_CALL(s->concurrency_submit, favim_concurrency_submit_proto, context);	
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_REBOOT:
        {
        		int i;
        		//Halt the VM / TODO needed? vm is always halted (due to arriving capsule, arriving context, ...)
        		//halt(0);
	        	// Reset the VM (inc. globals,locks)
			for(i=0; i < FAVIM_GLOBAL_COUNT; ++i){
				s->vm.globals[i] = NULL;
				s->vm.locks[i].holder = NULL;
			}
			// Reset contexts
			for(i=0; i < 8; ++i){
				FavimContextHashEntry * current;
				for(current = s->vm.contexts[i]; current != NULL; current = current->next){
					SOS_CALL(s->context_reset, favim_context_reset_proto, current->context);
				}
			}
	        	// Submit the reboot context for execution
	        SOS_CALL(s->concurrency_submit,favim_concurrency_submit_proto, get_context_from_id(REBOOT_CONTEXT_ID));
	        	return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the library id from an instruction
static FavimLibraryId get_library(func_cb_ptr proto, FavimOpCode instr)
{
    return ((instr >> OPCODE_BITS) & LIBRARY_MASK);	
}

//! Get the opcode from an instruction
static FavimLibraryId get_opcode(func_cb_ptr proto, FavimOpCode instr)
{
    return (instr & OPCODE_MASK);
}

//! Get the needed libs (global ids) from the encoding at the start of a capsule
static FavimLibraryBitmask get_needed_libs(func_cb_ptr proto, FavimOpCode libs)
{
	FavimLibraryBitmask ret = libs;
	return ret;
}

//! Lock the given global variable
static int8_t lock_global(func_cb_ptr proto, FavimContext* context, FavimVariableId id)
{
	DEBUG("FAVIM VM: locking %i, context %i\n", id, context->which);
	if(id >= FAVIM_GLOBAL_COUNT){
		return -FAVIM_ERROR_ID_INVALID;
	}
	context->vm->locks[id].holder = context;
	context->heldSet[id / 8] |= (1 << (id % 8));
	return FAVIM_OK;
}

//! Unlock the given global variable
static int8_t unlock_global(func_cb_ptr proto, FavimContext* context, FavimVariableId id)
{
	DEBUG("FAVIM VM: unlocking %i, context %i\n", id, context->which);
	if(id >= FAVIM_GLOBAL_COUNT){
		return -FAVIM_ERROR_ID_INVALID;
	}
	context->vm->locks[id].holder = NULL;
	context->heldSet[id / 8] &= ~(1 << (id % 8));
	return FAVIM_OK;
}

//! Check whether a global is locked
static BOOL is_locked(func_cb_ptr proto, FavimVariableId id)
{
	DEBUG("FAVIM VM: checking lock %i\n", id);
	vm_state_t* s = (vm_state_t *) ker_get_module_state(FAVIM_VM_PID);
	
	if(id >= FAVIM_GLOBAL_COUNT){
		//TODO error?
		return FALSE;
	}
	return (s->vm.locks[id].holder!=NULL);
}

//! Halt the VM
static void halt(func_cb_ptr proto)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	DEBUG("FAVIM VM: halting the VM\n");
	
	SOS_CALL(s->event_man_signal_halt, favim_event_manager_signal_halt_proto);
	
	int i;
	for(i=0; i < 8; ++i){
		FavimContextHashEntry * current;
		for(current = s->vm.contexts[i]; current != NULL; current = current->next){
			SOS_CALL(s->concurrency_halt, favim_concurrency_halt_proto, current->context);		
		}
	}
}

//! Get the currently present contexts
//! The memory will be allocated dynamically and will be owned by the
//! the vm module. Changing ownership will be necessary!
//! The array is sorted to make comparing efficient
static FavimContextId* get_present_contexts(func_cb_ptr proto){
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	FavimContextId * ids = my_malloc(s->nb_contexts*sizeof(FavimContextId),s->pid);
	uint8_t index = 0;
	int i;
	FavimContextHashEntry * current;
	for(i=0; i < 8; ++i){
		current = s->vm.contexts[i];
		while(current != NULL){
			insert_sorted(ids, current->context->which, index++);
			current = current->next;
		}
	}
	return ids;
}

//! Insert an element in a sorted array
static void insert_sorted(FavimContextId * array, FavimContextId elem, uint8_t index){
	while(index > 0 && array[index-1] > elem){
		array[index] = array[index-1];
		index--;
	}
	array[index] = elem;
}

//! Get the number of currently present contexts
static int8_t get_nb_contexts(func_cb_ptr proto){
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	return s->nb_contexts;
}

//! Get the context associated with the given event id
// For now contextid==eventid
static FavimContext* get_context_from_id(FavimEventId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	uint8_t hid = id & HASH_MASK;
	FavimContextHashEntry * current;
	for(current = s->vm.contexts[hid]; current != NULL; current = current->next){
		if(current->context->which == id){
			return current->context;
		}
	}
	//Nothing found with the given id
	return NULL;
}

//! Add a context to the VM datastructures
static void add_context(FavimContext * context){
	DEBUG("FAVIM VM: adding context %i\n",context->which);
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	SOS_CALL(s->cap_man_add_capsule, favim_capsule_manager_add_capsule_proto, context->which, context->capsule);
	
	FavimContextHashEntry * he = (FavimContextHashEntry *) my_malloc(sizeof(FavimContextHashEntry),s->pid);
	if(he == NULL){
		//TODO error handling
		DEBUG("VM: ADD CONTEXT: something very very wrong here!\n");
	}
	he->context = context;
	he->next = NULL;
	insert_he(he);
	s->nb_contexts++;
}


//! Remove a context from the VM datastructures
static void remove_context(FavimContext * context){
	remove_context_by_id(context->which);
}

//! Remove a context from the VM datastructures
static void remove_context_by_id(FavimContextId id){
	//TODO release memory etc ...
	DEBUG("FAVIM VM: removing context %i\n",id);
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	if(s->vm.contexts[hid] == NULL){
		DEBUG("FAVIM VM: tried to remove a non-existent context %i\n", id);
		//TODO error handling
		//return -EINVAL;
	} else if(s->vm.contexts[hid]->context->which == id) {
		FavimContextHashEntry * prev = s->vm.contexts[hid];
		s->vm.contexts[hid] = s->vm.contexts[hid]->next;
		release_context_mem(prev->context);
		ker_free(prev);
		s->nb_contexts--;
	} else {
		FavimContextHashEntry * current = s->vm.contexts[hid];
		while(current->next != NULL && current->next->context->which != id){
			current = current->next;
		}
		if(current->next == NULL){
			DEBUG("FAVIM VM: tried to remove a non-existent context %i\n", id);
			//return -EINVAL;
			//TODO error handling
		} else {
			FavimContextHashEntry * prev = current->next;
			current->next = current->next->next;
			release_context_mem(prev->context);
			ker_free(prev);
			s->nb_contexts--;
		}
	}	
}

//! Release the memory occupied by a context.
static inline void release_context_mem(FavimContext * context)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);

	SOS_CALL(s->cap_man_remove_capsule, favim_capsule_manager_remove_capsule_proto, context->which);
	//TODO mss delegeren nr context, stack en capsule?
	ker_free(context->capsule);
	ker_free(context->opstack);
	ker_free(context);
}


//! insert a entry in the hash table
static inline void insert_he(FavimContextHashEntry * he)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	
	uint8_t hid = (he->context->which & HASH_MASK);
	if(s->vm.contexts[hid] == NULL){
		s->vm.contexts[hid] = he;
	} else {
		FavimContextHashEntry * current = s->vm.contexts[hid];
		while(current->next != NULL){
			current = current->next;
		}
		current->next = he;
	}
}

#ifndef _MODULE_
mod_header_ptr vm_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

