/**
 * @file capsule_manager.c
 * @brief The implementation of the FAViM Capsule Manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <capsulestore/capsule_manager.h>
#include <vmstore/vm.h>
#include <vmstore/concurrency.h>

#define HASH_MASK 0x07

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t add_capsule(func_cb_ptr, FavimVMId, FavimContextId, FavimCapsule*);
static int8_t remove_capsule(func_cb_ptr, FavimVMId, FavimContextId);
static FavimCapsule * get_capsule(func_cb_ptr, FavimCapsuleId);
static uint8_t get_nb_capsules(func_cb_ptr);
static FavimCapsuleIdVersionPair* get_all_versions(func_cb_ptr);

// helper functions
static inline void insert_he(HashEntry * he);
static inline void capsule_arrived(FavimCapsuleId, FavimCapsule *);

//! The module header of ...
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_CAPSULE_MANAGER_PID,
	.state_size     = sizeof(capsule_manager_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_CAPSULE_MANAGER_PID),
    .num_timers     = 0,
	.num_sub_func   = 2,
	.num_prov_func  = 5,
	.module_handler = module,
	.funct = {
		//TODO ERROR HANDLING
		{NULL, "vCv1", FAVIM_VM_PID, FAVIM_VM_HALT_FID},
		{NULL, "vCa2", FAVIM_CONC_PID, FAVIM_CONC_ANALYSE_CAPSULE_FID},
		{add_capsule, "cCC3", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_ADD_CAPSULE_FID},
		{remove_capsule, "cCC2", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_REMOVE_CAPSULE_FID},
		{get_capsule, "ayv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_CAPSULE_FID},
		{get_nb_capsules, "Cvv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_NB_CAPSULES_FID},
		{get_all_versions, "avv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_ALL_VERSIONS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    capsule_manager_state_t *s = (capsule_manager_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			s->pid = msg->did;
			s->nb_capsules = 0;
			for(i=0; i < 8; ++i){
				s->capsules[i] = NULL;
			}
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_CAPS_MANAGER_CAPS_ARRIVED:
        {
        		FavimCapsuleMsg * m = (FavimCapsuleMsg *) msg->data;
        		FavimCapsuleId id = m->id;
        		FavimCapsule * capsule = &(m->capsule);
        		capsule_arrived(id,capsule);
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Add the given capsule with the given version
static int8_t add_capsule(func_cb_ptr proto, FavimVMId vm, FavimContextId con, FavimCapsule* capsule)
{
	FavimCapsuleId id = (vm << FAVIM_CONTEXT_ID_BITS) + con;
	DEBUG("FAVIM CAPSULE MANAGER: adding capsule %i\n",id);
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);

	HashEntry * he = (HashEntry*) my_malloc(sizeof(HashEntry),FAVIM_CAPSULE_MANAGER_PID);
	if(he == NULL){
		//TODO error handling
		panic();
		return -EINVAL;
	}
	he->id = id;
	he->capsule = capsule;
	he->next = NULL;
	insert_he(he);
	s->nb_capsules++;
	return FAVIM_OK;
}

//! Remove the capsule with the specified id
static int8_t remove_capsule(func_cb_ptr proto, FavimVMId vm, FavimContextId con)
{
	FavimCapsuleId id = (vm << FAVIM_CONTEXT_ID_BITS) + con;
	DEBUG("FAVIM CAPSULE MANAGER: removing capsule %i\n",id);
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	if(s->capsules[hid] == NULL){
		DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i\n", id);
		//TODO error handling
		return -EINVAL;
	} else if(s->capsules[hid]->id == id) {
		HashEntry * prev = s->capsules[hid];
		s->capsules[hid] = s->capsules[hid]->next;
		ker_free(prev);
		s->nb_capsules--;
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL && current->next->id != id){
			current = current->next;
		}
		if(current->next == NULL){
			DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i\n", id);
			return -EINVAL;
			//TODO error handling
		} else {
			HashEntry * prev = current->next;
			current->next = current->next->next;
			ker_free(prev);
			s->nb_capsules--;
		}
	}
	return FAVIM_OK;
}

//! Get the capsule with the given id, returns NULL if not found
static FavimCapsule * get_capsule(func_cb_ptr proto, FavimCapsuleId id)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	HashEntry * current = s->capsules[hid];
	while(current != NULL && current->id != id){
		current = current->next;
	}
	if(current == NULL) {
		DEBUG("FAVIM CAPSULEMANAGER: get capsule %i: non-existent, returning NULL\n", id);
		return NULL;
	} else {
		return current->capsule;
	}		
}

//! Get the number of installed capsules
static uint8_t get_nb_capsules(func_cb_ptr proto)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	return s->nb_capsules;
}

//! Return all ids and versions
//! The memory will be allocated dynamically and will be owned by the
//! the capsule manager. Changing ownership will be necessary!
static FavimCapsuleIdVersionPair* get_all_versions(func_cb_ptr proto)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	FavimCapsuleIdVersionPair * versions= (FavimCapsuleIdVersionPair *) my_malloc(s->nb_capsules * sizeof(FavimCapsuleIdVersionPair),FAVIM_CAPSULE_MANAGER_PID);
	if(versions == NULL){
		//TODO error handling;
		panic();
		return NULL;
	}
	FavimCapsuleIdVersionPair * v_orig = versions;
	int i;
	HashEntry * current;
	for(i=0; i < 8; ++i){
		current = s->capsules[i];
		while(current != NULL){
			versions->id = current->id;
			versions->version = current->capsule->version;
			current = current->next;
			versions++;
		}
	}
	return v_orig;
}

//----------------//
// HELP FUNCTIONS //
//----------------//

static inline void insert_he(HashEntry * he)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = (he->id & HASH_MASK);
	if(s->capsules[hid] == NULL){
		s->capsules[hid] = he;
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL){
			current = current->next;
		}
		current->next = he;
	}
}

static inline void capsule_arrived(FavimCapsuleId id, FavimCapsule * capsule)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	capsule_state_t* cap_s = (capsule_state_t*) ker_get_module_state(FAVIM_CAPSULE_PID);
	
	
	// Halt the VM: this is a function call
	// The VM has to halt immediately
	FavimVMId vm = id >> FAVIM_CONTEXT_ID_BITS;
	SOS_CALL(s->vm_halt, favim_vm_halt_proto, vm);

	// Install the capsule
	uint8_t hid = (id & HASH_MASK);
	HashEntry * current = s->capsules[hid];
	while(current != NULL && current->id != id) {
		current = current->next;
	}
	if(current != NULL){
//		int i;
		current->capsule->dataSize = capsule->dataSize;
//		for(i=0; i < FAVIM_CAPSULE_SIZE; ++i){
//			current->capsule->data[i] = capsule->data[i];
//		}
		ker_change_own(capsule->data,s->pid);
		if(current->capsule->data != cap_s->default_capsule){
			ker_free(current->capsule->data);
		}
		current->capsule->data = capsule->data;
		current->capsule->version = capsule->version;
		// Analyse the capsule to initialize usedVars
		SOS_CALL(s->analyse_capsule, favim_concurrency_analyse_capsule_proto, vm, current->capsule);
	} else {
		ker_free(capsule->data);
	}
	
	// Reboot the VM: this is a message
	// SOS delivers messages FIFO, so any pending EXECUTE messages to a lib will be discarded
	// by the lib because the VM is still halted
	post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REBOOT,vm,0,0);
}

#ifndef _MODULE_
mod_header_ptr capsule_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

