/**
 * @file event_manager.c
 * @brief The FAViM event manager
 * @author Wouter Horr&eacute;
 */

#include <engine/event_manager.h>

#define EVENT_ID_MASK 0x1f

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t start_timer(func_cb_ptr,FavimEventType,FavimVMId,FavimEventId,FavimTimerType,FavimTimerInterval);
static int8_t stop_timer(func_cb_ptr,FavimEventType,FavimVMId,FavimEventId);
//static int8_t init_event(func_cb_ptr,FavimEventId);
static int8_t trigger_event(func_cb_ptr,FavimEventType,FavimVMId,FavimEventId);
static void signal_halt(func_cb_ptr,FavimVMId);
static int8_t insert_into_timer_list(FavimEventTimer*);
static void tick(event_man_state_t*);
static void remove_timer(FavimEventType, FavimVMId, FavimEventId, BOOL);

//! The module header of the event manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_EVENT_MANAGER_PID,
	.state_size     = sizeof(event_man_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_EVENT_MANAGER_PID),
    .num_timers     = 1,
	.num_sub_func   = 0,
	.num_prov_func  = 4,
	.module_handler = module,
	.funct = {
		{start_timer,"cCC5",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_START_TIMER_FID},
		{stop_timer,"cCC3",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_STOP_TIMER_FID},
		{trigger_event,"cCC3",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_TRIGGER_EVENT_FID},
		{signal_halt, "vCv1", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_SIGNAL_HALT_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    event_man_state_t *s = (event_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("FAVIM: EVENT MANAGER INITIALIZATION\n");
			s->pid = msg->did;
			s->timerlist = NULL;
			
			ker_timer_init(s->pid, FAVIM_EVENT_TIMER_ID, TIMER_REPEAT);
			
			return SOS_OK;
		}

        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case MSG_TIMER_TIMEOUT:
        {
        		tick(s);
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Signal the halt of the VM
static void signal_halt(func_cb_ptr proto, FavimVMId vm){
	DEBUG("FAVIM: EVENT MANAGER HALTING VM %i\n", vm);
	remove_timer(FAVIM_EVENT_SINGLE_VM,vm,0,TRUE);
}

//! Start a timer
//! TODO This isn't very exact :S
static int8_t start_timer(func_cb_ptr proto, FavimEventType etype, FavimVMId vm, FavimEventId id, FavimTimerType type, FavimTimerInterval interval)
{
	FavimEventTimer* newTimer = (FavimEventTimer*) my_malloc(sizeof(FavimEventTimer),FAVIM_EVENT_MANAGER_PID);
	if(newTimer == NULL){
		//TODO error handling
		panic();
		return -ENOMEM;
	}
	if(etype == FAVIM_EVENT_SINGLE_VM){
		newTimer->id = (vm << FAVIM_CONTEXT_ID_BITS) + id;
	} else {
		newTimer->id = id;
	}
	newTimer->ticks.var = interval;
	newTimer->next = NULL;
	newTimer->type = type;
	if(type == FAVIM_TIMER_REPEAT) {
		DEBUG("FAVIM EVENT MANAGER: STARTING TIMER %i (TYPE %i/VM %i) AT RATE %i\n",id,type,vm,interval);
		newTimer->interval.var=interval;
	} else {
		DEBUG("FAVIM EVENT MANAGER: STARTING TIMER %i (TYPE %i/VM %i) ONCE AT %i TICKS\n",id,type,vm,interval);
		newTimer->interval.var=0;
	}
	return insert_into_timer_list(newTimer);
}

//! Stop a timer
static int8_t stop_timer(func_cb_ptr proto, FavimEventType type, FavimVMId vm, FavimEventId id)
{
	DEBUG("FAVIM: EVENT MANAGER STOPPING TIMER %i (TYPE %i/VM %i)\n", id, type, vm);
	remove_timer(type,vm,id,FALSE);
	return FAVIM_OK;
}

//! Remove one or all timers of a vm from the timerlist
static void remove_timer(FavimEventType type, FavimVMId vm, FavimEventId id, BOOL all)
{
    event_man_state_t *s = (event_man_state_t *)ker_get_module_state(FAVIM_EVENT_MANAGER_PID);
    
    if(type == FAVIM_EVENT_SINGLE_VM){
    		id += (vm << FAVIM_CONTEXT_ID_BITS);
    }
    	
	FavimEventTimer* current = s->timerlist;
	FavimEventTimer** previous = &(s->timerlist);
	while(current != NULL) {
		//if((type!=FAVIM_EVENT_SINGLE_VM || current->vm==vm) && (all || current->id == id)){
		if(current->type == type && (current->id == id || (all && (current->id >> FAVIM_CONTEXT_ID_BITS)==vm))){
			if(current->next != NULL){
				current->next->ticks.var += current->ticks.var;
			}
			*previous = current->next;
			ker_free(current);
			//! If we removed the last timer, stop the kernel timer
			if(s->timerlist == NULL){
				ker_timer_stop(s->pid, FAVIM_EVENT_TIMER_ID);
			}
			if(!all){
				break;
			}
		} else {
			previous = &(current->next);
			current = current->next;
		}
	}
}

//! Trigger an event
static int8_t trigger_event(func_cb_ptr proto, FavimEventType type, FavimVMId vm, FavimEventId id)
{
	if(type == FAVIM_EVENT_SINGLE_VM){
		DEBUG("FAVIM EVENT MANAGER: TRIGGER EVENT %i (VM %i)\n",id, vm);
		return post_short(FAVIM_VM_PID,FAVIM_EVENT_MANAGER_PID,FAVIM_CMD_VM_EVENT_OCCURED,id,vm,0);
	} else if (type == FAVIM_EVENT_GLOBAL){
		DEBUG("FAVIM EVENT MANAGER: TRIGGER GLOBAL EVENT %i\n", id);
		return post_short(FAVIM_VM_MANAGER_PID, FAVIM_EVENT_MANAGER_PID, FAVIM_CMD_VM_MAN_TRIGGER_GLOBAL,id,0,0);
	}
	return -EINVAL;
}

//! Insert a timer into the timerlist
static int8_t insert_into_timer_list(FavimEventTimer* new)
{
	DEBUG("FAVIM: EVENT MANAGER ADDING TIMER %i (TYPE %i) TO TIMER LIST\n",new->id, new->type);
    event_man_state_t *s = (event_man_state_t *)ker_get_module_state(FAVIM_EVENT_MANAGER_PID);
	
	//! If the new timer is the only one, start the kernel timer
	//! Apparently interval 1024 == 1 second, we want 0.1 seconds
	if(s->timerlist == NULL){
		s->timerlist = new;
		ker_timer_start(s->pid,FAVIM_EVENT_TIMER_ID,102);
	} else {
		FavimEventTimer* current = s->timerlist;
		FavimEventTimer** previous = &(s->timerlist);
		while(current!=NULL && (current->type!=new->type || current->id != new->id) && current->ticks.var < new->ticks.var){
			new->ticks.var -= current->ticks.var;
			previous = &(current->next);
			current = current->next;
		}
		if(current!=NULL && current->type==new->type && current->id == new->id){
			//! There is already a timer with the same id!
			//! Cleanup and return error
			ker_free(new);
			return -FAVIM_ERROR_TIMER_EXISTS;
		} else {
			*previous = new;
			new->next = current;
		}
	}
	return FAVIM_OK;
}

//! Advance one tick
static void tick(event_man_state_t *s){
	DEBUG("FAVIM: EVENT MANAGER TICK\n");
	s->timerlist->ticks.var--;
	while(s->timerlist->ticks.var==0){
		FavimEventTimer* current = s->timerlist;
		//TODO maybe a better way to remove the vm id from the event id
		trigger_event(0,current->type,current->id >> FAVIM_CONTEXT_ID_BITS, current->id & EVENT_ID_MASK);
		//! Remove the current timer from the timerlist
		s->timerlist = current->next;
		//! Reschedule the timer if needed, else cleanup
		if(current->interval.var != 0){
			current->ticks.var = current->interval.var;
			if(insert_into_timer_list(current)){
				//THIS SHOULD NOT HAPPEN
				//TODO raise error
			}
		} else {
			ker_free(current);
		}
	}
}

#ifndef _MODULE_
mod_header_ptr event_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

