/**
 * @file library_manager.c
 * @brief The library manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <operationstore/library_manager.h>
#include <vmstore/vm.h>
#include <engine/event_manager.h>

#define LIBRARY_PID_MASK 0x3f

// Forward declarations
static int8_t module(void *state, Message *e);
static BOOL lib_present(func_cb_ptr, FavimLibraryId);
static BOOL libs_present(func_cb_ptr, FavimLibraryBitmask);
static uint8_t get_library_pid(func_cb_ptr, FavimLibraryId);


//! The module header of the library manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_LIB_MANAGER_PID,
	.state_size     = sizeof(lib_man_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_LIB_MANAGER_PID),
    .num_timers     = 0,
	.num_sub_func   = 1,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		//TODO error handling
		{NULL, "cCC3", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_TRIGGER_EVENT_FID},
		{lib_present, "CCv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIB_PRESENT_FID},
		{libs_present, "Cyv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{get_library_pid, "CCv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_GET_LIBRARY_PID_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    lib_man_state_t *s = (lib_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			//s->libraries = 1; // Basic lib is always there, but also registers with us
			s->libraries = 0;
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_LIB_MANAGER_ADD_LIBRARY:
        {
        		FavimLibraryId lib_id = msg->sid & LIBRARY_PID_MASK;
		    uint64_t lib_bit = 0x1 << lib_id;
		    s->libraries |= lib_bit;
		    DEBUG("FAVIM LIBRARY MANAGER : Adding library\n");
		    DEBUG("FAVIM LIBRARY MANAGER : Source of message: %i\n", msg->sid);
		    DEBUG("FAVIM LIBRARY MANAGER : Library id is %i\n", (int)lib_id);
		    DEBUG("FAVIM LIBRARY MANAGER : Library bitmask is now %i\n", (int)s->libraries);
		    SOS_CALL(s->trigger_event, favim_event_manager_trigger_event_proto, FAVIM_EVENT_GLOBAL, 0, FAVIM_EVENT_GLOBAL_ADD_LIB);
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY:
        {
		    FavimLibraryId lib_id = msg->sid & LIBRARY_PID_MASK;
		    uint64_t lib_bit = 0x1 << lib_id;
		    s->libraries &= (~lib_bit);
		    DEBUG("FAVIM LIBRARY MANAGER : Removing library %i\n", lib_id);
		    SOS_CALL(s->trigger_event, favim_event_manager_trigger_event_proto, FAVIM_EVENT_GLOBAL, 0, FAVIM_EVENT_GLOBAL_REMOVE_LIB);
		    return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Is the library with a given id present?
static BOOL lib_present(func_cb_ptr proto, FavimLibraryId lib_id)
{
	if(lib_id < 64){
		FavimLibraryBitmask bitmask = 0x1 << lib_id;
		return libs_present(0, bitmask);
	}
	return FALSE;
}

//! Are all the libraries in the given bitmask present
static BOOL libs_present(func_cb_ptr proto, FavimLibraryBitmask bitmask)
{
	lib_man_state_t* s = (lib_man_state_t*) ker_get_module_state(FAVIM_LIB_MANAGER_PID);
	
	DEBUG("FAVIM LIBRARY MANAGER: checking bitmask %i, libraries %i\n", (int)bitmask, (int)s->libraries);
	
	return ((bitmask & s->libraries) ==  bitmask);
}


//! Get the pid of a given library
static uint8_t get_library_pid(func_cb_ptr proto, FavimLibraryId lib_id)
{
	return (lib_id + FAVIM_LIB_MIN_PID);
}

#ifndef _MODULE_
mod_header_ptr library_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

