/**
 * @file context.c
 * @brief A execution context in a FAViM virtual machine
 * @author Wouter Horr&eacute;
 */



#include <module.h>
#include <vmstore/context.h>
#include <util/favim_queue.h>
#include <capsulestore/capsule_manager.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static FavimContext* new(func_cb_ptr, FavimVMId, FavimContextId);
static void finalize(func_cb_ptr, FavimContext*);
static void reset(func_cb_ptr, FavimContext*);

//! The module header of the context module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_CONTEXT_PID,
	.state_size     = sizeof(context_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_CONTEXT_PID),
    .num_timers     = 0,
	.num_sub_func   = 7,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		// TODO error handler
		{NULL, "avv0", FAVIM_STACK_PID, FAVIM_STACK_NEW_FID},
		{NULL, "vav1", FAVIM_STACK_PID, FAVIM_STACK_FINALIZE_FID},
		{NULL, "vav1", FAVIM_STACK_PID, FAVIM_STACK_RESET_FID},
		{NULL, "avv0", FAVIM_CAPSULE_PID, FAVIM_CAPSULE_NEW_FID},
		{NULL, "vav1", FAVIM_CAPSULE_PID, FAVIM_CAPSULE_FINALIZE_FID},
		{NULL, "cCC3", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_ADD_CAPSULE_FID},
		{NULL, "cCC2", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_REMOVE_CAPSULE_FID},
		{new, "aCC2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_NEW_FID},
		{finalize, "vav1", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_FINALIZE_FID},
		{reset, "vav1", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_RESET_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    context_state_t *s = (context_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Allocate and initialise a new context
static FavimContext* new(func_cb_ptr proto, FavimVMId vm, FavimContextId which)
{
	context_state_t * s = (context_state_t*) ker_get_module_state(FAVIM_CONTEXT_PID);
	
	FavimContext* new_con = (FavimContext*) my_malloc(sizeof(FavimContext),s->pid);
	if(new_con == 0){
		DEBUG("CONTEXT NEW: Something is very very wrong here\n");
		panic();
		return 0;
	}
	new_con->capsule = SOS_CALL(s->capsule_new, favim_capsule_new_proto);
	new_con->opstack = SOS_CALL(s->stack_new, favim_stack_new_proto);
	new_con->vm = vm;
	new_con->which = which;
	new_con->queue = NULL;
	reset(0,new_con);
	SOS_CALL(s->add_capsule, favim_capsule_manager_add_capsule_proto, vm, which, new_con->capsule);
	return new_con;
}

//! Finalize a context
static void finalize(func_cb_ptr proto, FavimContext* context)
{
	context_state_t * s = (context_state_t*) ker_get_module_state(FAVIM_CONTEXT_PID);
	
	SOS_CALL(s->remove_capsule, favim_capsule_manager_remove_capsule_proto, context->vm, context->which);
	SOS_CALL(s->stack_finalize, favim_stack_finalize_proto, context->opstack);
	SOS_CALL(s->capsule_finalize, favim_capsule_finalize_proto, context->capsule);
	ker_free(context);
}

//! Reset the given context
static void reset(func_cb_ptr proto, FavimContext* context)
{
	
	// Do not touch the capsule, leave that to the capsule manager!
	context_state_t * s = (context_state_t*) ker_get_module_state(FAVIM_CONTEXT_PID);
	int i;
	
	context->pc=0;
	context->state=FAVIM_CONTEXT_STATE_HALT;
	context->num_executed=0;
	if(context->queue){
		queue_remove(context, context->queue, context);
	}
	SOS_CALL(s->stack_reset, favim_stack_reset_proto, context->opstack);
	for(i=0; i < (FAVIM_GLOBAL_COUNT + 7)/8; ++i){
		context->heldSet[i] = 0;
		context->acquireSet[i] = 0;
		context->releaseSet[i] = 0;
	}
	for(i=0; i < FAVIM_LOCAL_COUNT; ++i){
		if(context->locals[i] != NULL){
			ker_free(context->locals[i]);
			context->locals[i] = NULL;
		}
	}
}

#ifndef _MODULE_
mod_header_ptr context_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

