/**
 * @file vm_manager.c
 * @brief The FAViM vm manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <engine/event_manager.h>
#include <operationstore/library_manager.h>
#include <virus/virus.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static FavimVM* get_vm(func_cb_ptr,FavimVMId);
static BOOL vm_present(func_cb_ptr,FavimVMId);
static FavimVMVersion get_vm_version(func_cb_ptr,FavimVMId);


// Forward declarations: helper functions
static void add_vm(FavimVM *);
static void remove_vm_by_id(FavimVMId);

//! The module header of the favim vm manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VM_MANAGER_PID,
	.state_size     = sizeof(vm_man_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_VM_MANAGER_PID),
    .num_timers     = 0,
	.num_sub_func   = 4,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		//TODO error handling
		{NULL, "aCv1", FAVIM_VM_PID, FAVIM_VM_NEW_FID},
		{NULL, "vav1", FAVIM_VM_PID, FAVIM_VM_FINALIZE_FID},
		{NULL, "YCC2",FAVIM_VM_PID, FAVIM_VM_GET_NEEDED_LIBS_FID},
		{NULL, "Cyv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{get_vm, "aCv1", FAVIM_VM_MANAGER_PID, FAVIM_VM_MAN_GET_VM_FID},
		{vm_present, "CCv1", FAVIM_VM_MANAGER_PID, FAVIM_VM_MAN_VM_PRESENT_FID},
		{get_vm_version, "LCv1", FAVIM_VM_MANAGER_PID, FAVIM_VM_MAN_GET_VM_VERSION_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    vm_man_state_t *s = (vm_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			
			s->pid = msg->did;
			for(i=0; i<8; ++i){
				s->vms[i] = NULL;
			}
			
			// ADD one VM to the system by default
	        //Make the VM
	        FavimVM * nvm = SOS_CALL(s->vm_new, favim_vm_new_proto, 0);
	        //Add it to the data structures
	        add_vm(nvm);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
//        case FAVIM_CMD_VM_MAN_ADD_VM:
//        {
//        		DEBUG("FAVIM VM MANAGER: add VM started\n");
//        		
//        		MsgParam *rcvd = (MsgParam *)msg->data;		
//	        FavimVMId id = (FavimVMId)rcvd->byte;
//	        
//	        //Make the VM
//	        FavimVM * nvm = SOS_CALL(s->vm_new, favim_vm_new_proto, id);
//	        
//	        //Add it to the data structures
//	        add_vm(nvm);
//	        
//	        //Reboot it
//	        	post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REBOOT,nvm->id,0,0);
//	                		
//        		return SOS_OK;
//        }
//
//        case FAVIM_CMD_VM_MAN_REMOVE_VM:
//        {
//        		DEBUG("FAVIM VM MANAGER: remove VM started\n");
//        		
//        		MsgParam *rcvd = (MsgParam *)msg->data;		
//	        FavimVMId id = (FavimVMId)rcvd->byte;
//	        
//	        remove_vm_by_id(id);
//	        
//        		return SOS_OK;
//        }
        case FAVIM_CMD_VM_MAN_VMMSG_ARRIVED:
        {
        		DEBUG("FAVIM VM MANAGER: VM message arrived\n");
        		FavimVMMsg * vmmsg = (FavimVMMsg*)ker_msg_take_data(s->pid,msg);
        		
        		if(vmmsg->desc.id >= FAVIM_VM_COUNT){
        			//TODO error handling
        			ker_free(vmmsg);
        			return -EINVAL;
        		}
        		
        		if(vmmsg->version > s->versions[vmmsg->desc.id]){
        			s->versions[vmmsg->desc.id] = vmmsg->version;
        			if(vmmsg->present){
        				// It's an add of update
        				if(!vm_present(0,vmmsg->desc.id)){
				        //Make the VM
				        FavimVM * nvm = SOS_CALL(s->vm_new, favim_vm_new_proto, vmmsg->desc.id);
				        
				        //Add it to the data structures
				        add_vm(nvm);
        				}
        				// Install the description (it will be rebooted also)
        				FavimVMDescription * desc = my_malloc(sizeof(FavimVMDescription),s->pid);
        				if(desc == NULL){
        					//TODO error handling
        					panic();
        					return -ENOMEM;
        				}
        				// Deep copy
        				*desc = vmmsg->desc;
        				post_long(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_INSTALL_DESCRIPTION,sizeof(FavimVMDescription),desc,SOS_MSG_RELEASE);
        			} else {
        				// It's a remove
        				if(vm_present(0,vmmsg->desc.id)){
        					remove_vm_by_id(vmmsg->desc.id);
        				}
        			}
        		} else {
        			DEBUG("FAVIM VM MANAGER: we are already up to date");
        		}
        		ker_free(vmmsg);
        		return SOS_OK;
        }
        
        case FAVIM_CMD_VM_MAN_TRIGGER_GLOBAL:
        {
        		FavimEventId id = (FavimEventId)((MsgParam *)msg->data)->byte;
        		if(id == FAVIM_EVENT_GLOBAL_ADD_LIB){
        			//TODO implement when error handling is implemented -> vm in missing lib state
        		} else if (id == FAVIM_EVENT_GLOBAL_REMOVE_LIB) {
        			//check whether vms have missing libs now -> reboot them
        			int i;
        			for(i=0; i < FAVIM_VM_COUNT; ++i){
    					FavimLibraryBitmask mask = SOS_CALL(s->vm_get_needed_libs, favim_vm_get_needed_libs_proto, s->vms[i]->id, 0xff);
    					if(!SOS_CALL(s->lib_man_libs_present, favim_lib_manager_libs_present_proto,mask)){
				        	post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REBOOT,s->vms[i]->id,0,0);
    					}
        			}
        		}
        		return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get a vm by its id
static FavimVM* get_vm(func_cb_ptr proto,FavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(FAVIM_VM_MANAGER_PID);

	if(id >= FAVIM_VM_COUNT){
		//TODO error handling
		return NULL;
	}
	return s->vms[id];
}

//! Check whether a VM is present
static BOOL vm_present(func_cb_ptr proto, FavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(FAVIM_VM_MANAGER_PID);
	return (id<FAVIM_VM_COUNT && s->vms[id]!=NULL);
}

//! Get the version of a VM
static FavimVMVersion get_vm_version(func_cb_ptr proto,FavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(FAVIM_VM_MANAGER_PID);
	if(id >= FAVIM_VM_COUNT){
		//TODO error handling
		return 0;
	} else {
		return s->versions[id];
	}
}

//! Add a VM to the data structures
static void add_vm(FavimVM * vm)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(FAVIM_VM_MANAGER_PID);
	DEBUG("FAVIM VM MANAGER: adding vm %i\n", vm->id);
	
	if(vm->id >= FAVIM_VM_COUNT){
		//TODO error handling
	} else {
		s->vms[vm->id] = vm;	
	}
}

//! Remove a VM by its id
static void remove_vm_by_id(FavimVMId id)
{
	DEBUG("FAVIM VM MANAGER: removing vm %i\n", id);
	
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(FAVIM_VM_MANAGER_PID);
	
	if(id >= FAVIM_VM_COUNT){
		//TODO error handling
	} else {
		SOS_CALL(s->vm_finalize, favim_vm_finalize_proto, s->vms[id]);
		s->vms[id]=NULL;
	}	
}

#ifndef _MODULE_
mod_header_ptr vm_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

