/**
 * @file favim.h
 * @brief Variable types in a FAViM virtual machine
 * @author Wouter Horr&eacute;
 */

#ifndef _FAVIM_TYPES_H_
#define _FAVIM_TYPES_H_

#include <led.h>
#include "favim_constants.h"
#include "sos_constants.h"

// Why doesn't C have booleans?
typedef unsigned char BOOL;
#define FALSE 0
#define TRUE 1

#define FAVIM_BUF_LEN 8
#define FAVIM_GLOBAL_COUNT 2

#define FAVIM_VM_COUNT 8
#define FAVIM_VM_ID_BITS 3
#define FAVIM_CONTEXT_COUNT 32
#define FAVIM_CONTEXT_ID_BITS 5

// typedefs
typedef struct FavimDataBuffer FavimDataBuffer;
typedef struct FavimBufferVariable FavimBufferVariable;
typedef struct FavimValueVariable FavimValueVariable;
typedef struct FavimVariable FavimVariable;
typedef uint8_t FavimVariableId;
typedef enum FavimVariableType FavimVariableType;
typedef enum FavimDataType FavimDataType;

//! An opcode in FAViM
typedef uint8_t FavimOpCode;

//! The global events in FAVIM
enum
{
	FAVIM_EVENT_GLOBAL_ADD_LIB,
	FAVIM_EVENT_GLOBAL_REMOVE_LIB,
};

//! The several variable types
enum FavimVariableType
{
	FAVIM_TYPE_NONE = 0,
	FAVIM_TYPE_BUFFER = 1,
	FAVIM_TYPE_INTEGER = 2,
};

//! The several data types
enum FavimDataType
{
	FAVIM_DATA_NONE = 0,
	FAVIM_DATA_INTEGER = 1,
};

//! A data buffer
struct FavimDataBuffer
{
	FavimDataType type;
	uint8_t size;
	int16_t entries[FAVIM_BUF_LEN];
}  __attribute__((packed));

//! A buffer variable
struct FavimBufferVariable
{
	FavimDataBuffer* var;
};

//! A value variable
struct FavimValueVariable
{
	int16_t var;
}  __attribute__((packed));

//! A variabele in FAViM
struct FavimVariable
{
	FavimVariableType type;
	union 
	{
		FavimValueVariable value;
		FavimBufferVariable buffer;
	};
} __attribute__((packed));


//! Set a chunk of memory to a given value
static inline void *nmemset(void *to, int val, size_t n);
static inline void *nmemset(void *to, int val, size_t n)
{
  char *cto = to;
  while (n--) *cto++ = val;
  return to;
}

//! Copy a chunk of memory to another
static inline void memcopy(char *dest, char *src, uint16_t length);
static inline void memcopy(char *dest, char *src, uint16_t length) 
{
	int i=0;
	for(i=0; i < length; i++)
		*(dest+i) = *(src+i);
	return;
}


//! Check the type of a FavimVariable
static inline bool type_check(FavimVariable * var, FavimVariableType type)
{
	return var->type==type;
}

#ifdef PC_PLATFORM
void * my_malloc(uint16_t size, sos_pid_t id);
#else
#define my_malloc(a,b) ker_malloc(a,b)
#endif

static inline void panic(){
#ifdef PC_PLATFORM
	DEBUG("*************************************************************\n");
	DEBUG("FAVIM PANIC *************************************************\n");
	DEBUG("*************************************************************\n");
#else
//	ker_led(LED_GREEN_OFF);
//	ker_led(LED_RED_OFF);
//	ker_led(LED_YELLOW_OFF);
//	ker_led(LED_GREEN_ON);
//	ker_led(LED_RED_ON);
//	ker_led(LED_YELLOW_ON);
//	ker_led(LED_YELLOW_OFF);
//	ker_led(LED_RED_OFF);
//	ker_led(LED_GREEN_OFF);
#endif
}

#endif

