/**
 * @file testlib.c
 * @brief The testing operation library
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <led.h>
#include <favim.h>
#include <operationstore/library.h>
#include <engine/scheduler.h>
#include <vmstore/vm.h>
#include <vmstore/context.h>
#include <engine/event_manager.h>

#define OP_TRIGGER 0

//! The state of the testing library
typedef struct {
	func_cb_ptr scheduler_opdone;
	func_cb_ptr stack_push;
	func_cb_ptr stack_pop;
	func_cb_ptr events_trigger;
	sos_pid_t pid;
} testlib_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static uint8_t bytelength(func_cb_ptr, FavimOpCode opcode);
static void uses_globals(func_cb_ptr, FavimOpCode opcode, uint8_t*);

//! The module header of the basic library
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_LIB_TESTLIB_PID,
	.state_size     = sizeof(testlib_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_LIB_TESTLIB_PID),
    .num_timers     = 0,
	.num_sub_func   = 4,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "vav1", FAVIM_SCHEDULER_PID, FAVIM_SCHEDULER_OPDONE_FID},
		{NULL, "cay2", FAVIM_STACK_PID, FAVIM_STACK_PUSH_FID},
		{NULL, "aav1", FAVIM_STACK_PID, FAVIM_STACK_POP_FID},
		{NULL, "cCC3", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_TRIGGER_EVENT_FID},
		{bytelength, "CCv1", FAVIM_LIB_TESTLIB_PID, FAVIM_LIB_BYTELENGTH_FID},
		{uses_globals, "vCD1", FAVIM_LIB_TESTLIB_PID, FAVIM_LIB_USES_GLOBALS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    testlib_state_t *s = (testlib_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("INIT TESTLIB: pid=%i\n", msg->did);
			s->pid = msg->did;
			ker_led(LED_YELLOW_TOGGLE);
			//ker_led(LED_YELLOW_TOGGLE);
			post_short(FAVIM_LIB_MANAGER_PID,s->pid,FAVIM_CMD_LIB_MANAGER_ADD_LIBRARY,0,0,0);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
        		post_short(FAVIM_LIB_MANAGER_PID,s->pid,FAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY,0,0,0);
			return SOS_OK;
        }

        case FAVIM_CMD_LIB_EXECUTE:
        {
        		int8_t ret = SOS_OK;
        		FavimExecutionMsg* exec = (FavimExecutionMsg*) msg->data;
			FavimContext* context = exec->context;
			FavimOpCode instr = exec->opcode;
						
			if(context->state == FAVIM_CONTEXT_STATE_RUN)
			{
				
				DEBUG("Instruction is %d\n",instr);
				
				switch(instr)
				{
					case OP_TRIGGER:
					{
//						ker_led(LED_YELLOW_TOGGLE);
//						ker_led(LED_YELLOW_TOGGLE);
//						ker_led(LED_YELLOW_TOGGLE);
//						ker_led(LED_YELLOW_TOGGLE);
						
						FavimVariable* arg = SOS_CALL(s->stack_pop, favim_stack_pop_proto, context->opstack);
						
						if ( ! type_check(arg, FAVIM_TYPE_INTEGER) )
						{
							DEBUG("FAVIM TESTLIB: type error in TRIGGER\n");
							ret = -EINVAL;
							break;
						}
						else 
						{
							uint8_t id = (uint8_t)arg->value.var;
							DEBUG("FAVIM TESTLIB: Context %i (vm %i) executed TRIGGER %i.\n", (int)context->which, (int)context->vm, (int)id);
							
							SOS_CALL(s->events_trigger, favim_event_manager_trigger_event_proto, FAVIM_EVENT_SINGLE_VM, context->vm, id);
						}
						
						break;
					}
				}
			}
			SOS_CALL(s->scheduler_opdone, favim_scheduler_opdone_proto, context);
					
			DEBUG("FAVIM TESTLIB: Instruction done, called opdone.\n");
			return ret;					
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the bytelength of the given opcode
static uint8_t bytelength(func_cb_ptr proto, FavimOpCode opcode)
{
	switch(opcode)
	{
		default:
			return 1;
	}	
}

//! Returns the globals used by the given opcode (bitmask)
static void uses_globals(func_cb_ptr proto, FavimOpCode opcode, uint8_t* uses)
{
	int i;
	for(i = 0; i < (FAVIM_GLOBAL_COUNT + 7)/8; ++i){
		uses[i]=0;
	}
}


#ifndef _MODULE_
mod_header_ptr testlib_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

