/**
 * @file library_manager.c
 * @brief The library manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <operationstore/library_manager.h>

#define LIBRARY_MASK 0x3f

//! The state of the library manager
typedef struct {
	sos_pid_t pid;
	FavimLibraryBitmask libraries;
} lib_man_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static BOOL lib_present(func_cb_ptr, FavimLibraryId);
static BOOL libs_present(func_cb_ptr, FavimLibraryBitmask);
static uint8_t get_library_pid(func_cb_ptr, FavimLibraryId);


//! The module header of the library manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_LIB_MANAGER_PID,
	.state_size     = sizeof(lib_man_state_t),
    .num_timers     = 0,
	.num_sub_func   = 0,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		{lib_present, "CCv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIB_PRESENT_FID},
		{libs_present, "Cyv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{get_library_pid, "CCv1", FAVIM_LIB_MANAGER_PID, FAVIM_LIB_MANAGER_GET_LIBRARY_PID_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    lib_man_state_t *s = (lib_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			s->libraries = 1; // Basic lib is always there
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_LIB_MANAGER_ADD_LIBRARY:
        {
        		FavimLibraryId lib_id = msg->sid & LIBRARY_MASK;
		    uint64_t lib_bit = 0x1 << lib_id;
		    s->libraries |= lib_bit;
		    //DEBUG("FAVIM LIBRARY MANAGER : Adding library id %d bit %016x. So libraries become %016x.\n",lib_id,(int)lib_bit,(int)s->libraries);
		    // TODO is reboot necessary?
		    // reboot(NULL);
        	return SOS_OK;
        }
        
        case FAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY:
        {
		    FavimLibraryId lib_id = msg->sid & LIBRARY_MASK;
		    uint64_t lib_bit = 0x1 << lib_id;
		    s->libraries ^= lib_bit;	//XOR
		    //DEBUG("FAVIM LIBRARY MANAGER : Removing library id %d bit %016x. So libraries become %016x.\n",lib_id,(int)lib_bit,(int)s->libraries);
		    // TODO Reboot ! Moet synchroon ? 
		    //reboot(NULL);
		    return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Is the library with a given id present?
static BOOL lib_present(func_cb_ptr proto, FavimLibraryId lib_id)
{
	if(lib_id < 64){
		FavimLibraryBitmask bitmask = 0x1 << lib_id;
		return libs_present(NULL, bitmask);
	}
	return FALSE;
}

//! Are all the libraries in the given bitmask present
static BOOL libs_present(func_cb_ptr proto, FavimLibraryBitmask bitmask)
{
	lib_man_state_t* s = (lib_man_state_t*) ker_get_module_state(FAVIM_LIB_MANAGER_PID);
	
	return ((bitmask & s->libraries) ==  bitmask);
}


//! Get the pid of a given library
static uint8_t get_library_pid(func_cb_ptr proto, FavimLibraryId lib_id)
{
	return (lib_id + FAVIM_LIB_MIN_PID);
}

#ifndef _MODULE_
mod_header_ptr library_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

