/**
 * @file vm.c
 * @brief A FAViM virtual machine
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <vmstore/vm.h>
#include <operationstore/library_manager.h>
#include <engine/event_manager.h>
#include <vmstore/concurrency.h>


//! The state a FAViM virtual machine
typedef struct {
	func_cb_ptr concurrency_submit;
	sos_pid_t pid;
	FavimVM vm;
} vm_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static FavimLibraryId get_library(func_cb_ptr,FavimOpCode);
static FavimLibraryId get_opcode(func_cb_ptr,FavimOpCode);

static FavimContext* get_context_from_id(FavimEventId);

//! The module header of the VM module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VM_PID,
	.state_size     = sizeof(vm_state_t),
    .num_timers     = 0,
	.num_sub_func   = 1,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "cav1", FAVIM_CONC_PID, FAVIM_CONC_SUBMIT_FID},
		{get_library, "CCv1",FAVIM_VM_PID,FAVIM_VM_GET_LIBRARY_FID},
		{get_opcode, "CCv1",FAVIM_VM_PID,FAVIM_VM_GET_OPCODE_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    vm_state_t *s = (vm_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("FAVIM: VM INITIALIZATION\n");
			s->pid = msg->did;
			//TODO decent implementation :)
			// Initialise the VM
			
			DEBUG("FAVIM VM: fake initializing vm\n");
			// initialise to something stupid for now
			s->vm.contexts = (FavimContext**) ker_malloc(2*sizeof(FavimContext*),FAVIM_VM_PID);
			s->vm.globals = NULL;
			
			DEBUG("FAVIM VM: context 0\n");
			// initialise the contexts
			s->vm.contexts[0] = (FavimContext*)ker_malloc(sizeof(FavimContext),FAVIM_VM_PID);
			s->vm.contexts[0]->pc = 0;
			s->vm.contexts[0]->state = 0;
			s->vm.contexts[0]->num_executed = 0;
			s->vm.contexts[0]->which = REBOOT_CONTEXT_ID;
			s->vm.contexts[0]->vm = &(s->vm);
			s->vm.contexts[0]->capsule = (FavimCapsule*) ker_malloc(sizeof(FavimCapsule),FAVIM_VM_PID);
			s->vm.contexts[0]->capsule->dataSize = 5;
			s->vm.contexts[0]->capsule->data[0] = 0x01; // only basiclib
			s->vm.contexts[0]->capsule->data[1] = 0x01; //pushc6
			s->vm.contexts[0]->capsule->data[2] = 0x0a; //10
			s->vm.contexts[0]->capsule->data[3] = 0x0d; //settimer0 repeat
			s->vm.contexts[0]->capsule->data[4] = 0x00; //halt
			s->vm.contexts[0]->opstack = (FavimOperandStack*) ker_malloc(sizeof(FavimOperandStack),FAVIM_VM_PID);
			s->vm.contexts[0]->opstack->sp = 0;
			s->vm.contexts[0]->queue = NULL;
			
			DEBUG("FAVIM VM: context 1\n");
			s->vm.contexts[1] = (FavimContext*)ker_malloc(sizeof(FavimContext),FAVIM_VM_PID);
			s->vm.contexts[1]->pc = 0;
			s->vm.contexts[1]->state = 0;
			s->vm.contexts[1]->num_executed = 0;
			s->vm.contexts[1]->which = TIMER0_CONTEXT_ID;
			s->vm.contexts[1]->vm = &(s->vm);
			s->vm.contexts[1]->capsule = (FavimCapsule*) ker_malloc(sizeof(FavimCapsule),FAVIM_VM_PID);
			s->vm.contexts[1]->capsule->dataSize = 5;
			s->vm.contexts[1]->capsule->data[0] = 0x01; // only basiclib
			s->vm.contexts[1]->capsule->data[1] = 0x01; //pushc6
			s->vm.contexts[1]->capsule->data[2] = 0x1f; //31
			s->vm.contexts[1]->capsule->data[3] = 0x02; //led
			s->vm.contexts[1]->capsule->data[4] = 0x00; //halt
			s->vm.contexts[1]->opstack = (FavimOperandStack*) ker_malloc(sizeof(FavimOperandStack),FAVIM_VM_PID);
			s->vm.contexts[1]->opstack->sp = 0;
			s->vm.contexts[1]->queue = NULL;
			
			DEBUG("FAVIM VM: sending reboot message to self\n");
			post_short(FAVIM_VM_PID,FAVIM_VM_PID,FAVIM_CMD_VM_REBOOT,0,0,0);
			
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_VM_ADD_CONTEXT:
        {
	        	//TODO implementation
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_REMOVE_CONTEXT:
        {
	        	//TODO implementation
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_EVENT_OCCURED:
        {
	        	// Lookup the context associated with the event
	        	MsgParam *rcvd = (MsgParam *)msg->data;		
	        	FavimEventId id = (FavimEventId)rcvd->byte;
	        	FavimContext* context = get_context_from_id(id);
	        	// Submit the context for execution
	        	SOS_CALL(s->concurrency_submit, favim_concurrency_submit_proto, context);	
	        	return SOS_OK;
        }
        
        case FAVIM_CMD_VM_REBOOT:
        {
	        	//TODO implementation
	        	// Cancel execution of all contexts
	        	// Reset the VM
	        	// Submit the reboot context for execution
	        SOS_CALL(s->concurrency_submit,favim_concurrency_submit_proto, s->vm.contexts[REBOOT_CONTEXT_ID]);
	        	return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the library id from an instruction
static FavimLibraryId get_library(func_cb_ptr proto, FavimOpCode instr)
{
    return ((instr >> OPCODE_BITS) & LIBRARY_MASK);	
}

//! Get the opcode from an instruction
static FavimLibraryId get_opcode(func_cb_ptr proto, FavimOpCode instr)
{
    return (instr & OPCODE_MASK);
}

//! Get the context associated with the given event id
// For now contextid==eventid
static FavimContext* get_context_from_id(FavimEventId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(FAVIM_VM_PID);
	return s->vm.contexts[id];
}

#ifndef _MODULE_
mod_header_ptr vm_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

