/**
 * @file capsule_manager.c
 * @brief The implementation of the FAViM Capsule Manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <capsulestore/capsule_manager.h>

typedef struct HashEntry HashEntry;

struct HashEntry {
	FavimCapsuleId id;
	FavimCapsule* capsule;
	HashEntry* next;
};

#define HASH_MASK 0x07

//! The state of the capsule manager
typedef struct {
	sos_pid_t pid;
	HashEntry * capsules[8];
} capsule_manager_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t add_capsule(func_cb_ptr, FavimCapsuleId, FavimCapsule*);
static int8_t remove_capsule(func_cb_ptr, FavimCapsuleId);

static inline void insert_he(HashEntry * he);

//! The module header of ...
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_CAPSULE_MANAGER_PID,
	.state_size     = sizeof(capsule_manager_state_t),
    .num_timers     = 0,
	.num_sub_func   = 0,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		{add_capsule, "cya2", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_ADD_CAPSULE_FID},
		{remove_capsule, "cyv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_REMOVE_CAPSULE_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    capsule_manager_state_t *s = (capsule_manager_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			s->pid = msg->did;
			for(i=0; i < 8; ++i){
				s->capsules[i] = NULL;
			}
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Add the given capsule with the given version
static int8_t add_capsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule)
{
	HashEntry * he = (HashEntry*) ker_malloc(sizeof(HashEntry),FAVIM_CAPSULE_MANAGER_PID);
	if(he == NULL){
		//TODO error handling
		return -EINVAL;
	}
	he->id = id;
	he->capsule = capsule;
	he->next = NULL;
	insert_he(he);
	//TODO add in Virus
	return FAVIM_OK;
}

//! Remove the capsule with the specified id
static int8_t remove_capsule(func_cb_ptr proto, FavimCapsuleId id)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	if(s->capsules[hid] == NULL){
		DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i", id);
		//TODO error handling
		return -EINVAL;
	} else if(s->capsules[hid]->id == id) {
		HashEntry * prev = s->capsules[hid];
		s->capsules[hid] = s->capsules[hid]->next;
		ker_free(prev);
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL && current->next->id != id){
			current = current->next;
		}
		if(current->next == NULL){
			DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i", id);
			return -EINVAL;
			//TODO error handling
		} else {
			HashEntry * prev = current->next;
			current->next = current->next->next;
			ker_free(prev);
		}
	}
	//TODO remove from Virus
	return FAVIM_OK;
}

static inline void insert_he(HashEntry * he)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = (he->id & HASH_MASK);
	if(s->capsules[hid] == NULL){
		s->capsules[hid] = he;
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL){
			current = current->next;
		}
		current->next = he;
	}
}

#ifndef _MODULE_
mod_header_ptr capsule_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

