/**
 * @file event_manager.c
 * @brief The FAViM event manager
 * @author Wouter Horr&eacute;
 */

#include <engine/event_manager.h>

//! The state of the FAViM event manager
typedef struct {
	sos_pid_t pid;
	FavimEventTimer* timerlist;
//	uint8_t timercount;
} event_man_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t start_timer(func_cb_ptr,FavimEventId,FavimTimerType,FavimTimerInterval);
static int8_t stop_timer(func_cb_ptr,FavimEventId);
//static int8_t init_event(func_cb_ptr,FavimEventId);
static int8_t trigger_event(func_cb_ptr,FavimEventId);
static int8_t insert_into_timer_list(FavimEventTimer*);
static void tick(event_man_state_t*);

//! The module header of the event manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_EVENT_MANAGER_PID,
	.state_size     = sizeof(event_man_state_t),
    .num_timers     = 1,
	.num_sub_func   = 0,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		{start_timer,"cCC3",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_START_TIMER_FID},
		{stop_timer,"cCv1",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_STOP_TIMER_FID},
//		{init_event,"cCv1",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_INIT_EVENT_FID},
		{trigger_event,"cCv1",FAVIM_EVENT_MANAGER_PID,FAVIM_EVENT_MANAGER_TRIGGER_EVENT_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    event_man_state_t *s = (event_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("FAVIM: EVENT MANAGER INITIALIZATION\n");
			s->pid = msg->did;
			s->timerlist = NULL;
			
			ker_timer_init(s->pid, FAVIM_EVENT_TIMER_ID, TIMER_REPEAT);
			
			return SOS_OK;
		}

        case MSG_FINAL:
        {
        	//TODO is memory cleaned up automatically ?
			return SOS_OK;
        }
        
        case MSG_TIMER_TIMEOUT:
        {
        		tick(s);
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Start a timer
//! TODO This isn't very exact :S
static int8_t start_timer(func_cb_ptr proto, FavimEventId id, FavimTimerType type, FavimTimerInterval interval)
{
	FavimEventTimer* newTimer = (FavimEventTimer*) ker_malloc(sizeof(FavimEventTimer),FAVIM_EVENT_MANAGER_PID);
	newTimer->id = id;
	newTimer->ticks.var = interval;
	newTimer->next = NULL;
	if(type == FAVIM_TIMER_REPEAT) {
		DEBUG("FAVIM: EVENT MANAGER STARTING TIMER %i AT RATE %i\n",id,interval);
		newTimer->interval.var=interval;
	} else {
		DEBUG("FAVIM: EVENT MANAGER STARTING TIMER %i ONCE AT %i TICKS\n",id,interval);
		newTimer->interval.var=0;
	}
	return insert_into_timer_list(newTimer);
}

//! Stop a timer
static int8_t stop_timer(func_cb_ptr proto,FavimEventId id)
{
	DEBUG("FAVIM: EVENT MANAGER STOPPING TIMER %i\n", id);
    event_man_state_t *s = (event_man_state_t *)ker_get_module_state(FAVIM_EVENT_MANAGER_PID);
    	
	FavimEventTimer* current = s->timerlist;
	FavimEventTimer** previous = &(s->timerlist);
	while(current != NULL) {
		if(current->id == id){
			current->next->ticks.var += current->ticks.var;
			*previous = current->next;
			ker_free(current);
			//! If we removed the last timer, stop the kernel timer
			if(s->timerlist == NULL){
				ker_timer_stop(s->pid, FAVIM_EVENT_TIMER_ID);
			}
			break;
		} else {
			previous = &(current->next);
			current = current->next;
		}
	}
	return FAVIM_OK;
}

//! Init an event
//static int8_t init_event(func_cb_ptr proto, FavimEventId id)
//{
//	//TODO implementation
//	return FAVIM_OK;
//}

//! Trigger an event
static int8_t trigger_event(func_cb_ptr proto, FavimEventId id)
{
	DEBUG("FAVIM: EVENT MANAGER TRIGGER EVENT %i\n",id);
	return post_short(FAVIM_VM_PID,FAVIM_EVENT_MANAGER_PID,FAVIM_CMD_VM_EVENT_OCCURED,id,0,0);
}

//! Insert a timer into the timerlist
static int8_t insert_into_timer_list(FavimEventTimer* new)
{
	DEBUG("FAVIM: EVENT MANAGER ADDING TIMER %i TO TIMER LIST\n",new->id);
    event_man_state_t *s = (event_man_state_t *)ker_get_module_state(FAVIM_EVENT_MANAGER_PID);
	
	//! If the new timer is the only one, start the kernel timer
	//! Apparently interval 1024 == 1 second, we want 0.1 seconds
	if(s->timerlist == NULL){
		s->timerlist = new;
		ker_timer_start(s->pid,FAVIM_EVENT_TIMER_ID,102);
	} else {
		FavimEventTimer* current = s->timerlist;
		FavimEventTimer** previous = &(s->timerlist);
		while(current->id != new->id && current->ticks.var < new->ticks.var){
			new->ticks.var -= current->ticks.var;
			previous = &(current->next);
			current = current->next;
		}
		if(current->id == new->id){
			//! There is already a timer with the same id!
			//! Cleanup and return error
			ker_free(new);
			return -FAVIM_ERROR_TIMER_EXISTS;
		} else {
			*previous = new;
			new->next = current;
		}
	}
	return FAVIM_OK;
}

//! Advance one tick
static void tick(event_man_state_t *s){
	DEBUG("FAVIM: EVENT MANAGER TICK\n");
	s->timerlist->ticks.var--;
	while(s->timerlist->ticks.var==0){
		FavimEventTimer* current = s->timerlist;
		trigger_event(0,current->id);
		//! Remove the current timer from the timerlist
		s->timerlist = current->next;
		//! Reschedule the timer if needed, else cleanup
		if(current->interval.var != 0){
			current->ticks.var = current->interval.var;
			if(insert_into_timer_list(current)){
				//THIS SHOULD NOT HAPPEN
				//TODO raise error
			}
		} else {
			ker_free(current);
		}
	}
}

#ifndef _MODULE_
mod_header_ptr event_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

