/**
 * @file basiclib.c
 * @brief The basic operation library
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <led.h>
#include <favim.h>
#include <operationstore/library.h>
#include <engine/scheduler.h>
#include <vmstore/vm.h>
#include <vmstore/context.h>
#include <engine/event_manager.h>

#define OP_HALT 0
#define OP_PUSHC6 1
#define OP_LED 2
#define OP_SETTIMER0 0x0c

//! The state of the basic library
typedef struct {
	func_cb_ptr scheduler_opdone;
	func_cb_ptr stack_push;
	func_cb_ptr stack_pop;
	func_cb_ptr events_start_timer;
	func_cb_ptr events_stop_timer;
	sos_pid_t pid;
} basiclib_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static uint8_t bytelength(func_cb_ptr, FavimOpCode opcode);
static void uses_globals(func_cb_ptr, FavimOpCode opcode, uint8_t*);

//! The module header of the basic library
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_LIB_BASICLIB_PID,
	.state_size     = sizeof(basiclib_state_t),
    .num_timers     = 0,
	.num_sub_func   = 5,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "vav1", FAVIM_SCHEDULER_PID, FAVIM_SCHEDULER_OPDONE_FID},
		{NULL, "cay2", FAVIM_STACK_PID, FAVIM_STACK_PUSH_FID},
		{NULL, "aav1", FAVIM_STACK_PID, FAVIM_STACK_POP_FID},
		{NULL, "cCC3", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_START_TIMER_FID},
		{NULL, "cCv1", FAVIM_EVENT_MANAGER_PID, FAVIM_EVENT_MANAGER_STOP_TIMER_FID},
		{bytelength, "CCv1", FAVIM_LIB_BASICLIB_PID, FAVIM_LIB_BYTELENGTH_FID},
		{uses_globals, "vCD1", FAVIM_LIB_BASICLIB_PID, FAVIM_LIB_USES_GLOBALS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    basiclib_state_t *s = (basiclib_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			post_short(FAVIM_LIB_MANAGER_PID,s->pid,FAVIM_CMD_LIB_MANAGER_ADD_LIBRARY,0,0,0);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
        		post_short(FAVIM_LIB_MANAGER_PID,s->pid,FAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY,0,0,0);
			return SOS_OK;
        }

        case FAVIM_CMD_LIB_EXECUTE:
        {
        		FavimExecutionMsg* exec = (FavimExecutionMsg*) msg->data;
			FavimContext* context = exec->context;
			FavimOpCode instr = exec->opcode;
						
			if(context->state == FAVIM_CONTEXT_STATE_RUN)
			{
				
				DEBUG("Instruction is %d\n",instr);
				
				switch(instr)
				{
					case OP_HALT:
					{
						DEBUG("FAVIM BASICLIB: Context %i executed HALT.\n", (int)context->which);
						context->state = FAVIM_CONTEXT_STATE_HALT;
//						context->pc = 0;
//						SOS_CALL(s->synch_haltContext, func_vp_t, context);
//						
						break;
					}
					case OP_PUSHC6:
					{
						context->pc++;
						
						FavimOpCode arg = context->capsule->data[context->pc];
						FavimVariable var;
						var.type = FAVIM_TYPE_INTEGER;
						var.value.var = arg;
						DEBUG("FAVIM BASICLIB: Context %i: Executing pushc6 with arg %hi\n", (int)context->which, arg);

						SOS_CALL(s->stack_push, favim_stack_push_proto, context->opstack, var);
						break;
					}
					case OP_LED:
					{
						FavimVariable* arg = SOS_CALL(s->stack_pop, favim_stack_pop_proto, context->opstack);
	
						DEBUG("Inside OPled\n");
						
						//TODO typechecking
//						if ( !SOS_CALL(s->types_checkTypes, func_u8zzu8_t, context, arg, MATE_TYPE_INTEGER) )
//						{
//							DEBUG("Problem in OPled\n");
//							break;
//						}
//						else 
//						{
							uint16_t val = arg->value.var;
							uint8_t op = (val >> 3) & 3;
							uint8_t led	= val & 7;
							DEBUG("FAVIM BASICLIB: Context %i EXECUTING OP_LED with op %i and led %i.\n", (int)context->which, (int)op, (int)led);
		      
							switch (op) 
							{
								case 0:			/* set */
								{
									if (led & 1) 
										ker_led(LED_RED_ON);
									else 
										ker_led(LED_RED_OFF);
									if (led & 2)
										ker_led(LED_GREEN_ON);
									else
										ker_led(LED_GREEN_OFF);
									if (led & 4)
										ker_led(LED_YELLOW_ON);
									else
										ker_led(LED_YELLOW_OFF);
									break;
								}
								case 1:			/* OFF 0 bits */
								{
									if (!(led & 1)) ker_led(LED_RED_OFF);
									if (!(led & 2)) ker_led(LED_GREEN_OFF);
									if (!(led & 4)) ker_led(LED_YELLOW_ON);
									break;
								}
	      						case 2:			/* on 1 bits */
	      						{
									if (led & 1) ker_led(LED_RED_ON);
									if (led & 2) ker_led(LED_GREEN_ON);
									if (led & 4) ker_led(LED_YELLOW_ON);
									break;
								}
								case 3:			/* TOGGLE 1 bits */
								{
									if (led & 1) ker_led(LED_RED_TOGGLE);
									if (led & 2) ker_led(LED_GREEN_TOGGLE);
									if (led & 4) ker_led(LED_YELLOW_TOGGLE);
									break;
								}
								default:
								{
									DEBUG("FAVIM BASICLIB: LED command had unknown operations.\n");
									break;
								}
							}
//						}
						break;
					}
					case OP_SETTIMER0 + 0: 
					case OP_SETTIMER0 + 1:
					{
//						uint32_t msec;
						uint8_t timer_type = (instr-OP_SETTIMER0)?TIMER_REPEAT:TIMER_ONE_SHOT;
						FavimVariable* arg = SOS_CALL(s->stack_pop, favim_stack_pop_proto, context->opstack);
						
						DEBUG("FAVIM BASICLIB: Setting timer 0 rate.\n");
						//TODO typechecking
//						if (!SOS_CALL(s->types_checkTypes, func_u8zzu8_t, context, arg, MATE_TYPE_INTEGER)) {return -EINVAL;}
//						msec = 102 * arg->value.var + (4 * arg->value.var) / 10; 
						
						SOS_CALL(s->events_stop_timer, favim_event_manager_stop_timer_proto, TIMER0_CONTEXT_ID);
						if (arg->value.var > 0) {
							SOS_CALL(s->events_start_timer, favim_event_manager_start_timer_proto, TIMER0_CONTEXT_ID, timer_type, arg->value.var);
						}
						
						break;
					}
				}
			}
			SOS_CALL(s->scheduler_opdone, favim_scheduler_opdone_proto, context);
					
			DEBUG("FAVIM BASICLIB: Instruction done, called opdone.\n");
			return SOS_OK;					
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the bytelength of the given opcode
static uint8_t bytelength(func_cb_ptr proto, FavimOpCode opcode)
{
	switch(opcode)
	{
		case OP_PUSHC6:
//		case OP_JNZ:
		{
			return 2;
		}
//		case OP_CALL:
//		case OP_POST:
//		{
//			return 3;
//		}
		default:
			return 1;
	}	
}

//! Returns the globals used by the given opcode (bitmask)
static void uses_globals(func_cb_ptr proto, FavimOpCode opcode, uint8_t* uses)
{
	//TODO
	//Nothing to do for the moment, only clearing out uses, change when implementing other
	int i;
	for(i = 0; i < (FAVIM_GLOBAL_COUNT + 7)/8; ++i){
		uses[i]=0;
	}
}


#ifndef _MODULE_
mod_header_ptr basic_lib_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

