/**
 * @file capsule_manager.c
 * @brief The implementation of the FAViM Capsule Manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <capsulestore/capsule_manager.h>
#include <vmstore/vm.h>
#include <vmstore/concurrency.h>

typedef struct HashEntry HashEntry;

struct HashEntry {
	FavimCapsuleId id;
	FavimCapsule* capsule;
	HashEntry* next;
};

#define HASH_MASK 0x07

//! The state of the capsule manager
typedef struct {
	func_cb_ptr vm_halt;
	func_cb_ptr analyse_capsule;
	sos_pid_t pid;
	HashEntry * capsules[8];
} capsule_manager_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t add_capsule(func_cb_ptr, FavimCapsuleId, FavimCapsule*);
static int8_t remove_capsule(func_cb_ptr, FavimCapsuleId);

// helper functions
static inline void insert_he(HashEntry * he);
static inline void capsule_arrived(FavimCapsuleId, FavimCapsule *);

//! The module header of ...
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_CAPSULE_MANAGER_PID,
	.state_size     = sizeof(capsule_manager_state_t),
    .num_timers     = 0,
	.num_sub_func   = 2,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO ERROR HANDLING
		{NULL, "vvv0", FAVIM_VM_PID, FAVIM_VM_HALT_FID},
		{NULL, "vav1", FAVIM_CONC_PID, FAVIM_CONC_ANALYSE_CAPSULE_FID},
		{add_capsule, "cya2", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_ADD_CAPSULE_FID},
		{remove_capsule, "cyv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_REMOVE_CAPSULE_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    capsule_manager_state_t *s = (capsule_manager_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			s->pid = msg->did;
			for(i=0; i < 8; ++i){
				s->capsules[i] = NULL;
			}
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
        case FAVIM_CMD_CAPS_MANAGER_CAPS_ARRIVED:
        {
        		FavimCapsuleMsg * m = (FavimCapsuleMsg *) msg->data;
        		FavimCapsuleId id = m->id;
        		FavimCapsule * capsule = &(m->capsule);
        		capsule_arrived(id,capsule);
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Add the given capsule with the given version
static int8_t add_capsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule)
{
	DEBUG("FAVIM CAPSULE MANAGER: adding capsule %i\n",id);
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);

	HashEntry * he = (HashEntry*) ker_malloc(sizeof(HashEntry),FAVIM_CAPSULE_MANAGER_PID);
	if(he == NULL){
		//TODO error handling
		return -EINVAL;
	}
	he->id = id;
	he->capsule = capsule;
	he->next = NULL;
	insert_he(he);
	//TODO dit is niet echt nodig, enkel hier om even te testen of dit ook werkt
	SOS_CALL(s->analyse_capsule, favim_concurrency_analyse_capsule_proto, capsule);
	//TODO add in Virus
	return FAVIM_OK;
}

//! Remove the capsule with the specified id
static int8_t remove_capsule(func_cb_ptr proto, FavimCapsuleId id)
{
	DEBUG("FAVIM CAPSULE MANAGER: removing capsule %i\n",id);
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	if(s->capsules[hid] == NULL){
		DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i", id);
		//TODO error handling
		return -EINVAL;
	} else if(s->capsules[hid]->id == id) {
		HashEntry * prev = s->capsules[hid];
		s->capsules[hid] = s->capsules[hid]->next;
		ker_free(prev);
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL && current->next->id != id){
			current = current->next;
		}
		if(current->next == NULL){
			DEBUG("FAVIM CAPSULEMANAGER: tried to remove a non-existent capsule %i", id);
			return -EINVAL;
			//TODO error handling
		} else {
			HashEntry * prev = current->next;
			current->next = current->next->next;
			ker_free(prev);
		}
	}
	//TODO remove from Virus
	return FAVIM_OK;
}

static inline void insert_he(HashEntry * he)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	uint8_t hid = (he->id & HASH_MASK);
	if(s->capsules[hid] == NULL){
		s->capsules[hid] = he;
	} else {
		HashEntry * current = s->capsules[hid];
		while(current->next != NULL){
			current = current->next;
		}
		current->next = he;
	}
}

static inline void capsule_arrived(FavimCapsuleId id, FavimCapsule * capsule)
{
	capsule_manager_state_t * s = (capsule_manager_state_t*) ker_get_module_state(FAVIM_CAPSULE_MANAGER_PID);
	
	// Halt the VM: this is a function call
	// The VM has to halt immediately
	SOS_CALL(s->vm_halt, favim_vm_halt_proto);

	// Install the capsule
	uint8_t hid = (id & HASH_MASK);
	HashEntry * current = s->capsules[hid];
	while(current != NULL && current->id != id) {
		current = current->next;
	}
	if(current != NULL){
		int i;
		current->capsule->dataSize = capsule->dataSize;
		for(i=0; i < FAVIM_CAPSULE_SIZE; ++i){
			current->capsule->data[i] = capsule->data[i];
		}
		current->capsule->version = capsule->version;
		// Analyse the capsule to initialize usedVars
		SOS_CALL(s->analyse_capsule, favim_concurrency_analyse_capsule_proto, current->capsule);
	}
	
	// Reboot the VM: this is a message
	// SOS delivers messages FIFO, so any pending EXECUTE messages to a lib will be discarded
	// by the lib because the VM is still halted
	post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REBOOT,0,0,0);
}

#ifndef _MODULE_
mod_header_ptr capsule_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

