/**
 * @file context.c
 * @brief A execution context in a FAViM virtual machine
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <vmstore/context.h>
#include <util/favim_queue.h>

//! The state of the context module
typedef struct {
	func_cb_ptr stack_new;
	func_cb_ptr stack_reset;
	func_cb_ptr capsule_new;
	sos_pid_t pid;
} context_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
//static void set_pc(func_cb_ptr, FavimContext*, FavimCapsuleIndex);
//static void set_local(func_cb_ptr, FavimContext*, FavimVariableId, FavimVariable*);
//static void set_capsule(func_cb_ptr, FavimContext*, FavimCapsule*);
////op niveau individuele lock?
//static void set_use_set(func_cb_ptr, FavimContext*, uint8_t*);
//static void set_acquire_set(func_cb_ptr, FavimContext*, uint8_t*);
//static void set_release_set(func_cb_ptr, FavimContext*, uint8_t*);

static FavimContext* new(func_cb_ptr, FavimVM*, FavimContextId);
static void reset(func_cb_ptr, FavimContext*);

//! The module header of the context module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_CONTEXT_PID,
	.state_size     = sizeof(context_state_t),
    .num_timers     = 0,
	.num_sub_func   = 3,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		// TODO error handler
		{NULL, "avv0", FAVIM_STACK_PID, FAVIM_STACK_NEW_FID},
		{NULL, "vav1", FAVIM_STACK_PID, FAVIM_STACK_RESET_FID},
		{NULL, "avv0", FAVIM_CAPSULE_PID, FAVIM_CAPSULE_NEW_FID},
		{new, "aaC2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_NEW_FID},
//		{set_pc, "vay2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_PC_FID},
//		{set_local, "vay3", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_LOCAL_FID},
//		{set_capsule, "vaa2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_CAPSULE_FID},
//		{set_use_set, "vaE2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_USE_SET_FID},
//		{set_acquire_set, "vaE2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_ACQUIRE_SET_FID},
//		{set_release_set, "vaE2", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_SET_RELEASE_SET_FID},
		{reset, "vav1", FAVIM_CONTEXT_PID, FAVIM_CONTEXT_RESET_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    context_state_t *s = (context_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

////! Set the program counter
//static void set_pc(func_cb_ptr proto, FavimContext* context, FavimCapsuleIndex pc)
//{
//	//TODO implementation
//}
//
////! Set a variable local to the context
//static void set_local(func_cb_ptr proto, FavimContext* context, FavimVariableId id, FavimVariable* value)
//{
//	//TODO implementation
//}
//
////! Set the capsule executing in this context
//static void set_capsule(func_cb_ptr proto, FavimContext* context, FavimCapsule* capsule)
//{
//	//TODO implementation
//}
//
////op niveau individuele lock?
////! Set the set of used locks
//static void set_use_set(func_cb_ptr proto, FavimContext* context, uint8_t* set)
//{
//	//TODO implementation
//}
//
////! Set the set of locks to acquire
//static void set_acquire_set(func_cb_ptr proto, FavimContext* context, uint8_t* set)
//{
//	//TODO implementation
//}
//
////! Set the set of locks to release
//static void set_release_set(func_cb_ptr proto, FavimContext* context, uint8_t* set)
//{
//	//TODO implementation
//}


//! Allocate and initialise a new context
static FavimContext* new(func_cb_ptr proto, FavimVM* vm, FavimContextId which)
{
	context_state_t * s = (context_state_t*) ker_get_module_state(FAVIM_CONTEXT_PID);
	
	FavimContext* new_con = (FavimContext*) ker_malloc(sizeof(FavimContext),s->pid);
	if(new_con == 0){
		DEBUG("CONTEXT NEW: Something is very very wrong here\n");
		return 0;
	}
	new_con->capsule = SOS_CALL(s->capsule_new, favim_capsule_new_proto);
	new_con->opstack = SOS_CALL(s->stack_new, favim_stack_new_proto);
	new_con->vm = vm;
	new_con->which = which;
	new_con->queue = NULL;
	reset(0,new_con);
	return new_con;
}

//! Reset the given context
static void reset(func_cb_ptr proto, FavimContext* context)
{
	
	// Do not touch the capsule, leave that to the capsule manager!
	
	context_state_t * s = (context_state_t*) ker_get_module_state(FAVIM_CONTEXT_PID);
	int i;
	
	context->pc=0;
	context->state=FAVIM_CONTEXT_STATE_HALT;
	context->num_executed=0;
	if(context->queue){
		queue_remove(context, context->queue, context);
	}
	SOS_CALL(s->stack_reset, favim_stack_reset_proto, context->opstack);
	for(i=0; i < (FAVIM_GLOBAL_COUNT + 7)/8; ++i){
		context->heldSet[i] = 0;
		context->acquireSet[i] = 0;
		context->releaseSet[i] = 0;
	}
	for(i=0; i < FAVIM_LOCAL_COUNT; ++i){
		context->locals[i] = NULL;
	}
}

#ifndef _MODULE_
mod_header_ptr context_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

