/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file scheduler.c
 * @brief The FAViM scheduler
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <engine/scheduler.h>
#include <vmstore/context.h>
#include <vmstore/vm.h>
#include <vmstore/concurrency.h>
#include <operationstore/library_manager.h>
#include <operationstore/library.h>
#include <util/davim_queue.h>
#include <capsulestore/capsule.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t submit(func_cb_ptr, DavimContext*);
//static void block(func_cb_ptr proto, DavimContext* context);
//static void unblock(func_cb_ptr proto, DavimContext* context);
static void run_task();
//static void reboot(func_cb_ptr);
static void opdone(func_cb_ptr, DavimContext*);
static int8_t execute(DavimContext*);
static inline int8_t computeInstruction(DavimContext*);


//! The module header of the scheduler
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_SCHEDULER_PID,
	.state_size     = sizeof(scheduler_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_SCHEDULER_PID),
    .num_timers     = 0,
	.num_sub_func   = 8,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "CCC2",DAVIM_VM_PID,DAVIM_VM_GET_LIBRARY_FID},
		{NULL, "CCv1",DAVIM_VM_PID,DAVIM_VM_GET_OPCODE_FID},
		{NULL, "YCC2",DAVIM_VM_PID, DAVIM_VM_GET_NEEDED_LIBS_FID},
		{NULL, "CCv1", DAVIM_LIB_MANAGER_PID, DAVIM_LIB_MANAGER_GET_LIBRARY_PID_FID},
		{NULL, "Cyv1", DAVIM_LIB_MANAGER_PID, DAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{NULL, "vav1", DAVIM_CONC_PID, DAVIM_CONC_HALT_FID},
		{NULL, "cav1", DAVIM_CONC_PID, DAVIM_CONC_RESUME_FID},
		{NULL, "vav1", DAVIM_CONC_PID, DAVIM_CONC_YIELD_FID},
		{submit, "cav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_SUBMIT_FID},
		{opdone, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_OPDONE_FID},
//		{block, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_BLOCK_FID},
//		{unblock, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_UNBLOCK_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    scheduler_state_t *s = (scheduler_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("DAVIM: scheduler initialization\n");
			s->pid = msg->did;
//			s->flags.inErrorState = FALSE;
			s->flags.halted = FALSE;
			s->flags.taskRunning = FALSE;
//			s->flags.errorFlipFlop = 0;
//			s->libraries = 1;	//Basic library is already there.
			s->runningContext = NULL;
//			ker_timer_init(s->pid, ERROR_TIMER, TIMER_REPEAT);
//			reboot(0);
			queue_init(&(s->runQueue));
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
//        case DAVIM_CMD_SCHEDULER_HALT:
//        {
//			DEBUG("DAVIM: scheduler halted.\n");
//			s->flags.halted = TRUE;
//        		return SOS_OK;
//        }
        
//        case DAVIM_CMD_SCHEDULER_RESUME:
//        {
//			s->flags.halted = FALSE;
//			DEBUG("DAVIM: scheduler resumes....\n");
//			if (!s->flags.taskRunning) {
//			    s->flags.taskRunning = TRUE;
//			    DEBUG("DAVIM SCHEDULER: run_task called...\n");
//			    run_task();
//			}
//        		return SOS_OK;
//        }

	    case DAVIM_CMD_SCHEDULER_RUN_TASK:
	    {
			run_task();
			return SOS_OK;
	    }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Submit a context for execution
static int8_t submit(func_cb_ptr proto, DavimContext* context)
{
    DEBUG("DAVIM SCHEDULER: VM %i, Context %i submitted to run.\n", (int)context->vm, (int)context->which);
    if(context->state != DAVIM_CONTEXT_STATE_UNBLOCKED) { // If this is not a context that is resumed after blocking
    		//Reset program counter and execution counter
	    context->pc = 0;
	    context->num_executed = 0;		
    }
    context->state = DAVIM_CONTEXT_STATE_READY;
    return execute(context);
}


static void opdone(func_cb_ptr proto, DavimContext* caller)
{
	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);

	DEBUG("DAVIM SCHEDULER: opdone starts running...\n");
	
	if(caller->state == DAVIM_CONTEXT_STATE_RUN){
		caller->pc++;
		caller->num_executed++;
		if(caller == s->runningContext){
			run_task();
		} else {
			//should not happen!
			DEBUG("DAVIM SCHEDULER: Context executing opdone is not the running context!\n");
		}
	} else if(caller->state == DAVIM_CONTEXT_STATE_HALT) {
		// The caller has executed the halt instruction, signal to concurrency
		DEBUG("DAVIM SCHEDULER (opdone): The calling context has executed halt, signaling to concurrency\n");
		SOS_CALL(s->concurrency_halt, davim_concurrency_halt_proto, caller);
		if(caller == s->runningContext){
			s->runningContext = NULL;
			run_task();
		} else {
			//should not happen!
			DEBUG("DAVIM SCHEDULER: Halted context executing opdone is not the running context!\n");
		}
	} else if(caller->state == DAVIM_CONTEXT_STATE_BLOCKED) {
		DEBUG("DAVIM SCHEDULER (opdone): The calling context has been blocked, signaling to concurrency\n");
		SOS_CALL(s->concurrency_yield, davim_concurrency_yield_proto, caller);
		if(caller == s->runningContext){
			s->runningContext = NULL;
			run_task();
		} else {
			//should not happen!
			DEBUG("DAVIM SCHEDULER: Blocked context executing opdone is not the running context!\n");
		}
	} else if(caller->state == DAVIM_CONTEXT_STATE_UNBLOCKED) {
		DEBUG("DAVIM SCHEDULER: The running context has been unblocked, signaling to concurrency\n");
		caller->pc++;
		caller->num_executed++;
		// Concurrency will decide whether context is still runnable and eventually submit it back to us
		SOS_CALL(s->concurrency_resume, davim_concurrency_resume_proto, caller);
	}
}	

////! Block a context and perform scheduling, call this instead of opdone if the operation is blocking the context
//static void block(func_cb_ptr proto, DavimContext* context)
//{
//	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);
//
//	
//	context->state = DAVIM_CONTEXT_STATE_BLOCKED;
//	SOS_CALL(s->concurrency_yield, davim_concurrency_yield_proto, context);
//	s->runningContext = NULL;
//	run_task();
//}
//
////! Unblock a blocked context
//static void unblock(func_cb_ptr proto, DavimContext* context)
//{
//	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);
//
//	DEBUG("DAVIM SCHEDULER: unblock context starts running... context %i (VM %i)\n",context->which,context->vm);
//	
//	SOS_CALL(s->concurrency_resume, davim_concurrency_resume_proto, context);
//}

//! Schedule
static void run_task()
{
	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);
	
	DEBUG("DAVIM SCHEDULER: run_task starts running...\n");
	
	if (s->flags.halted == TRUE) { //TODO kan dit nog?
		DEBUG("DAVIM SCHEDULER: Halted, don't run.\n");
		s->flags.taskRunning = FALSE;
		return;
	}

	if (s->runningContext != NULL && s->runningContext->num_executed >= DAVIM_CPU_SLICE) { // Slice has expired
		DEBUG("DAVIM SCHEDULER: Slice for context %i (vm %i) expired, re-enqueue.\n", (int)s->runningContext->which, (int)s->runningContext->vm);
		s->runningContext->state = DAVIM_CONTEXT_STATE_READY;
		queue_enqueue(s->runningContext, &(s->runQueue), s->runningContext);

		s->runningContext = NULL;
	}
	
	
	if (s->runningContext == NULL /*&& !s->flags.inErrorState*/ && !queue_empty(&s->runQueue) ) 
	{
		DEBUG("DAVIM SCHEDULER: Need a new context.\n");
		s->runningContext = queue_dequeue(NULL, &s->runQueue );
		s->runningContext->num_executed = 0;
		s->runningContext->state = DAVIM_CONTEXT_STATE_RUN;
	}
	
	if (s->runningContext != NULL) {
		DEBUG("DAVIM SCHEDULER: Running context was not NULL.. So some task is running already..\n");
		
		if (s->runningContext->state != DAVIM_CONTEXT_STATE_RUN) {
			DEBUG("DAVIM SCHEDULER: No task is running even though runningContext was not NULL..\n");
			s->runningContext = NULL;
			if (post_short(s->pid, s->pid, DAVIM_CMD_SCHEDULER_RUN_TASK, 0, 0, 0) != SOS_OK) {
				s->flags.taskRunning = FALSE;
			}
		} else {
			computeInstruction(s->runningContext);
		}
	} else {
		DEBUG("DAVIM SCHEDULER: Running context was NULL\n");
		s->flags.taskRunning = FALSE;
	}
	DEBUG("DAVIM SCHEDULER: run_task call over\n");
}

//! Signal the reboot of the VM
//static void reboot(func_cb_ptr proto)
//{
//	scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);
////	MateHandlerID id;
//
//	DEBUG("DAVIM SCHEDULER: rebooting.\n");
//	s->runningContext = NULL;
//
//	queue_init( &s->runQueue );
//
////	SOS_CALL(s->synch_reset, func_vv_t);
//
////	for (id = 0; id < MATE_HANDLER_NUM; id++) {
////		SOS_CALL(s->analysis_clearAnalysis, func_vy_t, id);
////	}
////	DEBUG("MATE ENGINE: VM: Analyzing lock sets.\n");
////	for (id = 0; id < MATE_HANDLER_NUM; id++) {
////		SOS_CALL(s->analysis_analyzeVars, func_vy_t, id);
////	}
////	s->flags.inErrorState = FALSE;
//	s->flags.halted = FALSE;
//
////	DEBUG("MATE ENGINE: VM: Signaling reboot to contexts and libraries.\n");    
////	SOS_CALL(s->timer0_rebooted, func_vv_t);
////	SOS_CALL(s->reboot_rebooted, func_vv_t);
////	SOS_CALL(s->basiclib_rebooted, func_vv_t);
//
//	//post_short(M_TIMER0_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
//	//post_short(M_BASIC_LIB, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
//	//post_short(M_REBOOT_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
//	//post_short(M_CLIFF_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
//	//post_short(M_PUSHBUTTON_CONTEXT, MATE_ENGINE_M, REBOOTED, 0, 0, 0);
////	ker_led(LED_RED_OFF);
////	ker_led(LED_GREEN_OFF);
////	ker_led(LED_YELLOW_OFF);
//}

//! Execute a context
static int8_t execute(DavimContext* context) {
    scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);

    if (context->state != DAVIM_CONTEXT_STATE_READY) {
      DEBUG("DAVIM SCHEDULER: Failed to submit context %i (vm %i): not in READY state.\n", (int)context->which, (int)context->vm);
      return -EINVAL;
    }
    
    queue_enqueue(context, &s->runQueue, context);

    if (!s->flags.taskRunning) {
        s->flags.taskRunning = TRUE;
        DEBUG("DAVIM SCHEDULER: Executing context.. calling run_task.\n");
		run_task();
    }
    return SOS_OK;
}

static inline int8_t computeInstruction(DavimContext* context) {
    scheduler_state_t *s = (scheduler_state_t *) ker_get_module_state(DAVIM_SCHEDULER_PID);
    DavimLibraryId lib_id;
    uint8_t lib_mod;
    uint8_t opcode;
    

    DavimOpCode instr = context->capsule->data[context->pc];
    if (context->state != DAVIM_CONTEXT_STATE_RUN) {
      DEBUG("DAVIM SCHEDULER: Context %i (vm %i) Tried to execute instruction in non-run state: %hhi\n", context->which, context->vm, context->state);
      return -EINVAL;
    }
    
    DEBUG("DAVIM SCHEDULER: Inside compute_instruction \n");
	//! If the context just starts executing: check if libraries are present
    if (context->pc == 0) {
    		DavimLibraryBitmask libs = SOS_CALL(s->get_needed_libs,davim_vm_get_needed_libs_proto,context->vm,instr);
		if (!SOS_CALL(s->libs_present,davim_lib_manager_libs_present_proto,libs)) {
			DEBUG("DAVIM SCHEDULER: Library missing for context %i (vm %i). Halting execution.\n", context->which, context->vm);
			SOS_CALL(s->concurrency_halt, davim_concurrency_halt_proto, context);
			s->runningContext = NULL;
			run_task();
			//TODO error handling
			return -EINVAL;
		}
		context->pc++;
		instr = context->capsule->data[context->pc];
    }
	
	//! Get the library from the instruction
    lib_id = SOS_CALL(s->get_library,davim_vm_get_library_proto,context->vm,instr);
    if(lib_id == DAVIM_LIBRARY_ID_INVALID){
    		//TODO error handling
    		return -EINVAL;
    }
    lib_mod = SOS_CALL(s->get_library_pid, davim_lib_manager_get_library_pid_proto,lib_id);
    //! Get the opcode from the instruction
    opcode = SOS_CALL(s->get_opcode,davim_vm_get_opcode_proto,instr);

	//! Execute the instruction
    DEBUG("DAVIM SCHEDULER: executing instruction %i in library %i\n", opcode, lib_mod);

	DavimExecutionMsg * msg = my_malloc(sizeof(DavimExecutionMsg),s->pid);
	if(msg == NULL){
		//TODO error handling
		panic();
		return -ENOMEM;
	}
	msg->context = context;
	msg->opcode = opcode;
    post_long(lib_mod, s->pid, DAVIM_CMD_LIB_EXECUTE, sizeof(DavimExecutionMsg), msg, SOS_MSG_RELEASE);
    
    return SOS_OK;
}


#ifndef _MODULE_
mod_header_ptr scheduler_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

