/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file concurrency.c
 * @brief The concurrency manager of FAViM
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <vmstore/concurrency.h>
#include <engine/scheduler.h>
#include <util/davim_queue.h>
#include <operationstore/library.h>
#include <operationstore/library_manager.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static int8_t submit(func_cb_ptr, DavimContext*);
static int8_t resume(func_cb_ptr, DavimContext*);
static void halt(func_cb_ptr, DavimContext*);
static void yield(func_cb_ptr, DavimContext*);
static void analyse_capsule(func_cb_ptr, DavimVMId, DavimCapsule*);

// helper functions
static BOOL contextLocks_isRunnable(DavimContext* context);
static int8_t contextLocks_obtainLocks(DavimContext* caller, DavimContext* obtainer);
static int8_t contextLocks_releaseLocks(DavimContext* caller, DavimContext* releaser);
static int8_t contextLocks_releaseAllLocks(DavimContext* caller, DavimContext* releaser);
static inline int8_t locks_lock(concurrency_state_t* s, DavimContext * context, DavimVariableId id);
static inline int8_t locks_unlock(concurrency_state_t* s, DavimContext * context, DavimVariableId id);
static inline BOOL locks_is_locked(concurrency_state_t* s, DavimVMId, DavimVariableId id);

//! The module header of the concurrency manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_CONC_PID,
	.state_size     = sizeof(concurrency_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_CONC_PID),
    .num_timers     = 0,
	.num_sub_func   = 11,
	.num_prov_func  = 5,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "CCv1", DAVIM_LIB_BASICLIB_PID, DAVIM_LIB_BYTELENGTH_FID}, //Basiclib is always there, this will change later
		{NULL, "vCD1", DAVIM_LIB_BASICLIB_PID, DAVIM_LIB_USES_GLOBALS_FID}, // idem
		{NULL, "cav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_SUBMIT_FID},
		{NULL, "CCC2",DAVIM_VM_PID,DAVIM_VM_GET_LIBRARY_FID},
		{NULL, "CCv1",DAVIM_VM_PID,DAVIM_VM_GET_OPCODE_FID},
		{NULL, "YCC2",DAVIM_VM_PID, DAVIM_VM_GET_NEEDED_LIBS_FID},
		{NULL, "cCC2", DAVIM_VM_PID, DAVIM_VM_LOCK_GLOBAL_FID},
		{NULL, "cCC2", DAVIM_VM_PID, DAVIM_VM_UNLOCK_GLOBAL_FID},
		{NULL, "CCC2", DAVIM_VM_PID, DAVIM_VM_IS_LOCKED_FID},
		{NULL, "Cyv1", DAVIM_LIB_MANAGER_PID, DAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{NULL, "CCv1", DAVIM_LIB_MANAGER_PID, DAVIM_LIB_MANAGER_GET_LIBRARY_PID_FID},
		{submit, "cav1", DAVIM_CONC_PID, DAVIM_CONC_SUBMIT_FID},
		{halt, "vav1", DAVIM_CONC_PID, DAVIM_CONC_HALT_FID},
		{resume, "cav1", DAVIM_CONC_PID, DAVIM_CONC_RESUME_FID},
		{yield, "vav1", DAVIM_CONC_PID, DAVIM_CONC_YIELD_FID},
		{analyse_capsule, "vCa2", DAVIM_CONC_PID, DAVIM_CONC_ANALYSE_CAPSULE_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    concurrency_state_t *s = (concurrency_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			queue_init(&(s->readyQueue));
			return SOS_OK;
		}

        case MSG_FINAL:
        {
			return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

static int8_t submit(func_cb_ptr proto, DavimContext* context)
{

	// Set the acquire set to the set of variables used by the capsule
	int i;
	for(i=0; i < (DAVIM_GLOBAL_COUNT + 7)/8; ++i){
		context->acquireSet[i] = context->capsule->usedVars[i];
	}
	
	// Set the state of the context to waiting
	context->state = DAVIM_CONTEXT_STATE_WAITING;
	// Resume the context
	// TODO return value van resume?
	if(resume(0, context)==-EINVAL){
		return -EINVAL;
	}
	return DAVIM_OK;
}

//! Resume a context
static int8_t resume(func_cb_ptr proto, DavimContext* context)
{
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);

	DEBUG("DAVIM CONCURRENCY (%i): resuming context %i\n", context->vm, context->which);
	// Test if context is runnable
	if(contextLocks_isRunnable(context)){
		//Yes: obtain locks
		if(contextLocks_obtainLocks(context, context)!=DAVIM_OK){
			//TODO some error occured.
			return -EINVAL;
		}			
		// Pass it to the scheduler
		if(SOS_CALL(s->scheduler_submit, davim_scheduler_submit_proto, context) == SOS_OK) {
			DEBUG("DAVIM CONCURRENCY (%i): succesfully submitted context %i to scheduler\n", context->vm, context->which);
			return TRUE;
		} else {
			DEBUG("DAVIM CONCURRENCY (%i): submitting context %i to scheduler failed\n", context->vm, context->which);
			return -EINVAL;
		}
	} else {
		DEBUG("DAVIM CONCURRENCY (%i): context %i is not runnable\n", context->vm, context->which);
		//No: enqueue in the ready queue
		queue_enqueue(context,&s->readyQueue,context);
		return FALSE;
	}
	return TRUE;
}

//! Halt a context
static void halt(func_cb_ptr proto, DavimContext* context) 
{
	if (context->queue && context->state != DAVIM_CONTEXT_STATE_HALT) 
	{
		queue_remove( context, context->queue, context);
	}
	if (context->state != DAVIM_CONTEXT_STATE_HALT) 
	{
		context->state = DAVIM_CONTEXT_STATE_HALT;
	}
	contextLocks_releaseAllLocks(context, context);
	yield(0,context);
}

// TODO mss queue per VM ?
//! Yield a context
static void yield(func_cb_ptr proto, DavimContext* context) 
{
	DavimContext* start = NULL;
	DavimContext* current = NULL;
	
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	
	DEBUG("DAVIM CONCURRENCY (%i): Yielding context %i.\n", (int)context->vm, (int)context->which);
	contextLocks_releaseLocks(context, context);
	
	if( !queue_empty(&s->readyQueue) )
	{
		do 
		{
			current = queue_dequeue(context, &s->readyQueue);
			if (context->vm != current->vm || !resume(0, current)) //First check the VM, then try to resume
			{
				if (context->vm != current->vm){
					queue_enqueue(context,&s->readyQueue,context);
				} else {
					DEBUG("DAVIM CONCURRENCY (%i): Context %i not runnable.\n", (int)context->vm, (int)current->which);
				}
				if (start == NULL) 
				{
					start = current;
				}
				else if (start == current) 
				{
					DEBUG("DAVIM CONCURRENCY (%i): Looped on ready queue. End checks.\n", (int)context->vm);
					break;
				}
			}
		}
		while ( !queue_empty( &s->readyQueue ) );
	}
	else 
	{
		DEBUG("DAVIM CONCURRENCY (%i): Ready queue empty.\n", (int)context->vm);
	}
}

static void analyse_capsule(func_cb_ptr proto, DavimVMId vm, DavimCapsule* capsule) 
{
	uint16_t i;
	DavimCapsuleLength capsulelen;
	DavimOpCode instr = 0;
	DavimLibraryId lib_id = 0;
	uint8_t lib_mod;
	uint8_t uses[(DAVIM_GLOBAL_COUNT+7)/8];

	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	
	DEBUG("DAVIM CONCURRENCY (%i): Analyzing used vars for capsule\n", vm);
	// clear the usedvars
	for (i = 0; i < ((DAVIM_GLOBAL_COUNT + 7) / 8); i++) 
	{
		capsule->usedVars[i]=0;
	}

	capsulelen = capsule->dataSize;
	
	DEBUG("DAVIM CONCURRENCY: Handler length is %i.\n", capsule->dataSize);
	
	DavimLibraryBitmask libs = SOS_CALL(s->get_needed_libs,davim_vm_get_needed_libs_proto, vm, capsule->data[0]);
	if (!SOS_CALL(s->libs_present,davim_lib_manager_libs_present_proto,libs)) {			//library missing
		DEBUG("DAVIM CONCURRENCY: Library required %0llx. missing!\n",libs);
		return;
	}
	
	for (i = 1; i < capsulelen; i+=SOS_CALL(s->bytelength, davim_lib_bytelength_proto, instr) )
	{
		instr = capsule->data[i];

		//! Get the library from the instruction
	    lib_id = SOS_CALL(s->get_library,davim_vm_get_library_proto, vm, instr);
	    if(lib_id == DAVIM_LIBRARY_ID_INVALID){
	    		//TODO erro handling
	    		return;
	    }
	    lib_mod = SOS_CALL(s->get_library_pid, davim_lib_manager_get_library_pid_proto,lib_id);
	    //! Get the opcode from the instruction
	    instr = SOS_CALL(s->get_opcode,davim_vm_get_opcode_proto,instr);
		
		ker_fntable_subscribe(s->pid, lib_mod, DAVIM_LIB_BYTELENGTH_FID, 0);
		if(s->bytelength == 0){
			//TODO error handling
			DEBUG("DAVIM CONCURRENCY: pointer to bytelength == 0 for libraryid %i\n", lib_id);
			return;
		}
		DEBUG("i=%d instr=%x mod=%x\n",i, instr, lib_mod);
		ker_fntable_subscribe(s->pid, lib_mod, DAVIM_LIB_USES_GLOBALS_FID, 1);
		if(s->uses_globals != 0)		//SOMETHING TO CHECK IF THAT LIBRARY HAS A FUNCTION
		{
			SOS_CALL(s->uses_globals, davim_lib_uses_globals_proto, instr, uses);
			if (uses > 0) 
			{
				int j;
				DEBUG("DAVIM CONCURRENCY (analyse capsule): adding vars used by current instruction to usedVars\n");
				for(j=0; j < (DAVIM_GLOBAL_COUNT + 7)/8; ++j){
					capsule->usedVars[j] |= uses[j];
				}
			}
		} 
	}
}

//! Check whether a context is runnable
static BOOL contextLocks_isRunnable(DavimContext* context) 
{ 
	int8_t i;
	uint8_t* neededLocks = (context->acquireSet);
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	
	DEBUG("DAVIM CONCURRENCY (%i): Checking whether context %i runnable: ", (int)context->vm, (int)context->which);

	for (i = 0; i < DAVIM_GLOBAL_COUNT; i++) 
	{
		DEBUG_SHORT("%i,", (int)i); 
		if ((neededLocks[i / 8]) & (1 << (i % 8))) 
		{
			if (locks_is_locked(s,context->vm,i))
			{
				DEBUG_SHORT(" - no\n");
				return FALSE;
			}
		}
	}
	DEBUG_SHORT(" - yes\n");
	return TRUE;
}

//! Obtain locks for the given context
static int8_t contextLocks_obtainLocks(DavimContext* caller, DavimContext* obtainer) 
{ 
	int i;
	uint8_t* neededLocks = (obtainer->acquireSet);
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	//ker_led(LED_RED_TOGGLE);	
	DEBUG("DAVIM CONCURRENCY (%i): Attempting to obtain necessary locks for context %i: ", obtainer->vm, obtainer->which);
	for (i = 0; i < DAVIM_GLOBAL_COUNT; i++) 
	{
		DEBUG_SHORT("%i", (int)i);
		if ((neededLocks[i / 8]) & (1 << (i % 8))) 
		{
			DEBUG_SHORT("+"); 
			if(locks_lock(s, obtainer, i)!=DAVIM_OK){
				//TODO error handling
				return DAVIM_ERROR_FAILED_LOCKING;
			}
		}
		DEBUG_SHORT(","); 
	}
	for (i = 0; i < (DAVIM_GLOBAL_COUNT + 7) / 8; i++) 
	{
		obtainer->acquireSet[i] = 0;
	}
	DEBUG_SHORT("\n");
	return DAVIM_OK;		
}

//! Release the locks in the releaseSet of the given context
static int8_t contextLocks_releaseLocks(DavimContext* caller, DavimContext* releaser) 
{
	int i;
	uint8_t* lockSet = (releaser->releaseSet);
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	
	DEBUG("DAVIM CONCURRENCY (%i): Attempting to release specified locks for context %i.\n", releaser->vm, releaser->which);
	for (i = 0; i < DAVIM_GLOBAL_COUNT; i++) 
	{
		if ((lockSet[i / 8]) & (1 << (i % 8))) 
		{
			locks_unlock(s, releaser, i);
		}
	}
	for (i = 0; i < (DAVIM_GLOBAL_COUNT + 7) / 8; i++) 
	{
		releaser->releaseSet[i] = 0;
	}
	return SOS_OK;		
}

//! Release all locks the given context holds.
static int8_t contextLocks_releaseAllLocks(DavimContext* caller, DavimContext* releaser) 
{
	int i;
	uint8_t* lockSet = (releaser->heldSet);
	concurrency_state_t * s = (concurrency_state_t*) ker_get_module_state(DAVIM_CONC_PID);
	
	DEBUG("DAVIM CONCURRENCY (%i): Attempting to release all locks for context %i.\n", releaser->vm, releaser->which);
	for (i = 0; i < DAVIM_GLOBAL_COUNT; i++) 
	{
		if ((lockSet[i / 8]) & (1 << (i % 8))) 
		{
			locks_unlock(s, releaser, i);
		}
	}
	for (i = 0; i < (DAVIM_GLOBAL_COUNT + 7) / 8; i++) 
	{
		releaser->releaseSet[i] = 0;
	}
	return SOS_OK;
}

//! Lock the given variable in the VM
static inline int8_t locks_lock(concurrency_state_t* s, DavimContext * context, DavimVariableId id)
{
	return SOS_CALL(s->lock, davim_vm_lock_global_proto, context, id);
}

//! Unlock the given variable in the VM
static inline int8_t locks_unlock(concurrency_state_t* s, DavimContext * context, DavimVariableId id)
{
	return SOS_CALL(s->unlock, davim_vm_unlock_global_proto, context, id);
}

static inline BOOL locks_is_locked(concurrency_state_t* s, DavimVMId vm, DavimVariableId id)
{
	return SOS_CALL(s->is_locked, davim_vm_is_locked_proto, vm, id);
}

#ifndef _MODULE_
mod_header_ptr concurrency_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

