/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file stack.c
 * @brief The implementation of the FAViM stack interface
 * @author Wouter Horr&eacute;
 */
#include <module.h>

#include <vmstore/stack.h>

//forward declarations
static int8_t module(void *state, Message *e);
static DavimOperandStack* new(func_cb_ptr);
static void finalize(func_cb_ptr, DavimOperandStack*);
static int8_t push(func_cb_ptr proto, DavimOperandStack* stack, void* var);
static void* pop(func_cb_ptr proto, DavimOperandStack* stack);
static void reset(func_cb_ptr proto, DavimOperandStack* stack);

static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_STACK_PID,
	.state_size     = sizeof(stack_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_STACK_PID),
	.num_timers     = 0,
	.num_sub_func   = 0,
	.num_prov_func  = 5,
	.module_handler = module,
	.funct = {
		{new, "avv0", DAVIM_STACK_PID, DAVIM_STACK_NEW_FID},
		{finalize, "vav1", DAVIM_STACK_PID, DAVIM_STACK_FINALIZE_FID},
		{reset, "vav1", DAVIM_STACK_PID, DAVIM_STACK_RESET_FID},
		{push, "caa2", DAVIM_STACK_PID, DAVIM_STACK_PUSH_FID},
		{pop, "aav1", DAVIM_STACK_PID, DAVIM_STACK_POP_FID},
	},
};

/// Main event handling function
static int8_t module(void *state, Message *msg)
{
	stack_state_t *s = (stack_state_t *) state;

	switch (msg->type){
		case MSG_INIT:
			{
				s->pid = msg->did;
				DEBUG("FAViM: Stack initialized\n");
				return SOS_OK;
			}
		case MSG_FINAL:
			{
				return SOS_OK;
			}
		default:
			return -EINVAL;
	}
	return SOS_OK;
}

//! Create a new operand stack
static DavimOperandStack* new(func_cb_ptr proto)
{
	DavimOperandStack * s = (DavimOperandStack*) my_malloc(sizeof(DavimOperandStack), DAVIM_STACK_PID);
	if(s == 0){
		//TODO error handling
		DEBUG("STACK NEW: Something very very wrong here\n");
		panic();
		return 0;
	}
	reset(0,s);
	return s;
}

//! Finalize a stack
static void finalize(func_cb_ptr proto, DavimOperandStack* stack)
{
	int i = 0;
	for(i=0; i< DAVIM_STACK_DEPTH; ++i){
		if(stack->stack[i] != NULL){
			ker_free(stack->stack[i]);
		}
	}
	ker_free(stack);
}

//! Push a variable onto a FAViM stack
static int8_t push(func_cb_ptr proto, DavimOperandStack* stack, void* var)
{
	if(stack->sp >= DAVIM_STACK_DEPTH){
		/// TODO raise an error
		return -EINVAL;
	} else {
		stack->stack[(int)stack->sp] = var;
		stack->sp++;
		return DAVIM_OK;
	}
}

/// Pop a variable of a FAViM stack
static void* pop(func_cb_ptr proto, DavimOperandStack* stack)
{
	if(stack->sp == 0) {
		///TODO raise an error / implement properly
		return NULL;
	} else {
		stack->sp--;
		void* temp = stack->stack[(int)stack->sp];
		stack->stack[(int)stack->sp] = NULL;
		return temp;
	}
}

/// Reset a FAViM stack
static void reset(func_cb_ptr proto, DavimOperandStack* stack)
{
	stack->sp=0;
	// TODO clear stack? stack uses value semantics!
}

#ifndef _MODULE_
mod_header_ptr stack_get_header()
{
	return sos_get_header_address(mod_header);
}
#endif

