/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file vm.c
 * @brief A FAViM virtual machine
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <vmstore/vm.h>
#include <operationstore/library_manager.h>
#include <engine/event_manager.h>
#include <vmstore/concurrency.h>
#include <capsulestore/capsule_manager.h>

#define HASH_MASK 0x07

// Forward declarations
static int8_t module(void *state, Message *e);
static DavimVM * new(func_cb_ptr,DavimVMId);
static void finalize(func_cb_ptr, DavimVM*);
static DavimLibraryId get_library(func_cb_ptr, DavimVMId, DavimOpCode);
static DavimOpCode get_opcode(func_cb_ptr, DavimOpCode); //Same for all VMs
static DavimLibraryBitmask get_needed_libs(func_cb_ptr, DavimVMId, DavimOpCode);
static int8_t lock_global(func_cb_ptr, DavimContext*, DavimVariableId); //pointer to vm is also present in context
static int8_t unlock_global(func_cb_ptr, DavimContext*, DavimVariableId);
static BOOL is_locked(func_cb_ptr, DavimVMId, DavimVariableId);
static void halt(func_cb_ptr, DavimVMId);
static void get_description(func_cb_ptr, DavimVMId, DavimVMDescription*);

static DavimContext* get_context_from_id(DavimVM*, DavimEventId);
static void install_description(DavimVM*, DavimVMDescription*);
static void add_context(DavimVM*, DavimContextId);
static void remove_context(DavimVM*, DavimContextId);
static inline void insert_he(DavimVM*, DavimContextHashEntry *);
static inline void finalize_context(DavimContext *);

//! The module header of the VM module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_VM_PID,
	.state_size     = sizeof(vm_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_VM_PID),
    .num_timers     = 0,
	.num_sub_func   = 7,
	.num_prov_func  = 10,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "cav1", DAVIM_CONC_PID, DAVIM_CONC_SUBMIT_FID},
		{NULL, "vav1", DAVIM_CONC_PID, DAVIM_CONC_HALT_FID},
		{NULL, "aCC2", DAVIM_CONTEXT_PID, DAVIM_CONTEXT_NEW_FID},
		{NULL, "vav1", DAVIM_CONTEXT_PID, DAVIM_CONTEXT_FINALIZE_FID},
		{NULL, "vav1", DAVIM_CONTEXT_PID, DAVIM_CONTEXT_RESET_FID},
		{NULL, "cCC3",DAVIM_EVENT_MANAGER_PID,DAVIM_EVENT_MANAGER_TRIGGER_EVENT_FID},
		{NULL, "aCv1", DAVIM_VM_MANAGER_PID, DAVIM_VM_MAN_GET_VM_FID},
		{new, "aCv1", DAVIM_VM_PID, DAVIM_VM_NEW_FID},
		{finalize, "vav1", DAVIM_VM_PID, DAVIM_VM_FINALIZE_FID},
		{get_library, "CCC2",DAVIM_VM_PID,DAVIM_VM_GET_LIBRARY_FID},
		{get_opcode, "CCv1",DAVIM_VM_PID,DAVIM_VM_GET_OPCODE_FID},
		{get_needed_libs, "YCC2",DAVIM_VM_PID, DAVIM_VM_GET_NEEDED_LIBS_FID},
		{lock_global, "cCC2", DAVIM_VM_PID, DAVIM_VM_LOCK_GLOBAL_FID},
		{unlock_global, "cCC2", DAVIM_VM_PID, DAVIM_VM_UNLOCK_GLOBAL_FID},
		{is_locked, "CCC2", DAVIM_VM_PID, DAVIM_VM_IS_LOCKED_FID},
		{halt, "vCv1", DAVIM_VM_PID, DAVIM_VM_HALT_FID},
		{get_description, "vCa1", DAVIM_VM_PID, DAVIM_VM_GET_DESCRIPTION_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    vm_state_t *s = (vm_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			DEBUG("DAVIM: VM INITIALIZATION\n");
			s->pid = msg->did;
//			s->nb_contexts=0;
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
//        case DAVIM_CMD_VM_ADD_CONTEXT:
//        {
//	        	//Unmarshall the parameter
//	        	DavimVMId vmid = (DavimVMId)((MsgParam*)msg->data)->word);
//	        	DavimContextId id = (DavimContextId)((MsgParam*)msg->data)->byte);
//	        DavimVM* vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto,vmid); 
//
//	        DEBUG("DAVIM VM (%i): ADD CONTEXT %i STARTED\n", vm->id, id);
//	        	//Stop the VM
//	        	halt(0,vm);
//	        //Make the new context
//	        DavimContext* con = SOS_CALL(s->context_new, davim_context_new_proto, vm, id);
//	        //Add it to the data structures
//	        add_context(vm,con);
//	        //Reboot the VM
//	        	post_short(s->pid,s->pid,DAVIM_CMD_VM_REBOOT,vmid,0,0);
//        	
//	        	return SOS_OK;
//        }
//        
//        case DAVIM_CMD_VM_REMOVE_CONTEXT:
//        {
//	        	//Unmarshall the parameter
//	        	DavimVMId vmid = (DavimVMId)((MsgParam*)msg->data)->word);
//	        	DavimContextId id = (DavimContextId)((MsgParam*)msg->data)->byte);
//	        DavimVM* vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto,vmid); 
//        		
//	        DEBUG("DAVIM VM (%i): REMOVE CONTEXT %i STARTED\n", vm->id, id);
//	        	//Stop the VM
//	        	halt(0, vm);
//	        //Remove the context
//	        if(id > 0){ // We don't want to remove the reboot context!
//	        		remove_context(vm, id);
//	        }
//	        //Reboot the VM
//	        	post_short(s->pid,s->pid,DAVIM_CMD_VM_REBOOT,vmid,0,0);
//	       
//	        	return SOS_OK;
//        }
        
        case DAVIM_CMD_VM_INSTALL_DESCRIPTION:
        {
        		DavimVMDescription* desc = (DavimVMDescription*)msg->data;
	        DavimVM* vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto,desc->id);
	        
	        DEBUG("DAVIM VM (%i): INSTALL DESCRIPTION STARTED\n", vm->id);
	        	//Stop the VM
	        	halt(0, vm->id);
        		// install the description
        		install_description(vm,desc);
        		//Reboot the VM
	        	post_short(s->pid,s->pid,DAVIM_CMD_VM_REBOOT,vm->id,0,0);
        		
        		return SOS_OK;
        }
        
        case DAVIM_CMD_VM_EVENT_OCCURED:
        {
	        	// Lookup the context associated with the event
	        	DavimVMId vmid = (DavimVMId)((MsgParam*)msg->data)->word;
	        	DavimEventId id = (DavimEventId)((MsgParam*)msg->data)->byte;
	        DavimVM* vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto,vmid); 
			if(!vm->halted){
		        	DavimContext* context = get_context_from_id(vm,id);
		        	// Submit the context for execution
		        	SOS_CALL(s->concurrency_submit, davim_concurrency_submit_proto, context);
			}
	        	return SOS_OK;
        }
        
        case DAVIM_CMD_VM_REBOOT:
        {
        		int i;
        		DavimVMId id = (DavimVMId)((MsgParam*)msg->data)->byte;
        		DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto,id); 
        		
        		DEBUG("DAVIM VM: Rebooting VM %i\n", id);
	        	// Reset the VM (inc. globals,locks)
			for(i=0; i < DAVIM_GLOBAL_COUNT; ++i){
				if(vm->globals[i] != NULL){
					ker_free(vm->globals[i]);
					vm->globals[i] = NULL;
				}
				vm->locks[i].holder = NULL;
			}
			// Reset contexts
			for(i=0; i < 8; ++i){
				DavimContextHashEntry * current;
				for(current = vm->contexts[i]; current != NULL; current = current->next){
					SOS_CALL(s->context_reset, davim_context_reset_proto, current->context);
				}
			}
			vm->halted = FALSE;
	        	// Submit the reboot context for execution
	        SOS_CALL(s->concurrency_submit,davim_concurrency_submit_proto, get_context_from_id(vm, REBOOT_CONTEXT_ID));
	        	return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Create a new VM
static DavimVM * new(func_cb_ptr proto, DavimVMId id)
{
	int i;

	DEBUG("DAVIM VM: creating a new VM\n");
	// Allocate memory
	DavimVM * nvm = my_malloc(sizeof(DavimVM),DAVIM_VM_PID);
	if(nvm == NULL){
		//TODO error handling
		panic();
		return NULL;
	}	

	// Initialize the members
	nvm->id = id;	
	for(i=0; i < DAVIM_GLOBAL_COUNT; ++i){
		nvm->globals[i] = NULL;
		nvm->locks[i].holder = NULL;
	}
	for(i=0; i < 8; ++i){
		nvm->contexts[i] = NULL;
		// Initialize the library mapping
		nvm->lib_mapping[i] = DAVIM_LIBRARY_ID_INVALID;
	}
	// In a default VM, we want basiclib ;)
	nvm->lib_mapping[0] = 0;
		
	// Add a reboot context
	add_context(nvm, REBOOT_CONTEXT_ID);
	
	return nvm;
}

//! Finalize a VM
static void finalize(func_cb_ptr proto, DavimVM * vm)
{
	int i;
	for(i=0; i < 8; ++i){
		DavimContextHashEntry * current = vm->contexts[i];
		while(current != NULL){
			DavimContextHashEntry * prev = current;
			finalize_context(current->context);
			current = current->next;
			ker_free(prev);
		}
	}
	for(i=0; i<8; ++i){
		if(vm->globals[i] != NULL){
			ker_free(vm->globals[i]);
		}
	}
	ker_free(vm);
}

//! Get the library id from an instruction
static DavimLibraryId get_library(func_cb_ptr proto, DavimVMId id, DavimOpCode instr)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, id);

    uint8_t lib = ((instr >> OPCODE_BITS) & LIBRARY_MASK);	
   	return vm->lib_mapping[lib];
}

//! Get the opcode from an instruction
static DavimOpCode get_opcode(func_cb_ptr proto, DavimOpCode instr)
{
    return (instr & OPCODE_MASK);
}

//! Get the needed libs (global ids) from the encoding at the start of a capsule
static DavimLibraryBitmask get_needed_libs(func_cb_ptr proto, DavimVMId id, DavimOpCode libs)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, id);

	int i;
	DavimLibraryBitmask ret = 0;
	for(i=0; i<8; ++i){
		uint8_t bit = (1 << i);
		if(libs & bit){
			//This lib is present
			if(vm->lib_mapping[i] == DAVIM_LIBRARY_ID_INVALID){
				//TODO something wrong, error handling
				// Is this an error? maybe we want to know all needed libs and call this with 0xff ?
			} else {
				ret |= (1 << vm->lib_mapping[i]);
			}
		}
	}
	return ret;
}

//! Lock the given global variable
static int8_t lock_global(func_cb_ptr proto, DavimContext* context, DavimVariableId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, context->vm);

	DEBUG("DAVIM VM (%i): locking %i, context %i\n", vm->id, id, context->which);
	if(id >= DAVIM_GLOBAL_COUNT){
		return -DAVIM_ERROR_ID_INVALID;
	}
	vm->locks[id].holder = context;
	context->heldSet[id / 8] |= (1 << (id % 8));
	return DAVIM_OK;
}

//! Unlock the given global variable
static int8_t unlock_global(func_cb_ptr proto, DavimContext* context, DavimVariableId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, context->vm);

	DEBUG("DAVIM VM (%i): unlocking %i, context %i\n", vm->id, id, context->which);
	if(id >= DAVIM_GLOBAL_COUNT){
		return -DAVIM_ERROR_ID_INVALID;
	}
	vm->locks[id].holder = NULL;
	context->heldSet[id / 8] &= ~(1 << (id % 8));
	return DAVIM_OK;
}

//! Check whether a global is locked
static BOOL is_locked(func_cb_ptr proto, DavimVMId vmid, DavimVariableId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, vmid);

	DEBUG("DAVIM VM (%i): checking lock %i\n", vm->id, id);
	
	if(id >= DAVIM_GLOBAL_COUNT){
		//TODO error?
		return FALSE;
	}
	return (vm->locks[id].holder!=NULL);
}

//! Halt the VM
static void halt(func_cb_ptr proto, DavimVMId id)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, id);
	
	DEBUG("DAVIM VM (%i): halting the VM\n", id);
	
	vm->halted = TRUE;
	SOS_CALL(s->event_man_trigger_event, davim_event_manager_trigger_event_proto, DAVIM_EVENT_GLOBAL,0,DAVIM_EVENT_GLOBAL_VM_HALTED+id);
	
	int i;
	for(i=0; i < 8; ++i){
		DavimContextHashEntry * current;
		for(current = vm->contexts[i]; current != NULL; current = current->next){
			SOS_CALL(s->concurrency_halt, davim_concurrency_halt_proto, current->context);		
		}
	}
}

//! Get the description of a VM.
//! The memory will be owned by the VM module
//! Changing ownership will be necessary
static void get_description(func_cb_ptr proto, DavimVMId id, DavimVMDescription* desc)
{
	int i;
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	DavimVM * vm = SOS_CALL(s->get_vm, davim_vm_man_get_vm_proto, id);
	
	if(vm == NULL){
		desc->id = DAVIM_VM_INVALID;
		return;
	}
	
	desc->id = vm->id;
	for(i=0; i<8; ++i){
		desc->lib_mapping[i] = vm->lib_mapping[i];
	}
	desc->contexts = 0;
	for(i=0; i < 8; ++i){
		DavimContextHashEntry * current;
		for(current = vm->contexts[i]; current != NULL; current = current->next){
			desc->contexts |= (0x1 << current->context->which);
		}
	}
}


//! Get the context associated with the given event id
// For now contextid==eventid
static DavimContext* get_context_from_id(DavimVM * vm, DavimEventId id)
{
	uint8_t hid = id & HASH_MASK;
	DavimContextHashEntry * current;
	for(current = vm->contexts[hid]; current != NULL; current = current->next){
		if(current->context->which == id){
			return current->context;
		}
	}
	//Nothing found with the given id
	return NULL;
}

//! Update a vm to a given description
static void install_description(DavimVM* vm, DavimVMDescription* desc)
{
	int i;
	//Update the library mapping
	for(i = 0; i<8; ++i){
		vm->lib_mapping[i] = desc->lib_mapping[i];
	}
	//Add/Remove contexts
	//Construct bitmask of present contexts
	DavimContextBitmask mask=0;
	for(i=0; i < 8; ++i){
		DavimContextHashEntry * current;
		for(current = vm->contexts[i]; current != NULL; current = current->next){
			mask |= (0x1 << current->context->which);
		}
	}
	//Check whether it differs from the bitmask in desc
	if(mask ^ desc->contexts){
		// Yes - Contexts have to change
		// Check for contexts present in mask, but not in desc->contexts
		DavimContextBitmask diff = mask & (~desc->contexts);
		if(diff){
			//We must delete contexts
			for(i=1; i<DAVIM_CONTEXT_COUNT; ++i){ // Start from 1 to prevent deletion of reboot context!
				if(diff & (0x1 << i)){
					remove_context(vm,i);
				}
			}
		}
		// Check for contexts present in desc->contexts but not in mask
		diff = (~mask) & desc->contexts;
		if(diff){
			//We must add contexts
			for(i=0; i<DAVIM_CONTEXT_COUNT; ++i){
				if(diff & (0x1 << i)){
					add_context(vm,i);
				}
			}
		} 
	}
}


//! Add a context to the VM datastructures
static void add_context(DavimVM * vm, DavimContextId id){
	DEBUG("DAVIM VM (%i): adding context %i\n", vm->id, id);
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);

    //Make the new context
    DavimContext* con = SOS_CALL(s->context_new, davim_context_new_proto, vm->id, id);
	
	DavimContextHashEntry * he = (DavimContextHashEntry *) my_malloc(sizeof(DavimContextHashEntry),s->pid);
	if(he == NULL){
		//TODO error handling
		DEBUG("VM: ADD CONTEXT: something very very wrong here!\n");
		panic();
		return;
	}
	he->context = con;
	he->next = NULL;
	insert_he(vm,he);
}

//! Remove a context from the VM datastructures
static void remove_context(DavimVM * vm, DavimContextId id){
	DEBUG("DAVIM VM (%i): removing context %i\n", vm->id, id);
//	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);
	
	uint8_t hid = id & HASH_MASK;
	
	if(vm->contexts[hid] == NULL){
		DEBUG("DAVIM VM (%i): tried to remove a non-existent context %i\n", vm->id, id);
		//TODO error handling
	} else if(vm->contexts[hid]->context->which == id) {
		DavimContextHashEntry * prev = vm->contexts[hid];
		vm->contexts[hid] = vm->contexts[hid]->next;
		finalize_context(prev->context);
		ker_free(prev);
	} else {
		DavimContextHashEntry * current = vm->contexts[hid];
		while(current->next != NULL && current->next->context->which != id){
			current = current->next;
		}
		if(current->next == NULL){
			DEBUG("DAVIM VM: tried to remove a non-existent context %i\n", id);
			//TODO error handling
		} else {
			DavimContextHashEntry * prev = current->next;
			current->next = current->next->next;
			finalize_context(prev->context);
			ker_free(prev);
		}
	}	
}

//! Release the memory occupied by a context.
static inline void finalize_context(DavimContext * context)
{
	vm_state_t * s = (vm_state_t*) ker_get_module_state(DAVIM_VM_PID);

	SOS_CALL(s->context_finalize, davim_context_finalize_proto, context);
}


//! insert a entry in the hash table of a VM
static inline void insert_he(DavimVM * vm, DavimContextHashEntry * he)
{
	uint8_t hid = (he->context->which & HASH_MASK);
	if(vm->contexts[hid] == NULL){
		vm->contexts[hid] = he;
	} else {
		DavimContextHashEntry * current = vm->contexts[hid];
		while(current->next != NULL && current->context->which!=he->context->which){
			current = current->next;
		}
		if(current->next==NULL){
			current->next = he;
		} else {
			//Context is already there???
			//TODO error handling
			finalize_context(he->context);
			ker_free(he);
		}
	}
}

#ifndef _MODULE_
mod_header_ptr vm_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

