/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file vm_manager.c
 * @brief The FAViM vm manager
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <engine/event_manager.h>
#include <operationstore/library_manager.h>
#include <virus/virus.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static DavimVM* get_vm(func_cb_ptr,DavimVMId);
static BOOL vm_present(func_cb_ptr,DavimVMId);
static DavimVMVersion get_vm_version(func_cb_ptr,DavimVMId);


// Forward declarations: helper functions
static void add_vm(DavimVM *);
static void remove_vm_by_id(DavimVMId);

//! The module header of the davim vm manager
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_VM_MANAGER_PID,
	.state_size     = sizeof(vm_man_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_VM_MANAGER_PID),
    .num_timers     = 0,
	.num_sub_func   = 4,
	.num_prov_func  = 3,
	.module_handler = module,
	.funct = {
		//TODO error handling
		{NULL, "aCv1", DAVIM_VM_PID, DAVIM_VM_NEW_FID},
		{NULL, "vav1", DAVIM_VM_PID, DAVIM_VM_FINALIZE_FID},
		{NULL, "YCC2",DAVIM_VM_PID, DAVIM_VM_GET_NEEDED_LIBS_FID},
		{NULL, "Cyv1", DAVIM_LIB_MANAGER_PID, DAVIM_LIB_MANAGER_LIBS_PRESENT_FID},
		{get_vm, "aCv1", DAVIM_VM_MANAGER_PID, DAVIM_VM_MAN_GET_VM_FID},
		{vm_present, "CCv1", DAVIM_VM_MANAGER_PID, DAVIM_VM_MAN_VM_PRESENT_FID},
		{get_vm_version, "LCv1", DAVIM_VM_MANAGER_PID, DAVIM_VM_MAN_GET_VM_VERSION_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    vm_man_state_t *s = (vm_man_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			int i;
			
			s->pid = msg->did;
			for(i=0; i<8; ++i){
				s->vms[i] = NULL;
			}
			
			// ADD one VM to the system by default
	        //Make the VM
	        DavimVM * nvm = SOS_CALL(s->vm_new, davim_vm_new_proto, 0);
	        //Add it to the data structures
	        add_vm(nvm);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
			return SOS_OK;
        }
        
//        case DAVIM_CMD_VM_MAN_ADD_VM:
//        {
//        		DEBUG("DAVIM VM MANAGER: add VM started\n");
//        		
//        		MsgParam *rcvd = (MsgParam *)msg->data;		
//	        DavimVMId id = (DavimVMId)rcvd->byte;
//	        
//	        //Make the VM
//	        DavimVM * nvm = SOS_CALL(s->vm_new, davim_vm_new_proto, id);
//	        
//	        //Add it to the data structures
//	        add_vm(nvm);
//	        
//	        //Reboot it
//	        	post_short(DAVIM_VM_PID,s->pid,DAVIM_CMD_VM_REBOOT,nvm->id,0,0);
//	                		
//        		return SOS_OK;
//        }
//
//        case DAVIM_CMD_VM_MAN_REMOVE_VM:
//        {
//        		DEBUG("DAVIM VM MANAGER: remove VM started\n");
//        		
//        		MsgParam *rcvd = (MsgParam *)msg->data;		
//	        DavimVMId id = (DavimVMId)rcvd->byte;
//	        
//	        remove_vm_by_id(id);
//	        
//        		return SOS_OK;
//        }
        case DAVIM_CMD_VM_MAN_VMMSG_ARRIVED:
        {
        		DEBUG("DAVIM VM MANAGER: VM message arrived\n");
        		DavimVMMsg * vmmsg = (DavimVMMsg*)ker_msg_take_data(s->pid,msg);
        		
        		if(vmmsg->desc.id >= DAVIM_VM_COUNT){
        			//TODO error handling
        			ker_free(vmmsg);
        			return -EINVAL;
        		}
        		
        		if(vmmsg->version > s->versions[vmmsg->desc.id]){
        			s->versions[vmmsg->desc.id] = vmmsg->version;
        			if(vmmsg->present){
        				// It's an add of update
        				if(!vm_present(0,vmmsg->desc.id)){
				        //Make the VM
				        DavimVM * nvm = SOS_CALL(s->vm_new, davim_vm_new_proto, vmmsg->desc.id);
				        
				        //Add it to the data structures
				        add_vm(nvm);
        				}
        				// Install the description (it will be rebooted also)
        				DavimVMDescription * desc = my_malloc(sizeof(DavimVMDescription),s->pid);
        				if(desc == NULL){
        					//TODO error handling
        					panic();
        					return -ENOMEM;
        				}
        				// Deep copy
        				*desc = vmmsg->desc;
        				post_long(DAVIM_VM_PID,s->pid,DAVIM_CMD_VM_INSTALL_DESCRIPTION,sizeof(DavimVMDescription),desc,SOS_MSG_RELEASE);
        			} else {
        				// It's a remove
        				if(vm_present(0,vmmsg->desc.id)){
        					remove_vm_by_id(vmmsg->desc.id);
        				}
        			}
        		} else {
        			DEBUG("DAVIM VM MANAGER: we are already up to date");
        		}
        		ker_free(vmmsg);
        		return SOS_OK;
        }
        
        case DAVIM_CMD_TRIGGER_GLOBAL:
        {
        		DavimEventId id = (DavimEventId)((MsgParam *)msg->data)->byte;
        		if(id == DAVIM_EVENT_GLOBAL_ADD_LIB){
        			//TODO implement when error handling is implemented -> vm in missing lib state
        		} else if (id == DAVIM_EVENT_GLOBAL_REMOVE_LIB) {
        			//check whether vms have missing libs now -> reboot them
        			int i;
        			for(i=0; i < DAVIM_VM_COUNT; ++i){
    					DavimLibraryBitmask mask = SOS_CALL(s->vm_get_needed_libs, davim_vm_get_needed_libs_proto, s->vms[i]->id, 0xff);
    					if(!SOS_CALL(s->lib_man_libs_present, davim_lib_manager_libs_present_proto,mask)){
				        	post_short(DAVIM_VM_PID,s->pid,DAVIM_CMD_VM_REBOOT,s->vms[i]->id,0,0);
    					}
        			}
        		}
        		return SOS_OK;
        }

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get a vm by its id
static DavimVM* get_vm(func_cb_ptr proto,DavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(DAVIM_VM_MANAGER_PID);

	if(id >= DAVIM_VM_COUNT){
		//TODO error handling
		return NULL;
	}
	return s->vms[id];
}

//! Check whether a VM is present
static BOOL vm_present(func_cb_ptr proto, DavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(DAVIM_VM_MANAGER_PID);
	return (id<DAVIM_VM_COUNT && s->vms[id]!=NULL);
}

//! Get the version of a VM
static DavimVMVersion get_vm_version(func_cb_ptr proto,DavimVMId id)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(DAVIM_VM_MANAGER_PID);
	if(id >= DAVIM_VM_COUNT){
		//TODO error handling
		return 0;
	} else {
		return s->versions[id];
	}
}

//! Add a VM to the data structures
static void add_vm(DavimVM * vm)
{
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(DAVIM_VM_MANAGER_PID);
	DEBUG("DAVIM VM MANAGER: adding vm %i\n", vm->id);
	
	if(vm->id >= DAVIM_VM_COUNT){
		//TODO error handling
	} else {
		s->vms[vm->id] = vm;	
	}
}

//! Remove a VM by its id
static void remove_vm_by_id(DavimVMId id)
{
	DEBUG("DAVIM VM MANAGER: removing vm %i\n", id);
	
	vm_man_state_t * s = (vm_man_state_t*) ker_get_module_state(DAVIM_VM_MANAGER_PID);
	
	if(id >= DAVIM_VM_COUNT){
		//TODO error handling
	} else {
		SOS_CALL(s->vm_finalize, davim_vm_finalize_proto, s->vms[id]);
		s->vms[id]=NULL;
	}	
}

#ifndef _MODULE_
mod_header_ptr vm_manager_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

