/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file basiclib.c
 * @brief The basic operation library
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <led.h>
#include <davim.h>
#include <operationstore/library.h>
#include <engine/scheduler.h>
#include <vmstore/vm.h>
#include <vmstore/context.h>
#include <engine/event_manager.h>
#include <util/davim_queue.h>
#include <sensor.h>

#define OP_HALT 0x00
#define OP_PUSHC6 0x01
#define OP_LED 0x02
//#define OP_POST_NET 0x03
#define OP_SETTIMER0 0x0c
#define OP_GET_PHOTO_DATA 0x0F

//enum {
//        MSG_DATA_FROM_REMOTE  = (MOD_MSG_START + 0),
//};

//! The state of the basic library
typedef struct {
	func_cb_ptr scheduler_opdone;
//	func_cb_ptr scheduler_block;
//	func_cb_ptr scheduler_unblock;
	func_cb_ptr stack_push;
	func_cb_ptr stack_pop;
	func_cb_ptr events_start_timer;
	func_cb_ptr events_stop_timer;
	sos_pid_t pid;
	uint8_t busy;                   //for GET_PHOTO_DATA
    DavimContext *executing;         //for GET_PHOTO_DATA
    DavimQueue getDataWaitQueue;     //for GET_PHOTO_DATA
//	int16_t sendbuf;					//for POST_NET
} basiclib_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static uint8_t bytelength(func_cb_ptr, DavimOpCode opcode);
static void uses_globals(func_cb_ptr, DavimOpCode opcode, uint8_t*);
static int8_t tryget(DavimContext *, uint8_t);
static void halted(DavimVMId);

//! The module header of the basic library
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_LIB_BASICLIB_PID,
	.state_size     = sizeof(basiclib_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_LIB_BASICLIB_PID),
    .num_timers     = 0,
	.num_sub_func   = 5,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_OPDONE_FID},
//		{NULL, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_BLOCK_FID},
//		{NULL, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_UNBLOCK_FID},
		{NULL, "caa2", DAVIM_STACK_PID, DAVIM_STACK_PUSH_FID},
		{NULL, "aav1", DAVIM_STACK_PID, DAVIM_STACK_POP_FID},
		{NULL, "cCC5", DAVIM_EVENT_MANAGER_PID, DAVIM_EVENT_MANAGER_START_TIMER_FID},
		{NULL, "cCC3", DAVIM_EVENT_MANAGER_PID, DAVIM_EVENT_MANAGER_STOP_TIMER_FID},
		{bytelength, "CCv1", DAVIM_LIB_BASICLIB_PID, DAVIM_LIB_BYTELENGTH_FID},
		{uses_globals, "vCD1", DAVIM_LIB_BASICLIB_PID, DAVIM_LIB_USES_GLOBALS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    basiclib_state_t *s = (basiclib_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			s->busy = FALSE;
			s->executing = NULL;
			queue_init(&(s->getDataWaitQueue));
//			s->sendbuf=0;
			post_short(DAVIM_LIB_MANAGER_PID,s->pid,DAVIM_CMD_LIB_MANAGER_ADD_LIBRARY,0,0,0);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
        		post_short(DAVIM_LIB_MANAGER_PID,s->pid,DAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY,0,0,0);
			return SOS_OK;
        }
        case DAVIM_CMD_TRIGGER_GLOBAL:
        {
        		// TODO shouldn't this happen synchronous?
        		DavimEventId id = (DavimEventId)((MsgParam *)msg->data)->byte;
        		if(id >= DAVIM_EVENT_GLOBAL_VM_HALTED && id < DAVIM_EVENT_GLOBAL_VM_HALTED + DAVIM_VM_COUNT){
        			// A VM rebooted
        			halted(id-DAVIM_EVENT_GLOBAL_VM_HALTED);
        		}
        		return SOS_OK;
        }

        case DAVIM_CMD_LIB_EXECUTE:
        {
        		int8_t ret = SOS_OK;
        		DavimExecutionMsg* exec = (DavimExecutionMsg*) msg->data;
			DavimContext* context = exec->context;
			DavimOpCode instr = exec->opcode;
						
			if(context->state == DAVIM_CONTEXT_STATE_RUN)
			{
				
				DEBUG("Instruction is %d\n",instr);
				
				switch(instr)
				{
					case OP_HALT:
					{
						DEBUG("DAVIM BASICLIB: Context %i (VM %i) executed HALT.\n", (int)context->which, (int)context->vm);
						context->state = DAVIM_CONTEXT_STATE_HALT;
						break;
					}
					case OP_PUSHC6:
					{
						context->pc++;
						
						DavimOpCode arg = context->capsule->data[context->pc];
						DavimVariable* var = my_malloc(sizeof(DavimVariable),s->pid);
						var->type = DAVIM_TYPE_INTEGER;
						var->value.var = arg;
						DEBUG("DAVIM BASICLIB: Context %i (VM %i): Executing pushc6 with arg %hi\n", (int)context->which,(int)context->vm, arg);

						SOS_CALL(s->stack_push, davim_stack_push_proto, context->opstack, var);
						break;
					}
					case OP_LED:
					{
						DavimVariable* arg = SOS_CALL(s->stack_pop, davim_stack_pop_proto, context->opstack);
	
						DEBUG("Inside OPled\n");
						
						if ( ! type_check(arg, DAVIM_TYPE_INTEGER) )
						{
							DEBUG("DAVIM BASICLIB: type error in LED\n");
							ret = -EINVAL;
						}
						else 
						{
							uint16_t val = arg->value.var;
							uint8_t op = (val >> 3) & 3;
							uint8_t led	= val & 7;
							DEBUG("DAVIM BASICLIB: Context %i (VM %i) EXECUTING OP_LED with op %i and led %i.\n", (int)context->which, (int)context->vm, (int)op, (int)led);
		      
							switch (op) 
							{
								case 0:			/* set */
								{
									if (led & 1) 
										ker_led(LED_RED_ON);
									else 
										ker_led(LED_RED_OFF);
									if (led & 2)
										ker_led(LED_GREEN_ON);
									else
										ker_led(LED_GREEN_OFF);
									if (led & 4)
										ker_led(LED_YELLOW_ON);
									else
										ker_led(LED_YELLOW_OFF);
									break;
								}
								case 1:			/* OFF 0 bits */
								{
									if (!(led & 1)) ker_led(LED_RED_OFF);
									if (!(led & 2)) ker_led(LED_GREEN_OFF);
									if (!(led & 4)) ker_led(LED_YELLOW_ON);
									break;
								}
	      						case 2:			/* on 1 bits */
	      						{
									if (led & 1) ker_led(LED_RED_ON);
									if (led & 2) ker_led(LED_GREEN_ON);
									if (led & 4) ker_led(LED_YELLOW_ON);
									break;
								}
								case 3:			/* TOGGLE 1 bits */
								{
									if (led & 1) ker_led(LED_RED_TOGGLE);
									if (led & 2) ker_led(LED_GREEN_TOGGLE);
									if (led & 4) ker_led(LED_YELLOW_TOGGLE);
									break;
								}
								default:
								{
									DEBUG("DAVIM BASICLIB: LED command had unknown operations.\n");
									break;
								}
							}
							ker_free(arg);
						}
						break;
					}
					case OP_SETTIMER0 + 0: 
					case OP_SETTIMER0 + 1:
					{
						uint8_t timer_type = (instr-OP_SETTIMER0)?TIMER_REPEAT:TIMER_ONE_SHOT;
						DavimVariable* arg = SOS_CALL(s->stack_pop, davim_stack_pop_proto, context->opstack);
						
						DEBUG("DAVIM BASICLIB: Setting timer 0 (VM %i) rate.\n", context->vm);
						if (!type_check(arg, DAVIM_TYPE_INTEGER)) {
							DEBUG("DAVIM BASICLIB: type error in SETTIMER0\n");
							ret = -EINVAL;
						} else {
							SOS_CALL(s->events_stop_timer, davim_event_manager_stop_timer_proto, DAVIM_EVENT_SINGLE_VM, context->vm, TIMER0_CONTEXT_ID);
							if (arg->value.var > 0) {
								SOS_CALL(s->events_start_timer, davim_event_manager_start_timer_proto, DAVIM_EVENT_SINGLE_VM, context->vm, TIMER0_CONTEXT_ID, timer_type, arg->value.var);
							}
						}
						ker_free(arg);
						break;
					}
                    case OP_GET_PHOTO_DATA:
                    {
                        context->state = DAVIM_CONTEXT_STATE_BLOCKED; // The context gets blocked
                        if (s->busy) {
                                queue_enqueue(context, &(s->getDataWaitQueue), context);
                                DEBUG("BASICLIB (VM %d / CON %d): Calling get photo data - Sensor busy\n",context->vm,context->which);
                        } else {
                                DEBUG("BASICLIB (VM %d / CON %d): Calling get photo data\n",context->vm,context->which);
                                ret = tryget(context, PHOTO);
                        }
                        break;
                    }
//                    case OP_POST_NET:
//                    {
//                    		//Just broadcast it, only for demonstration
//						DavimVariable* arg = SOS_CALL(s->stack_pop, davim_stack_pop_proto, context->opstack);
//	
//						DEBUG("Inside POST_NET\n");
//						
//						if ( ! type_check(arg, DAVIM_TYPE_INTEGER) )
//						{
//							DEBUG("DAVIM BASICLIB: type error in POST_NET\n");
//							ret = -EINVAL;
//						}
//						else 
//						{
//							s->sendbuf = arg->value.var;
//				            post_net(s->pid, 
//				                    s->pid, 
//				                    MSG_DATA_FROM_REMOTE,
//				                    sizeof(uint16_t),
//				                    (uint8_t*)&(s->sendbuf), 
//				                    0, 
//				                    BCAST_ADDRESS); 														
//						}
//						ker_free(arg);
//						break;
//                    }
					
				}
			}
			SOS_CALL(s->scheduler_opdone, davim_scheduler_opdone_proto, context);
					
			DEBUG("DAVIM BASICLIB: Instruction done, called opdone.\n");
			return ret;					
        }

        case MSG_DATA_READY: //DATA FROM GET_PHOTO_DATA ARE READY
        {
                MsgParam* param = (MsgParam*) (msg->data);
                uint16_t photo_data = param->word;
                s->busy = FALSE;
                if (s->executing != NULL) {
                        DEBUG("BASICLIB (VM %d / CON %d): Sensor returns - Context resumed.\n",s->executing->vm,s->executing->which);
						
						// Push the value on the stack
						DavimVariable* var = my_malloc(sizeof(DavimVariable),s->pid);
						var->type = DAVIM_TYPE_INTEGER;
						var->value.var = photo_data;
						SOS_CALL(s->stack_push, davim_stack_push_proto, s->executing->opstack, var);
						DEBUG("BASICLIB (VM %d / CON %d): Pushing sensor reading (%d) onto stack\n", s->executing->vm, s->executing->which, photo_data);
						
						// Resume context
                        s->executing->state = DAVIM_CONTEXT_STATE_UNBLOCKED;
                        SOS_CALL(s->scheduler_opdone, davim_scheduler_opdone_proto, s->executing);
                        s->executing = NULL;
                }

                if (!queue_empty(&(s->getDataWaitQueue))) {
                        DavimContext *current = queue_dequeue(NULL, &(s->getDataWaitQueue));
                        DEBUG("BASICLIB (VM %d / CON %d): Sensor returns - Some context waiting to get photo data.\n",current->vm, current->which);
                        tryget(current, PHOTO);
                }
                return SOS_OK;
        }

//		case MSG_DATA_FROM_REMOTE:
//		{
//			DEBUG("BASICLIB: Received data from remote\n");
//			return SOS_OK;
//		}

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the bytelength of the given opcode
static uint8_t bytelength(func_cb_ptr proto, DavimOpCode opcode)
{
	switch(opcode)
	{
		case OP_PUSHC6:
//		case OP_JNZ:
		{
			return 2;
		}
//		case OP_CALL:
//		case OP_POST:
//		{
//			return 3;
//		}
		default:
			return 1;
	}	
}

//! Returns the globals used by the given opcode (bitmask)
static void uses_globals(func_cb_ptr proto, DavimOpCode opcode, uint8_t* uses)
{
	//TODO
	//Nothing to do for the moment, only clearing out uses, change when implementing other
	int i;
	for(i = 0; i < (DAVIM_GLOBAL_COUNT + 7)/8; ++i){
		uses[i]=0;
	}
}

//! Try to get sensor data
static int8_t tryget(DavimContext *context, uint8_t sensor_type) {
		basiclib_state_t* s = (basiclib_state_t*)ker_get_module_state(DAVIM_LIB_BASICLIB_PID);
		
        if (ker_sensor_get_data(s->pid, sensor_type) == SOS_OK) {
        			DEBUG("BASICLIB: tryget: ker_sensor_get_data OK!\n");
                s->busy = TRUE;
                s->executing = context;
        } else {
        			DEBUG("BASICLIB: tryget: ker_sensor_get_data FAILED!!!!!!\n");
                queue_enqueue(context, &(s->getDataWaitQueue), context);
        }
        return SOS_OK;
}


static void halted(DavimVMId id) {
	basiclib_state_t* s = (basiclib_state_t*)ker_get_module_state(DAVIM_LIB_BASICLIB_PID);

	if(s->executing!=NULL && s->executing->vm == id){
		s->executing = NULL;
	}
	
	
	if(!queue_empty(&(s->getDataWaitQueue))){
		DavimContext* current = queue_dequeue(NULL, &(s->getDataWaitQueue));
		DavimContext* first = NULL;
		while(current!=first && !queue_empty(&(s->getDataWaitQueue))){
			if(current->vm != id){ //Queue again
				if(first==NULL){
					first = current;
				}
				queue_enqueue(current,&(s->getDataWaitQueue),current);
			} // otherwise: drop
			current = queue_dequeue(NULL, &(s->getDataWaitQueue));
		}
	}

}


#ifndef _MODULE_
mod_header_ptr basic_lib_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

