/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file netlib.c
 * @brief A network library for demonstration purposes
 * @author Wouter Horr&eacute;
 */
 
#include <module.h>
#include <led.h>
#include <davim.h>
#include <operationstore/library.h>
#include <engine/scheduler.h>
#include <vmstore/vm.h>
#include <vmstore/context.h>

#define OP_POST_NET 0x00

enum {
        MSG_DATA_FROM_REMOTE  = (MOD_MSG_START + 0),
};

//! The state of the network library
typedef struct {
	func_cb_ptr scheduler_opdone;
	func_cb_ptr stack_pop;
	sos_pid_t pid;
	int16_t sendbuf;					//for POST_NET
} netlib_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static uint8_t bytelength(func_cb_ptr, DavimOpCode opcode);
static void uses_globals(func_cb_ptr, DavimOpCode opcode, uint8_t*);

//! The module header of the basic library
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = DAVIM_LIB_NETLIB_PID,
	.state_size     = sizeof(netlib_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(DAVIM_LIB_NETLIB_PID),
    .num_timers     = 0,
	.num_sub_func   = 2,
	.num_prov_func  = 2,
	.module_handler = module,
	.funct = {
		//TODO error handler
		{NULL, "vav1", DAVIM_SCHEDULER_PID, DAVIM_SCHEDULER_OPDONE_FID},
		{NULL, "aav1", DAVIM_STACK_PID, DAVIM_STACK_POP_FID},
		{bytelength, "CCv1", DAVIM_LIB_NETLIB_PID, DAVIM_LIB_BYTELENGTH_FID},
		{uses_globals, "vCD1", DAVIM_LIB_NETLIB_PID, DAVIM_LIB_USES_GLOBALS_FID},
	},
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    netlib_state_t *s = (netlib_state_t *) state;

    switch (msg->type){

        case MSG_INIT:
		{
			s->pid = msg->did;
			s->sendbuf=0;
			post_short(DAVIM_LIB_MANAGER_PID,s->pid,DAVIM_CMD_LIB_MANAGER_ADD_LIBRARY,0,0,0);
			return SOS_OK;
		}


        case MSG_FINAL:
        {
        		post_short(DAVIM_LIB_MANAGER_PID,s->pid,DAVIM_CMD_LIB_MANAGER_REMOVE_LIBRARY,0,0,0);
			return SOS_OK;
        }
        case DAVIM_CMD_TRIGGER_GLOBAL:
        {
        		return SOS_OK;
        }

        case DAVIM_CMD_LIB_EXECUTE:
        {
        		int8_t ret = SOS_OK;
        		DavimExecutionMsg* exec = (DavimExecutionMsg*) msg->data;
			DavimContext* context = exec->context;
			DavimOpCode instr = exec->opcode;
						
			if(context->state == DAVIM_CONTEXT_STATE_RUN)
			{
				
				DEBUG("Instruction is %d\n",instr);
				
				switch(instr)
				{
                    case OP_POST_NET:
                    {
                    		//Just broadcast it, only for demonstration
						DavimVariable* arg = SOS_CALL(s->stack_pop, davim_stack_pop_proto, context->opstack);
	
						DEBUG("Inside POST_NET\n");
						
						if ( ! type_check(arg, DAVIM_TYPE_INTEGER) )
						{
							DEBUG("DAVIM NETLIB: type error in POST_NET\n");
							ret = -EINVAL;
						}
						else 
						{
							s->sendbuf = arg->value.var;
				            post_net(s->pid, 
				                    s->pid, 
				                    MSG_DATA_FROM_REMOTE,
				                    sizeof(uint16_t),
				                    (uint8_t*)&(s->sendbuf), 
				                    0, 
				                    BCAST_ADDRESS); 														
						}
						ker_free(arg);
						break;
                    }
				}
			}
			SOS_CALL(s->scheduler_opdone, davim_scheduler_opdone_proto, context);
					
			DEBUG("DAVIM NETLIB: Instruction done, called opdone.\n");
			return ret;					
        }

		case MSG_DATA_FROM_REMOTE:
		{
			DEBUG("NETLIB: Received data from remote\n");
			return SOS_OK;
		}

        default:
			return -EINVAL;
	}

    return SOS_OK;
}

//! Get the bytelength of the given opcode
static uint8_t bytelength(func_cb_ptr proto, DavimOpCode opcode)
{
	switch(opcode)
	{
		default:
			return 1;
	}	
}

//! Returns the globals used by the given opcode (bitmask)
static void uses_globals(func_cb_ptr proto, DavimOpCode opcode, uint8_t* uses)
{
	int i;
	for(i = 0; i < (DAVIM_GLOBAL_COUNT + 7)/8; ++i){
		uses[i]=0;
	}
}

#ifndef _MODULE_
mod_header_ptr netlib_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

