/**
 * @file virus.c
 * @brief The Virus module of FAViM
 * @author Wouter Horr&eacute;
 * 
 * FAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 * 
 * TODO
 * This code is an adaptation of the code of ... (license?)
 * 
 */
 
#include <module.h>

#include <virus/virus.h>
#include <random.h>
#include <led.h>
#include <sos_timer.h>

typedef enum {
    FAVIM_VIRUS_CAPSULE_MAINTAIN,    //!< Everything seems to be in order: default state
    FAVIM_VIRUS_CAPSULE_REQUEST,     //!< Somebody has something newer: receive capsules
    FAVIM_VIRUS_CAPSULE_RESPOND,     //!< Somebody has something older: send capsules
} FavimVirusCapsuleState;

typedef enum
{
    FAVIM_VIRUS_CONTEXT_IDLE,
    FAVIM_VIRUS_CONTEXT_RECV,
} FavimVirusContextState;

//! The state of the virus module
typedef struct {
	func_cb_ptr get_capsule;
	func_cb_ptr get_nb_capsules;
	func_cb_ptr get_all_capsule_versions;
	func_cb_ptr get_nb_contexts;
	func_cb_ptr get_present_contexts;
    sos_pid_t pid;
    FavimTrickleTimer versionTimer;
    FavimTrickleTimer capsuleTimer;
    FavimTrickleTimer contextTimer;
    //Capsules
    FavimVirusCapsuleState cap_state;
	FavimCapsule * currentCapsule; //!< Which capsule are we sending
	FavimCapsuleMsg * currentCapsuleMsg; //!< We are receiving a capsule, save it in here
    uint8_t needBitmask[FAVIM_VIRUS_BITMASK_SIZE];
    uint8_t sendBusy;
	uint8_t* versionNeedBitmask;
	int8_t comparison;
    FavimCapsuleChunkMsg chunk;
    FavimCapsuleStatusMsg statusMsg;
    //Contexts
    uint32_t contextVersion;
    FavimVirusContextState con_state;
    uint32_t contextRecvVersion;
    uint8_t contextRecvNbCon;
    FavimContextId * recvContexts;
} virus_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static void sendCapsulePacket() ;
static void sendVersionPacket(uint16_t to) ;
static void installChunk(FavimCapsuleChunkMsg* chunk) ;
static void chunkReceive(FavimCapsuleChunkMsg* chunk) ;
static void receiveVector(Message *msg) ;
static void sendCapsulePacket() ;
static void toMaintainState() ;
static void toResponseState(FavimCapsuleId capsuleNum) ;
//static int8_t registerCapsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule) ;
static void toRequestState(FavimCapsuleId capsuleNum, FavimCapsuleVersion version) ;
static void printNeedBitmask() ;
static uint8_t bitEntry(uint8_t* mask, uint8_t which) ;
static void newCounter(FavimTrickleTimer* timer) ;
static void decayNeedField() ;
static void clearInvalidBits() ;
static void removeNeededBit(uint16_t bit) ;
static int8_t getRandomChunkIndex() ;
static void sendContextMsg(uint16_t to);
static void receiveContextChunk(Message *msg);
static void checkContexts(FavimContextId *, uint8_t);
static void resetContextState(uint32_t, uint8_t);
static void installContextChunk(FavimContextMsg *);

//! The module header of the virus module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VIRUS_PID,
	.state_size     = sizeof(virus_state_t),
    .num_sub_func	= 5,
    .num_prov_func	= 0,
    .num_timers		= 3,
    .module_handler	= module,
    .funct = {
    		//TODO error handling
    		{NULL, "ayv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_CAPSULE_FID},
    		{NULL, "Cvv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_NB_CAPSULES_FID},
		{NULL, "avv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_ALL_VERSIONS_FID},
		{NULL, "Cvv0", FAVIM_VM_PID, FAVIM_VM_GET_NB_CONTEXTS_FID},
		{NULL, "avv0", FAVIM_VM_PID, FAVIM_VM_GET_PRESENT_CONTEXTS_FID},
		
//        {registerCapsule, "cya2", FAVIM_VIRUS_PID, FAVIM_VIRUS_REGISTERCAPSULE_FID},
    },
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    virus_state_t *s = (virus_state_t *) state;

    switch (msg->type){
        case MSG_INIT: {
            s->pid = msg->did;
            s->cap_state = FAVIM_VIRUS_CAPSULE_MAINTAIN;
            //TODO invalid capsuleid ?
            s->chunk.capsuleNum = 0; //MATE_CAPSULE_INVALID;
            s->currentCapsuleMsg = NULL;
            s->currentCapsule = NULL;
			s->chunk.version = 0;
            nmemset(s->needBitmask, 0, FAVIM_VIRUS_BITMASK_SIZE);
            s->versionNeedBitmask = NULL;
            s->comparison = 0;
			//nmemset(s->versions, 0, MATE_CAPSULE_NUM*sizeof(FavimCapsuleVersion));
            // Initialize (but do not start) the timers
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            s->capsuleTimer.interval = FAVIM_VIRUS_CAPSULE_TAU;
            s->contextTimer.interval = FAVIM_VIRUS_CONTEXT_TAU_MIN;
            newCounter(&s->versionTimer);
            newCounter(&s->capsuleTimer);
            newCounter(&s->contextTimer);

            DEBUG("FAVIM VIRUS: Initialized:\n\tchunk size=%i; bitmask size=%i\n", (int)FAVIM_VIRUS_CHUNK_SIZE, (int)FAVIM_VIRUS_BITMASK_SIZE);
            DEBUG("\tversion constants: quantum=%i;tau_l=%i;tau_h=%i;k=%i\n", (int)FAVIM_VIRUS_VERSION_TIMER, (int)FAVIM_VIRUS_VERSION_TAU_MIN, (int)FAVIM_VIRUS_VERSION_TAU_MAX, (int)FAVIM_VIRUS_VERSION_REDUNDANCY);
            DEBUG("\tcapsule constants: quantum=%i;tau=%i;repeat=%i;k=%i.\n", (int)FAVIM_VIRUS_CAPSULE_TIMER, (int)FAVIM_VIRUS_CAPSULE_TAU, (int)FAVIM_VIRUS_CAPSULE_REPEAT, (int)FAVIM_VIRUS_CAPSULE_REDUNDANCY);
            DEBUG("\tcontext constants: quantum=%i;tau_l=%i;tau_h=%i;k=%i\n", (int)FAVIM_VIRUS_CONTEXT_TIMER, (int)FAVIM_VIRUS_CONTEXT_TAU_MIN, (int)FAVIM_VIRUS_CONTEXT_TAU_MAX, (int)FAVIM_VIRUS_CONTEXT_REDUNDANCY);

            s->sendBusy = 0;
//            s->capsuleBusy = FALSE;

			s->con_state = FAVIM_VIRUS_CONTEXT_IDLE;
			s->contextVersion = 0;
			resetContextState(0,0);
            
            ker_timer_init(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, TIMER_REPEAT);
            ker_timer_init(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, TIMER_REPEAT);
            ker_timer_init(s->pid, FAVIM_VIRUS_CONTEXT_TIMER_ID, TIMER_REPEAT);
            ker_timer_start(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
            ker_timer_start(s->pid, FAVIM_VIRUS_CONTEXT_TIMER_ID, FAVIM_VIRUS_CONTEXT_TIMER);
			DEBUG("FAVIM VIRUS: ENTER HERE-------------------------Initialized\n");

			// TODO test code
//			FavimCapsuleChunkMsg * msg;
//			FavimCapsuleLength * bla;
			
//			msg = (FavimCapsuleChunkMsg *) my_malloc(sizeof(FavimCapsuleChunkMsg),FAVIM_VIRUS_PID);
//			msg->capsuleNum = 1;
//			msg->version = 1;
//			msg->piece = 0;
//			bla = (FavimCapsuleLength*) msg->chunk;
//			*bla = 5;
//			msg->chunk[2] = 0x01;
//			msg->chunk[3] = 0x01;
//			msg->chunk[4] = 0x1f;
//			msg->chunk[5] = 0x02;
//			msg->chunk[6] = 0x00;
//			post_long(s->pid,s->pid,FAVIM_MSG_VIRUS_CAPSULECHUNKMSG,sizeof(FavimCapsuleChunkMsg),msg,SOS_MSG_RELEASE);
			
//			msg = (FavimCapsuleChunkMsg *) my_malloc(sizeof(FavimCapsuleChunkMsg),FAVIM_VIRUS_PID);
//			if(msg == NULL){
//	    			ker_led(LED_GREEN_TOGGLE);
//	    			ker_led(LED_GREEN_TOGGLE);
//	    			ker_led(LED_GREEN_TOGGLE);
//	    			ker_led(LED_GREEN_TOGGLE);
//				return SOS_OK;
//			}
//			msg->capsuleNum = 0;
//			msg->version = 1;
//			msg->piece = 0;
//			bla = (FavimCapsuleLength*) msg->chunk;
//			*bla = 5;
//			msg->chunk[2] = 0x01;
//			msg->chunk[3] = 0x01;
//			msg->chunk[4] = 0x0a;
//			msg->chunk[5] = 0x0d;
//			msg->chunk[6] = 0x00;
//			post_long(s->pid,s->pid,FAVIM_MSG_VIRUS_CAPSULECHUNKMSG,sizeof(FavimCapsuleChunkMsg),msg,SOS_MSG_RELEASE);
			
			
			// TODO end of test code
            
            break;  
        }
        case MSG_FINAL: {
            ker_timer_stop(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID);
            ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
            break;
        }
        case FAVIM_CMD_VIRUS_VERSION_TIMER_TASK: {
            s->versionTimer.elapsed++;
            if (s->versionTimer.elapsed == s->versionTimer.threshold) {
                if (s->versionTimer.numHeard < FAVIM_VIRUS_VERSION_REDUNDANCY) {
                    sendVersionPacket(BCAST_ADDRESS);
                    s->versionTimer.numHeard++;
                }
            } else if (s->versionTimer.elapsed >= s->versionTimer.interval) {
                DEBUG("FAVIM VIRUS: Version tau elapsed, picking new t.\n");
                s->versionTimer.interval *= 2;
                if (s->versionTimer.interval > FAVIM_VIRUS_VERSION_TAU_MAX) {
                    s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MAX;
                }
                newCounter(&s->versionTimer);
                //TODO check this
                if(s->versionNeedBitmask != NULL){
                		ker_free(s->versionNeedBitmask);
                		s->versionNeedBitmask = NULL;
                }
                s->comparison = 0;
            }
            
             break; 
        }
        case FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK: {
            s->capsuleTimer.elapsed++;
            if (s->capsuleTimer.elapsed == s->capsuleTimer.threshold) {
                if (s->capsuleTimer.numHeard < FAVIM_VIRUS_CAPSULE_REDUNDANCY) {
                    sendCapsulePacket();
                    s->capsuleTimer.numHeard++;
                }
            } else if (s->capsuleTimer.elapsed >= s->capsuleTimer.interval) {
                newCounter(&s->capsuleTimer);
                decayNeedField();
            }
            
            break;
        }
        case FAVIM_CMD_VIRUS_CONTEXT_TIMER_TASK: {
            s->contextTimer.elapsed++;
            if (s->contextTimer.elapsed == s->contextTimer.threshold) {
                if (s->contextTimer.numHeard < FAVIM_VIRUS_CONTEXT_REDUNDANCY) {
                    sendContextMsg(BCAST_ADDRESS);
                    s->contextTimer.numHeard++;
                }
            } else if (s->contextTimer.elapsed >= s->contextTimer.interval) {
                DEBUG("FAVIM VIRUS: Context tau elapsed, picking new t.\n");
                s->contextTimer.interval *= 2;
                if (s->contextTimer.interval > FAVIM_VIRUS_CONTEXT_TAU_MAX) {
                    s->contextTimer.interval = FAVIM_VIRUS_CONTEXT_TAU_MAX;
                }
                newCounter(&s->contextTimer);
                s->con_state = FAVIM_VIRUS_CONTEXT_IDLE;
                resetContextState(0,0);
            }
        		break;
        }
        case FAVIM_CMD_VIRUS_CHECK_NEED_TASK: {
            uint8_t i;
            for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
                if (s->needBitmask[i] != 0) {
                    return -EINVAL;
                }
            }
            post_long(FAVIM_CAPSULE_MANAGER_PID, s->pid, FAVIM_CMD_CAPS_MANAGER_CAPS_ARRIVED, sizeof(FavimCapsuleMsg), s->currentCapsuleMsg, SOS_MSG_RELEASE);
            s->currentCapsuleMsg = NULL;
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
            toMaintainState();
            
            break;
        }
//		case MSG_FROM_USER: 
//		{
//			unsigned char a = (unsigned char)*(msg->data);
//			if (a == '1') {
//				DEBUG("Sending timer0 context\n");
//				s->chunk.version = 1;
//				s->chunk.capsuleNum = MATE_CONTEXT_TIMER0;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 5;
//				s->chunk.chunk[1] = 0;		//5 byte data length
//				s->chunk.chunk[2] = 0x01;	//requires basic lib
//				s->chunk.chunk[3] = 0x01;	//PUSH 31
//				s->chunk.chunk[4] = 0x1F;	//toggling all leds		
//				s->chunk.chunk[5] = 0x2;	//LED
//				s->chunk.chunk[6] = 0x0;	//HALT
//			}
//			else if (a == '2') { 
//				s->chunk.version++;
//				s->chunk.capsuleNum = MATE_CONTEXT_REBOOT;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 5;
//				s->chunk.chunk[1] = 0;
//				s->chunk.chunk[2] = 0x01;	//requires only basic library
//				s->chunk.chunk[3] = 0x01;	//PUSH
//				s->chunk.chunk[4] = 0x0A;	//10
//				s->chunk.chunk[5] = 0x0D;	//SETTIMER0 REPEAT
//				s->chunk.chunk[6] = 0x00;	//HALT
//			}
//			else { 
//				s->chunk.version++;
//				s->chunk.capsuleNum = MATE_CONTEXT_TIMER0;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 0x1D;
//				s->chunk.chunk[1] = 0x0;
//				s->chunk.chunk[2] = 0x3;	//requires only basic library
//				s->chunk.chunk[3] = 0x1;	//push
//				s->chunk.chunk[4] = 0x5;	//7
//				s->chunk.chunk[5] = 0x4;	//led
//				s->chunk.chunk[6] = 0x22;	//halt
//				s->chunk.chunk[7] = 0x21;	//halt
//				s->chunk.chunk[8] = 0x8;	//halt
//				s->chunk.chunk[9] = 0x4;	//halt
//				s->chunk.chunk[10] = 0xE;	//halt
//				s->chunk.chunk[11] = 0x1C;	//halt
//				s->chunk.chunk[12] = 0x1;	//halt
//				s->chunk.chunk[13] = 0xF;	//halt
//				s->chunk.chunk[14] = 0xF;	//halt
//				s->chunk.chunk[15] = 0x5;	//halt
//				s->chunk.chunk[16] = 0x6;	//halt
//				s->chunk.chunk[17] = 0x5;	//halt
//				s->chunk.chunk[18] = 0x22;	//halt
//				s->chunk.chunk[19] = 0x9;	//halt
//				s->chunk.chunk[20] = 0x11;	//halt
//				s->chunk.chunk[21] = 0x16;	//halt
//				s->chunk.chunk[22] = 0x11;	//halt
//				s->chunk.chunk[23] = 0x14;	//halt
//				s->chunk.chunk[24] = 0x14;	//halt
//				s->chunk.chunk[25] = 0x14;	//halt
//				s->chunk.chunk[26] = 0x14;	//halt
//				s->chunk.chunk[27] = 0x10;	//halt
//				s->chunk.chunk[28] = 0x8D;	//halt
//				s->chunk.chunk[29] = 0x23;	//halt
//				s->chunk.chunk[30] = 0x0;	//halt
//			}
//			post_long(s->pid, s->pid, FAVIM_MSG_VIRUS_CAPSULECHUNKMSG, sizeof(FavimCapsuleChunkMsg), &(s->chunk), 0);
//			break;
//		}
        case FAVIM_MSG_VIRUS_CAPSULECHUNKMSG: {
            FavimCapsuleChunkMsg* chunk = (FavimCapsuleChunkMsg*)msg->data;
            DEBUG("FAVIM VIRUS: Received chunk broadcast.\n");
	    		ker_led(LED_GREEN_TOGGLE);
            chunkReceive(chunk);
            break;
        }
        case FAVIM_MSG_VIRUS_CAPSULESTATUSMSG: {
            FavimCapsuleStatusMsg* statusMsg = (FavimCapsuleStatusMsg*)msg->data;
            FavimCapsuleVersion version = statusMsg->version;
            FavimCapsuleId which = statusMsg->capsuleNum;
            FavimCapsuleVersion installedVersion = (SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, which))->version;
            DEBUG("FAVIM VIRUS: Received status message for %d.%d\n", (int)which, (int)version);
            // If I'm requesting this capsule already...
            if ((s->cap_state == FAVIM_VIRUS_CAPSULE_REQUEST) &&
                (s->currentCapsuleMsg->id == which)) {
                // and this version, increment my counter.
                if (s->currentCapsuleMsg->capsule.version == version) {
                    s->versionTimer.numHeard++;
                }
                // and an older version, request the new version.
                else if (s->currentCapsuleMsg->capsule.version < version) {
                    toRequestState(which, version);
                }
            }
            // If I'm idle or responding, and need this capsule,
            // start requesting it.
            else if (((s->cap_state == FAVIM_VIRUS_CAPSULE_MAINTAIN) || (s->cap_state == FAVIM_VIRUS_CAPSULE_RESPOND)) &&
                (installedVersion < version)) {
                toRequestState(which, version);
            }
            // If I'm idle and have this capsule, respond with parts of it.
            else if (s->cap_state == FAVIM_VIRUS_CAPSULE_MAINTAIN &&
                (installedVersion == version)) {
                memcopy(s->needBitmask, statusMsg->bitmask, FAVIM_VIRUS_BITMASK_SIZE);
                toResponseState(which);
            }
    
            // If I'm responding, then add these to the parts I know are needed.
            if ((s->cap_state == FAVIM_VIRUS_CAPSULE_RESPOND) &&
                (s->chunk.capsuleNum == which) &&
                (s->currentCapsule->version == version)) {
                int i;
                for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
                    s->needBitmask[i] |= statusMsg->bitmask[i];
                }
            }
            
            break;
        }
        case FAVIM_MSG_VIRUS_VERSIONMSG: {
            receiveVector(msg);
            break;
        }
        case FAVIM_MSG_VIRUS_VERSIONREQUESTMSG: {
            sendVersionPacket(UART_ADDRESS);
            break;
        }
        case FAVIM_MSG_VIRUS_CONTEXTMSG: {
        		receiveContextChunk(msg);
        		break;
        }
        case MSG_PKT_SENDDONE: {
		    s->sendBusy--;
            break;
        }
        case MSG_TIMER_TIMEOUT: {
            MsgParam *t = (MsgParam *) msg->data;
            
            if (t->byte == FAVIM_VIRUS_VERSION_TIMER_ID) 
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_VERSION_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_CAPSULE_TIMER_ID)
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_CONTEXT_TIMER_ID)
            		post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_CONTEXT_TIMER_TASK, 0, 0, 0);
                
            break;
        }
        default:
            break;
    }
    
    return SOS_OK;
}

//static int8_t registerCapsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule) {
//    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
//    virus_state_t *s = (virus_state_t *)m->handler_state;
//    
//    DEBUG("FAVIM VIRUS: Registering capsule %i as %p\n", (int)id, capsule);
//    s->capsules[id] = capsule;
//    return SOS_OK;
//}

static void printNeedBitmask() {
#ifdef PC_PLATFORM
    int i;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    DEBUG("\tneedBitmask: ");
    for (i = 0; i < FAVIM_VIRUS_BITMASK_ENTRIES; i++) {
      DEBUG_SHORT("%s", (s->needBitmask[i/8] & (0x80 >> (i % 8)))? "1":"0");
    }
    DEBUG_SHORT("\n");
#endif
}
  
static uint8_t bitEntry(uint8_t* mask, uint8_t which) {
    return (mask[which >> 3] & (1 << (7 - (which & 7))));
}
  
  /* Select a new threshold, in the range [interval/2,interval], and
     clear out temporary state. */
static void newCounter(FavimTrickleTimer* timer) {
    timer->elapsed = 0;
    timer->threshold = timer->interval / 2;
    timer->threshold += ker_rand() % (timer->interval / 2);
    timer->numHeard = 0;
}

static void decayNeedField() {
    int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    DEBUG("FAVIM VIRUS: Decaying need field.\n");
    printNeedBitmask();
    for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
      s->needBitmask[i] &= (uint8_t)(ker_rand());
    }
    printNeedBitmask();
}

static void clearInvalidBits() {
    int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    uint8_t neededBits;
    uint16_t size = sizeof(FavimCapsule) - offsetof(FavimCapsule,dataSize)
    		+ sizeof(FavimOpCode) * (s->currentCapsuleMsg->capsule.dataSize - FAVIM_CAPSULE_SIZE);
    DEBUG("FAVIM VIRUS: Total size of capsule is %d-%d+%d*(%d-%d) = %d (%d).\n", (int)sizeof(FavimCapsule), (int)offsetof(FavimCapsule,dataSize), (int)sizeof(FavimOpCode), (int)s->currentCapsuleMsg->capsule.dataSize, (int)FAVIM_CAPSULE_SIZE, (int)(size), (int)FAVIM_VIRUS_CHUNK_SIZE);
    size += (FAVIM_VIRUS_CHUNK_SIZE - 1);
    size /= FAVIM_VIRUS_CHUNK_SIZE;
    neededBits = size;
    DEBUG("FAVIM VIRUS: Clearing need bits for caspule %i.%i. (%i)\n", (int)s->currentCapsuleMsg->id, (int)s->currentCapsuleMsg->capsule.version, (int)size);
    printNeedBitmask();
    // If neededBit isn't a multiple of 8, do some magic to clear some bits within a byte
    if (neededBits & 7) {
      uint8_t mask = (0x80 >> ((neededBits & 7) - 1));
      mask -= 1;
      mask = ~mask;
      s->needBitmask[neededBits / 8] &= mask;
      neededBits -= neededBits & 7;
      neededBits += 8;
    }
    // Clear whole bytes for the rest of the bitmask
    for (i = neededBits / 8; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
      s->needBitmask[i]  = 0;
    }
    printNeedBitmask();
}

static void removeNeededBit(uint16_t bit) {
    uint8_t mask = (1 << (7 - (bit & 7)));
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    DEBUG("FAVIM VIRUS: clear mask for %i: %hhx -> %hhx\n", bit/8, mask, ~mask);
    mask = ~mask;
    s->needBitmask[bit / 8] &= mask;
}

static void toRequestState(FavimCapsuleId capsuleNum, FavimCapsuleVersion version) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    // TODO moet/mag dit
    // Stop the capsule timer, we mustn't respond anymore
    ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
    DEBUG("FAVIM VIRUS: Moving to request state, capsule %i version %i.\n", (int)capsuleNum, (int)version);
    s->cap_state = FAVIM_VIRUS_CAPSULE_REQUEST;
    if(s->currentCapsuleMsg == NULL){
    		// allocate memory for it
    		uint8_t bla = sizeof(FavimCapsuleMsg);
    		s->currentCapsuleMsg = (FavimCapsuleMsg*) my_malloc(bla,FAVIM_VIRUS_PID);
    		if(s->currentCapsuleMsg == NULL){
    			// TODO error handling
    			ker_led(LED_GREEN_TOGGLE);
    			ker_led(LED_GREEN_TOGGLE);
    			ker_led(LED_GREEN_TOGGLE);
    			ker_led(LED_GREEN_TOGGLE);
    		}
    }
    s->currentCapsuleMsg->id = capsuleNum;
    s->currentCapsuleMsg->capsule.version = version;
    nmemset(s->needBitmask, 0xff, FAVIM_VIRUS_BITMASK_SIZE);
    
    s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
    newCounter(&s->versionTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
}

static void toResponseState(FavimCapsuleId capsuleNum) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->cap_state = FAVIM_VIRUS_CAPSULE_RESPOND;
    s->currentCapsule = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, capsuleNum);
    s->chunk.capsuleNum = capsuleNum;
    
    DEBUG("FAVIM VIRUS: Moving to response state, capsule %i version %i.\n", (int)capsuleNum, (int)s->currentCapsule->version);
    newCounter(&s->capsuleTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, FAVIM_VIRUS_CAPSULE_TIMER);
}

static void toMaintainState() {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->cap_state = FAVIM_VIRUS_CAPSULE_MAINTAIN;
    ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
}
 
static int8_t getRandomChunkIndex() {
    uint8_t idx;
    uint8_t count = 0;
    uint16_t size = sizeof(FavimCapsule) - offsetof(FavimCapsule, dataSize);
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    DEBUG("FAVIM VIRUS: Get random chunk index from\n");
    printNeedBitmask();
    size -= (FAVIM_CAPSULE_SIZE - s->currentCapsule->dataSize) * sizeof(FavimOpCode);
    size += (FAVIM_VIRUS_CHUNK_SIZE - 1);
    size /= FAVIM_VIRUS_CHUNK_SIZE;
    for (idx = 0; idx < size; idx++) {
        if (bitEntry(s->needBitmask, idx)) {
            count++;
        }
    }

    if (count == 0) {
      DEBUG("FAVIM VIRUS: Tried to select capsule chunk to send, but none were needed. Returning -1.\n");
      return -1;
    }

    count = (ker_rand() % count) + 1;

    for (idx = 0; idx < size; idx++) {
        if (bitEntry(s->needBitmask, idx)) {
            count--;
        }
        if (count == 0) {
            DEBUG("FAVIM VIRUS: Select capsule chunk %i to send.\n", (int)idx);
            return (int8_t)idx;
        }
    }
    DEBUG("FAVIM VIRUS: Tried to select capsule chunk to send, but encountered an error due to miscount. Returning -1.\n");
    return -1;
  }

static void sendCapsulePacket() {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    if (s->sendBusy) {
        DEBUG("FAVIM VIRUS: Tried to send capsule, but send busy.\n");
        return;
    }
    else {
        int8_t idx = getRandomChunkIndex();
        if (idx >= 0) {
            uint8_t* chunkPtr = (uint8_t*)&(s->currentCapsule->dataSize); // Send from dataSize to end of struct FavimCapsule
            chunkPtr += FAVIM_VIRUS_CHUNK_SIZE * idx;
            memcopy(s->chunk.chunk, chunkPtr, FAVIM_VIRUS_CHUNK_SIZE);

            s->chunk.piece = idx;
            s->chunk.version = s->currentCapsule->version;
            if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULECHUNKMSG, sizeof(FavimCapsuleChunkMsg), &s->chunk, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
                DEBUG("FAVIM VIRUS: Sent chunk %i of capsule %i.%i.\n", (int)idx, (int)s->chunk.capsuleNum, (int)s->currentCapsule->version);
                s->sendBusy++;
                removeNeededBit(idx);
            } else {
                DEBUG("FAVIM VIRUS: Sending chunk %i of capsule %i.%i FAILED.\n", (int)idx, (int)s->chunk.capsuleNum, (int)s->currentCapsule->version);
            }
        } else {
            DEBUG("FAVIM VIRUS: Tried to send capsule, but nothing to send. Return to maintain state.\n");
            toMaintainState();
        }
    }
}

static void sendVersionPacket(uint16_t to) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    if (s->sendBusy) {
      DEBUG("FAVIM VIRUS: Tried to send version packet, but send busy.\n");
      return;
    }
    
    if (s->cap_state == FAVIM_VIRUS_CAPSULE_MAINTAIN || s->cap_state == FAVIM_VIRUS_CAPSULE_RESPOND) {
    		uint8_t nb_cap = SOS_CALL(s->get_nb_capsules, favim_capsule_manager_get_nb_capsules_proto);
    		uint8_t send_cap = 0;
    		uint8_t piece = 0;
    		FavimCapsuleIdVersionPair * versions = SOS_CALL(s->get_all_capsule_versions, favim_capsule_manager_get_all_versions_proto);
    		ker_change_own(versions, FAVIM_VIRUS_PID);
    		uint8_t nb_chunks = (nb_cap + FAVIM_VIRUS_VERSION_CHUNK_SIZE - 1) / FAVIM_VIRUS_VERSION_CHUNK_SIZE;
    		uint8_t chunk_size;
    		FavimVersionChunkMsg * msg;
    		while(piece < nb_chunks){
	    		if((nb_cap-send_cap) > FAVIM_VIRUS_VERSION_CHUNK_SIZE){
	    			chunk_size = FAVIM_VIRUS_VERSION_CHUNK_SIZE;
	    		} else {
	    			chunk_size = (nb_cap-send_cap);
	    		}
			msg = (FavimVersionChunkMsg*) my_malloc(sizeof(FavimVersionChunkMsg)+chunk_size*sizeof(FavimCapsuleIdVersionPair),FAVIM_VIRUS_PID);
			if(msg == NULL){
				//TODO error handling
			}
			msg->nb_chunks = nb_chunks;
			msg->piece = piece++;
			msg->chunk_length = chunk_size;
	    		int i;
			for(i=0; i < chunk_size; ++i){
				msg->versions[i] = versions[send_cap++];
			}
	        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_VERSIONMSG, sizeof(FavimVersionChunkMsg)+chunk_size*sizeof(FavimCapsuleIdVersionPair), msg, SOS_MSG_RELIABLE | SOS_MSG_RELEASE, to) == SOS_OK) {
	            DEBUG("FAVIM VIRUS: Sent version vector chunk %i.\n", piece-1);
	            s->sendBusy++;
	        }
    		}
    		ker_free(versions);
    } else if (s->cap_state == FAVIM_VIRUS_CAPSULE_REQUEST && to == BCAST_ADDRESS) {
        s->statusMsg.capsuleNum = s->currentCapsuleMsg->id;
        s->statusMsg.version = s->currentCapsuleMsg->capsule.version;
        memcopy(s->statusMsg.bitmask, s->needBitmask, FAVIM_VIRUS_BITMASK_SIZE);
        DEBUG("FAVIM VIRUS: In state %i, sending bitmask:\n", (int)s->cap_state);
        printNeedBitmask();
        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULESTATUSMSG, sizeof(FavimCapsuleStatusMsg), &s->statusMsg, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
            DEBUG("FAVIM VIRUS: Sent capsule version bitmask.\n");
            s->sendBusy++;
        }
    } else {
        DEBUG("FAVIM VIRUS: In invalid state for sending version packet: %i\n", (int)s->cap_state);
        return;
    }
}
  
static void receiveVector(Message *msg) {
	DEBUG("Received version vector\n");
	ker_led(LED_RED_TOGGLE);
	int i = 0;
//    int8_t comparison = 0; // msg is: -1 older, 0 same, 1 newer
    FavimVersionChunkMsg* versionMsg = (FavimVersionChunkMsg*)msg->data;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    if(s->versionNeedBitmask == NULL){
    		uint8_t neededBits = versionMsg->nb_chunks;
    		uint8_t size = (neededBits+7)/8;
    		s->versionNeedBitmask = (uint8_t*)my_malloc(size,s->pid);
    		if(s->versionNeedBitmask == NULL){
    			//TODO error handling
    		}
    		for(i=0; i < size; ++i){
    			s->versionNeedBitmask[i] = 0xff;
    		}
    		// If needed, set some bits to zero in the last byte of the bitmask
    		if(neededBits & 7){
	      uint8_t mask = (0x80 >> ((neededBits & 7) - 1));
	      mask -= 1;
	      mask = ~mask;
	      s->versionNeedBitmask[size-1] &= mask;
    		}
    }
    
    // Remove bit of received chunk from needBitmask
    uint8_t mask = (1 << (7 - (versionMsg->piece & 7)));
    DEBUG("FAVIM VIRUS: clear version mask for %i: %hhx -> %hhx\n", (versionMsg->piece)/8, mask, ~mask);
    mask = ~mask;
    s->versionNeedBitmask[versionMsg->piece / 8] &= mask;
	
	// If we're requesting check version of the capsule we are requesting
    if (s->cap_state == FAVIM_VIRUS_CAPSULE_REQUEST) {
		FavimCapsuleVersion version = 0;
    		BOOL present = FALSE;
    		for(i=0; i < versionMsg->chunk_length; ++i){
    			if(versionMsg->versions[i].id == s->currentCapsuleMsg->id){
    				version = versionMsg->versions[i].version;
    				present = TRUE;
    				break;
    			}
    		}
    		if(!present){
    			DEBUG("FAVIM VIRUS: Requested capsule not in this version vector chunk\n");
    			return;
    		}
        if (version > s->currentCapsuleMsg->capsule.version) {
            toRequestState(s->currentCapsuleMsg->id, version);
        } else if (version == s->currentCapsuleMsg->capsule.version) {
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
            ker_timer_restart(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
        } else {
            DEBUG("FAVIM VIRUS: In request state for %i.%i already.\n", (int)s->currentCapsuleMsg->id, (int)s->currentCapsuleMsg->capsule.version);
        }
        return;
    }

	// Otherwise, investigate whole vector
	for(i=0; i < versionMsg->chunk_length; ++i){
		FavimCapsule * caps = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, versionMsg->versions[i].id);
		if(caps != NULL){
			FavimCapsuleVersion version = caps->version;
			if(versionMsg->versions[i].version > version){
				s->comparison = 1;
				break;
			} else if (versionMsg->versions[i].version < version){
				s->comparison = -1;
				break;
			}
		}
	}
		
    
    // Requests are given precedence over responses: there's no
    // point updating someone to something that's out of date.
    if (s->comparison == 0) {
    		//check bitmask, if all chunks heard, increment numHeard, otherwise return
    		for(i=0; i < (versionMsg->nb_chunks+7)/8; ++i){
    			if(s->versionNeedBitmask[i] != 0){
    				return;
    			}
    		}
        DEBUG("FAVIM VIRUS: Heard identical version vector, suppressing.\n");
        s->versionTimer.numHeard++;
    } else if (s->comparison == 1) {
        DEBUG("Heard newer vector, requesting.\n");
        if (s->cap_state != FAVIM_VIRUS_CAPSULE_REQUEST) {
            toRequestState(versionMsg->versions[i].id, versionMsg->versions[i].version);
        }
    } else if (s->comparison == -1) {
        DEBUG("Heard older vector, updating.\n");
        if (s->cap_state == FAVIM_VIRUS_CAPSULE_MAINTAIN) {
            nmemset(s->needBitmask, 0xff, FAVIM_VIRUS_BITMASK_SIZE);
            toResponseState(versionMsg->versions[i].id);
        }
    }
    return;
}

static void installChunk(FavimCapsuleChunkMsg* chunk) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
//    FavimCapsule * cap = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, chunk->capsuleNum);
    FavimOpCode* destPtr = (FavimOpCode*)&(s->currentCapsuleMsg->capsule.dataSize);

    destPtr += chunk->piece * FAVIM_VIRUS_CHUNK_SIZE;
    DEBUG("FAVIM VIRUS: Copying chunk %d of capsule %d version %d.\n", (int)chunk->piece, (int)chunk->capsuleNum, (int)chunk->version);
    memcopy(destPtr, chunk->chunk, FAVIM_VIRUS_CHUNK_SIZE);
    // Remove that bit entry
    if (chunk->piece == 0) { // If it's the chunk metadata, clear out bits
      clearInvalidBits();
    }
    removeNeededBit(chunk->piece);
    printNeedBitmask();
    post_short(s->pid, s->pid, FAVIM_CMD_VIRUS_CHECK_NEED_TASK, 0, 0, 0);
}


static void chunkReceive(FavimCapsuleChunkMsg* chunk) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    s->capsuleTimer.numHeard++;
    if (s->cap_state == FAVIM_VIRUS_CAPSULE_REQUEST) {
		
        if (chunk->capsuleNum == s->currentCapsuleMsg->id &&
            chunk->version == s->currentCapsuleMsg->capsule.version &&
            bitEntry(s->needBitmask, chunk->piece)) {
			
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        } else if (chunk->capsuleNum == s->currentCapsuleMsg->id 
        				&& chunk->version > s->currentCapsuleMsg->capsule.version) {
            toRequestState(chunk->capsuleNum, chunk->version);
			
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        }
    } else if (s->cap_state == FAVIM_VIRUS_CAPSULE_MAINTAIN) { //TODO This can't happen while in response state????
    		FavimCapsule * cap = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, chunk->capsuleNum);
        if (chunk->version > cap->version) {
            toRequestState(chunk->capsuleNum, chunk->version);
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        }
    } 
}

static void sendContextMsg(uint16_t to) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    if (s->sendBusy) {
      DEBUG("FAVIM VIRUS: Tried to send context packet, but send busy.\n");
      return;
    }
    
	uint8_t nb_con = SOS_CALL(s->get_nb_contexts, favim_vm_get_nb_contexts_proto);
	uint8_t send_con = 0;
	uint8_t piece = 0;
	FavimContextId * ids = SOS_CALL(s->get_present_contexts, favim_vm_get_present_contexts_proto);
	ker_change_own(ids, FAVIM_VIRUS_PID);
	uint8_t nb_chunks = (nb_con + FAVIM_VIRUS_CONTEXT_CHUNK_SIZE - 1) / FAVIM_VIRUS_CONTEXT_CHUNK_SIZE;
	uint8_t chunk_size;
	FavimContextMsg * msg;
	while(piece < nb_chunks){
		if((nb_con-send_con) > FAVIM_VIRUS_CONTEXT_CHUNK_SIZE){
			chunk_size = FAVIM_VIRUS_CONTEXT_CHUNK_SIZE;
		} else {
			chunk_size = (nb_con-send_con);
		}
		msg = (FavimContextMsg*) my_malloc(sizeof(FavimContextMsg)+chunk_size*sizeof(FavimContextId),FAVIM_VIRUS_PID);
		if(msg == NULL){
			//TODO error handling
		}
		msg->nb_con = nb_con;
		msg->piece = piece++;
		msg->chunk_length = chunk_size;
		msg->version = s->contextVersion;
		int i;
		for(i=0; i < chunk_size; ++i){
			msg->ids[i] = ids[send_con++];
		}
	    if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CONTEXTMSG, sizeof(FavimContextMsg)+chunk_size*sizeof(FavimContextId), msg, SOS_MSG_RELIABLE | SOS_MSG_RELEASE, to) == SOS_OK) {
	        DEBUG("FAVIM VIRUS: Sent context vector chunk %i, version %i.\n", piece-1, s->contextVersion);
	        s->sendBusy++;
	    }
	}
	ker_free(ids);
}

//! Receive a chunk of the context vector
static void receiveContextChunk(Message *msg){
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	FavimContextMsg * conmsg = (FavimContextMsg *) msg->data;

	DEBUG("FAVIM VIRUS: received context chunk, version: %i\n", (int)conmsg->version);

	//TODO check version inside one/multiple packet cases?
	if(conmsg->nb_con <= FAVIM_VIRUS_CONTEXT_CHUNK_SIZE){
		// Only one chunk
		if(conmsg->version == s->contextVersion){
			s->contextTimer.numHeard++;
	        DEBUG("FAVIM VIRUS: Heard identical context vector, suppressing.\n");
			return;
		} else if (conmsg->version < s->contextVersion) {
			s->contextTimer.interval = FAVIM_VIRUS_CONTEXT_TAU_MIN;
			newCounter(&(s->contextTimer));
			ker_timer_restart(s->pid, FAVIM_VIRUS_CONTEXT_TIMER_ID, FAVIM_VIRUS_CONTEXT_TIMER);
		} else {
				checkContexts(conmsg->ids, conmsg->nb_con);
				s->contextVersion = conmsg->version;
				s->contextTimer.interval = FAVIM_VIRUS_CONTEXT_TAU_MIN;
				newCounter(&(s->contextTimer));
				ker_timer_restart(s->pid, FAVIM_VIRUS_CONTEXT_TIMER_ID, FAVIM_VIRUS_CONTEXT_TIMER);
		}
	} else {
		// Multiple chunks
		if(s->con_state == FAVIM_VIRUS_CONTEXT_RECV){
			if(conmsg->version == s->contextRecvVersion){
				//copy to datastruct
				installContextChunk(conmsg);
			} else if (conmsg->version > s->contextRecvVersion) {
				// reset data structs and start receiving newer version
				resetContextState(conmsg->version,conmsg->nb_con);
				installContextChunk(conmsg);
			} //other case: old data, discard
			if(s->contextRecvNbCon == conmsg->nb_con){
				//This was the last chunk
				if(s->contextRecvVersion == s->contextVersion) {
					// identical one
					s->contextTimer.numHeard++;
			        DEBUG("FAVIM VIRUS: Heard identical context vector, suppressing.\n");
					return;
				} else {
					// other one
					checkContexts(s->recvContexts,s->contextRecvNbCon);
					s->contextVersion = s->contextRecvVersion;
					s->contextTimer.interval = FAVIM_VIRUS_CONTEXT_TAU_MIN;
					newCounter(&(s->contextTimer));
					ker_timer_restart(s->pid, FAVIM_VIRUS_CONTEXT_TIMER_ID, FAVIM_VIRUS_CONTEXT_TIMER);
					resetContextState(0,0);
				}
			}
		} else if(s->con_state == FAVIM_VIRUS_CONTEXT_IDLE){
			if(conmsg->version >= s->contextVersion){
				//Start receiving it (identical versions needed to prevent counting 2 chunks as 2 packets)
				s->con_state = FAVIM_VIRUS_CONTEXT_RECV;
				resetContextState(conmsg->version,conmsg->nb_con);
				installContextChunk(conmsg);
			} //else: older data, but we are also out of date, so don't send our status
		}
	}
}

static void resetContextState(uint32_t version, uint8_t nb_con){
	int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	
	s->contextRecvVersion = version;
	s->contextRecvNbCon = 0;
	if(s->recvContexts != NULL) {
		ker_free(s->recvContexts);
		s->recvContexts = NULL;
	}
	if(nb_con > 0){
		s->recvContexts = my_malloc(nb_con * sizeof(FavimContextId),s->pid);
		if(s->recvContexts == NULL){
			//TODO error handling
		}
		for(i=0; i<nb_con; ++i){
			s->recvContexts[nb_con] = 0xff;
		}
	}
}

static void installContextChunk(FavimContextMsg * conmsg){
	int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    FavimContextId * ptr = s->recvContexts + conmsg->piece*FAVIM_VIRUS_CONTEXT_CHUNK_SIZE;
    
    if(ptr[0] == 0xFF) {
    		//This is a new piece
    		s->contextRecvNbCon += conmsg->chunk_length;
    		for(i=0; i < conmsg->chunk_length; ++i){
    			ptr[i] = conmsg->ids[i];
    		}
    }
}

//! Compare the incoming context status and the currently present one
//! @pre Arrays are sorted!
static void checkContexts(FavimContextId * incoming, uint8_t inc_size){
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	
	FavimContextId * current = SOS_CALL(s->get_present_contexts, favim_vm_get_present_contexts_proto);
	ker_change_own(current, FAVIM_VIRUS_PID);
	uint8_t cur_size = SOS_CALL(s->get_nb_contexts, favim_vm_get_nb_contexts_proto);
	
	uint8_t inc_index = 0;
	uint8_t cur_index = 0;
	
	while(inc_index < inc_size && cur_index < cur_size){
		if(incoming[inc_index] == current[cur_index]){
			// Both have this one
			cur_index++;
			inc_index++;
		} else if(incoming[inc_index] > current[cur_index]){
			// Present, but no longer needed
			post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REMOVE_CONTEXT,current[cur_index],0,0);
			cur_index++;
		} else if(incoming[inc_index] < current[cur_index]){
			// Not present, but needed
			post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_ADD_CONTEXT,incoming[inc_index],0,0);
			inc_index++;
		}
	}
	while(inc_index < inc_size){
		// These are new ones.
		post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_ADD_CONTEXT,incoming[inc_index],0,0);
		inc_index++;
	}
	while(cur_index < cur_size){
		// These are no longer needed
		post_short(FAVIM_VM_PID,s->pid,FAVIM_CMD_VM_REMOVE_CONTEXT,current[cur_index],0,0);
		cur_index++;
	}
	
}


#ifndef _MODULE_
mod_header_ptr virus_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif


