/**
 * @file virus.c
 * @brief The Virus module of FAViM
 * @author Wouter Horr&eacute;
 * 
 * FAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 * 
 * TODO
 * This code is an adaptation of the code of ... (license?)
 * 
 */
 
#include <module.h>

#include <virus/virus.h>
#include <random.h>
#include <led.h>
#include <sos_timer.h>

// Forward declarations
static int8_t module(void *state, Message *e);
static void newCounter(FavimTrickleTimer* timer) ;
//Capsules
static void sendCapsulePacket() ;
static void sendCapsuleVersionPacket(uint16_t to) ;
static void installCapsuleChunk(FavimCapsuleChunkMsg* chunk) ;
static void capsuleChunkReceive(FavimCapsuleChunkMsg* chunk) ;
static void receiveCapsuleVersionVector(Message *msg) ;
static void sendCapsulePacket() ;
static void toCapsuleMaintainState() ;
static void toCapsuleResponseState(FavimCapsuleId capsuleNum) ;
//static int8_t registerCapsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule) ;
static void toCapsuleRequestState(FavimCapsuleId capsuleNum, FavimCapsuleVersion version) ;
static void printCapsuleNeedBitmask() ;
static uint8_t capsuleBitEntry(uint8_t* mask, uint8_t which) ;
static void capsuleDecayNeedField() ;
static void capsuleClearInvalidBits(FavimCapsuleLength dataSize) ;
static void capsuleRemoveNeededBit(uint16_t bit) ;
static int8_t capsuleGetRandomChunkIndex() ;
//VMs
static void toVMMaintainState();
static void toVMResponseState(FavimVMId id);
static void toVMRequestState(FavimVMId id, FavimVMVersion version);
static void sendVMPacket();
static void sendVMVersionPacket(uint16_t to);
static void receiveVMVersionVector(FavimVMVersionMsg *msg);
static void receiveVM(FavimVMMsg *msg);

//! The module header of the virus module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VIRUS_PID,
	.state_size     = sizeof(virus_state_t),
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id			= ehtons(FAVIM_VIRUS_PID),
    .num_sub_func	= 6,
    .num_prov_func	= 0,
    .num_timers		= 4,
    .module_handler	= module,
    .funct = {
    		//TODO error handling
		{NULL, "ayv1", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_CAPSULE_FID},
		{NULL, "Cvv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_NB_CAPSULES_FID},
		{NULL, "avv0", FAVIM_CAPSULE_MANAGER_PID, FAVIM_CAPSULE_MANAGER_GET_ALL_VERSIONS_FID},
		{NULL, "vCa1", FAVIM_VM_PID, FAVIM_VM_GET_DESCRIPTION_FID},
		{NULL, "CCv1", FAVIM_VM_MANAGER_PID, FAVIM_VM_MAN_VM_PRESENT_FID},
		{NULL, "LCv1", FAVIM_VM_MANAGER_PID, FAVIM_VM_MAN_GET_VM_VERSION_FID},
    },
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    virus_state_t *s = (virus_state_t *) state;

    switch (msg->type){
        case MSG_INIT: {
            s->pid = msg->did;
            
            // CAPSULES
            s->cap_state = FAVIM_VIRUS_MAINTAIN;
            //TODO invalid capsuleid ?
            s->cap_chunk.capsuleNum = 0; //MATE_CAPSULE_INVALID;
            s->cap_currentMsg = NULL;
            s->cap_current = NULL;
			s->cap_chunk.version = 0;
            nmemset(s->cap_needBitmask, 0, FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
            s->cap_versionNeedBitmask = NULL;
//            s->cap_comparison = 0;
            // Initialize (but do not start) the timers
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            s->cap_timer.interval = FAVIM_VIRUS_CAPSULE_TAU;
            newCounter(&s->cap_versionTimer);
            newCounter(&s->cap_timer);
            ker_timer_init(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID, TIMER_REPEAT);
            ker_timer_init(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, TIMER_REPEAT);
            ker_timer_start(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID, FAVIM_VIRUS_CAPSULE_VERSION_TIMER);

            DEBUG("FAVIM VIRUS: Initialized:\n\tcapsule chunk size=%i; bitmask size=%i\n", (int)FAVIM_VIRUS_CAPSULE_CHUNK_SIZE, (int)FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
            DEBUG("\tcapsule version constants: quantum=%i;tau_l=%i;tau_h=%i;k=%i\n", (int)FAVIM_VIRUS_CAPSULE_VERSION_TIMER, (int)FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN, (int)FAVIM_VIRUS_CAPSULE_VERSION_TAU_MAX, (int)FAVIM_VIRUS_CAPSULE_VERSION_REDUNDANCY);
            DEBUG("\tcapsule constants: quantum=%i;tau=%i;k=%i.\n", (int)FAVIM_VIRUS_CAPSULE_TIMER, (int)FAVIM_VIRUS_CAPSULE_TAU, (int)FAVIM_VIRUS_CAPSULE_REDUNDANCY);
            
            //VMS
            s->vm_state = FAVIM_VIRUS_MAINTAIN;
            s->vm_current_msg.desc.id = FAVIM_VM_INVALID;
            s->vm_versionTimer.interval = 2*FAVIM_VIRUS_VM_VERSION_TAU_MIN; //differentiate a little bit with the capsule timer
            s->vm_timer.interval = FAVIM_VIRUS_VM_TAU;
            newCounter(&s->vm_timer);
            newCounter(&s->vm_versionTimer);
            ker_timer_init(s->pid, FAVIM_VIRUS_VM_VERSION_TIMER_ID, TIMER_REPEAT);
            ker_timer_init(s->pid, FAVIM_VIRUS_VM_TIMER_ID, TIMER_REPEAT);
            ker_timer_start(s->pid, FAVIM_VIRUS_VM_VERSION_TIMER_ID, FAVIM_VIRUS_VM_VERSION_TIMER);

            s->sendBusy = 0;

			DEBUG("FAVIM VIRUS: ENTER HERE-------------------------Initialized\n");
            
            break;  
        }
        case MSG_FINAL: {
            ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID);
            ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
            break;
        }
        
        // Capsules
        
        case FAVIM_CMD_VIRUS_CAPSULE_VERSION_TIMER_TASK: {
            s->cap_versionTimer.elapsed++;
            if (s->cap_versionTimer.elapsed == s->cap_versionTimer.threshold) {
                if (s->cap_versionTimer.numHeard < FAVIM_VIRUS_CAPSULE_VERSION_REDUNDANCY) {
                    sendCapsuleVersionPacket(BCAST_ADDRESS);
                    s->cap_versionTimer.numHeard++;
                }
            } else if (s->cap_versionTimer.elapsed >= s->cap_versionTimer.interval) {
                DEBUG("FAVIM VIRUS: Capsule version tau elapsed, picking new t.\n");
                s->cap_versionTimer.interval *= 2;
                if (s->cap_versionTimer.interval > FAVIM_VIRUS_CAPSULE_VERSION_TAU_MAX) {
                    s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MAX;
                }
                newCounter(&s->cap_versionTimer);
                if(s->cap_versionNeedBitmask != NULL){
                		ker_free(s->cap_versionNeedBitmask);
                		s->cap_versionNeedBitmask = NULL;
                }
//                s->cap_comparison = 0;
            }
            
             break; 
        }
        case FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK: {
            s->cap_timer.elapsed++;
            if (s->cap_timer.elapsed == s->cap_timer.threshold) {
                if (s->cap_timer.numHeard < FAVIM_VIRUS_CAPSULE_REDUNDANCY) {
                    sendCapsulePacket();
                    s->cap_timer.numHeard++;
                }
            } else if (s->cap_timer.elapsed >= s->cap_timer.interval) {
                newCounter(&s->cap_timer);
                capsuleDecayNeedField();
            }
            
            break;
        }
        case FAVIM_CMD_VIRUS_CAPSULE_CHECK_NEED_TASK: {
            uint8_t i;
            for (i = 0; i < FAVIM_VIRUS_CAPSULE_BITMASK_SIZE; i++) {
                if (s->cap_needBitmask[i] != 0) {
                    return -EINVAL;
                }
            }
            post_long(FAVIM_CAPSULE_MANAGER_PID, s->pid, FAVIM_CMD_CAPS_MANAGER_CAPS_ARRIVED, sizeof(FavimCapsuleMsg), s->cap_currentMsg, SOS_MSG_RELEASE);
            s->cap_currentMsg = NULL;
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            newCounter(&s->cap_versionTimer);
            toCapsuleMaintainState();
            
            break;
        }
        case FAVIM_MSG_VIRUS_CAPSULECHUNKMSG: {
            FavimCapsuleChunkMsg* chunk = (FavimCapsuleChunkMsg*)msg->data;
            DEBUG("FAVIM VIRUS: Received chunk broadcast.\n");
//	    		ker_led(LED_GREEN_TOGGLE);
            capsuleChunkReceive(chunk);
            break;
        }
        case FAVIM_MSG_VIRUS_CAPSULESTATUSMSG: {
            FavimCapsuleStatusMsg* statusMsg = (FavimCapsuleStatusMsg*)msg->data;
            FavimCapsuleVersion version = statusMsg->version;
            FavimCapsuleId which = statusMsg->capsuleNum;
            FavimCapsuleVersion installedVersion = (SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, which))->version;
            DEBUG("FAVIM VIRUS: Received status message for %d.%d\n", (int)which, (int)version);
            // If I'm requesting this capsule already...
            if ((s->cap_state == FAVIM_VIRUS_REQUEST) &&
                (s->cap_currentMsg->id == which)) {
                // and this version, increment my counter.
                if (s->cap_currentMsg->capsule.version == version) {
                    s->cap_versionTimer.numHeard++;
                }
                // and an older version, request the new version.
                else if (s->cap_currentMsg->capsule.version < version) {
                    toCapsuleRequestState(which, version);
                }
            }
            // If I'm idle or responding, and need this capsule,
            // start requesting it.
            else if (((s->cap_state == FAVIM_VIRUS_MAINTAIN) || (s->cap_state == FAVIM_VIRUS_RESPOND)) &&
                (installedVersion < version)) {
                toCapsuleRequestState(which, version);
            }
            // If I'm idle and have this capsule, respond with parts of it.
            else if (s->cap_state == FAVIM_VIRUS_MAINTAIN &&
                (installedVersion == version)) {
                memcopy(s->cap_needBitmask, statusMsg->bitmask, FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
                toCapsuleResponseState(which);
            }
    
            // If I'm responding, then add these to the parts I know are needed.
            if ((s->cap_state == FAVIM_VIRUS_RESPOND) &&
                (s->cap_chunk.capsuleNum == which) &&
                (s->cap_current->version == version)) {
                int i;
                for (i = 0; i < FAVIM_VIRUS_CAPSULE_BITMASK_SIZE; i++) {
                    s->cap_needBitmask[i] |= statusMsg->bitmask[i];
                }
            }
            
            break;
        }
        case FAVIM_MSG_VIRUS_CAPSULE_VERSIONMSG: {
            receiveCapsuleVersionVector(msg);
            break;
        }
        
        // VMs
        
        case FAVIM_CMD_VIRUS_VM_VERSION_TIMER_TASK: {
            s->vm_versionTimer.elapsed++;
            if (s->vm_versionTimer.elapsed == s->vm_versionTimer.threshold) {
                if (s->vm_versionTimer.numHeard < FAVIM_VIRUS_VM_VERSION_REDUNDANCY) {
                    sendVMVersionPacket(BCAST_ADDRESS);
                    s->vm_versionTimer.numHeard++;
                }
            } else if (s->vm_versionTimer.elapsed >= s->vm_versionTimer.interval) {
                DEBUG("FAVIM VIRUS: VM version tau elapsed, picking new t.\n");
                s->vm_versionTimer.interval *= 2;
                if (s->vm_versionTimer.interval > FAVIM_VIRUS_VM_VERSION_TAU_MAX) {
                    s->vm_versionTimer.interval = FAVIM_VIRUS_VM_VERSION_TAU_MAX;
                }
                newCounter(&s->vm_versionTimer);
//                if(s->vm_versionNeedBitmask != NULL){
//                		ker_free(s->vm_versionNeedBitmask);
//                		s->vm_versionNeedBitmask = NULL;
//                }
            }
            
        		break;
        }
        case FAVIM_CMD_VIRUS_VM_TIMER_TASK: {
            s->vm_timer.elapsed++;
            if (s->vm_timer.elapsed == s->vm_timer.threshold) {
                if (s->vm_timer.numHeard < FAVIM_VIRUS_VM_REDUNDANCY) {
                    sendVMPacket();
                    s->vm_timer.numHeard++;
                }
            } else if (s->vm_timer.elapsed >= s->vm_timer.interval) {
            		//We only send VM's for one period
            		toVMMaintainState();
            }
        		break;
        }
        case FAVIM_MSG_VIRUS_VM_VERSIONMSG: {
        		receiveVMVersionVector((FavimVMVersionMsg*)msg->data);
        		break;
        }
        case FAVIM_MSG_VIRUS_VMMSG: {
        		receiveVM((FavimVMMsg*)ker_msg_take_data(s->pid,msg));
        		break;
        }
//        case FAVIM_MSG_VIRUS_CAPSULE_VERSIONREQUESTMSG: {
//            sendCapsuleVersionPacket(UART_ADDRESS);
//            break;
//        }
        case MSG_PKT_SENDDONE: {
		    s->sendBusy--;
            break;
        }
        case MSG_TIMER_TIMEOUT: {
            MsgParam *t = (MsgParam *) msg->data;
            
            if (t->byte == FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID) 
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_CAPSULE_VERSION_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_CAPSULE_TIMER_ID)
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_VM_VERSION_TIMER_ID)
				post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_VM_VERSION_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_VM_TIMER_ID)
				post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_VM_TIMER_TASK, 0, 0, 0);
                
            break;
        }
        default:
            break;
    }
    
    return SOS_OK;
}

  /* Select a new threshold, in the range [interval/2,interval], and
     clear out temporary state. */
static void newCounter(FavimTrickleTimer* timer) {
    timer->elapsed = 0;
    timer->threshold = timer->interval / 2;
    timer->threshold += ker_rand() % (timer->interval / 2);
    timer->numHeard = 0;
}

static void printCapsuleNeedBitmask() {
#ifdef PC_PLATFORM
    int i;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    DEBUG("\tneedBitmask: ");
    for (i = 0; i < FAVIM_VIRUS_CAPSULE_BITMASK_ENTRIES; i++) {
      DEBUG_SHORT("%s", (s->cap_needBitmask[i/8] & (0x80 >> (i % 8)))? "1":"0");
    }
    DEBUG_SHORT("\n");
#endif
}
  
static uint8_t capsuleBitEntry(uint8_t* mask, uint8_t which) {
    return (mask[which >> 3] & (1 << (7 - (which & 7))));
}
  
static void capsuleDecayNeedField() {
    int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    DEBUG("FAVIM VIRUS: Decaying need field.\n");
    printCapsuleNeedBitmask();
    for (i = 0; i < FAVIM_VIRUS_CAPSULE_BITMASK_SIZE; i++) {
      s->cap_needBitmask[i] &= (uint8_t)(ker_rand());
    }
    printCapsuleNeedBitmask();
}

static void capsuleClearInvalidBits(FavimCapsuleLength dataSize) {
    int i;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    uint8_t neededBits = (dataSize + FAVIM_VIRUS_CAPSULE_CHUNK_SIZE - 1) / FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
//    FavimCapsuleLength size = s->cap_currentMsg->dataSize
//    DEBUG("FAVIM VIRUS: Total size of capsule is %d-%d+%d*(%d-%d) = %d (%d).\n", (int)sizeof(FavimCapsule), (int)offsetof(FavimCapsule,dataSize), (int)sizeof(FavimOpCode), (int)s->cap_currentMsg->capsule.dataSize, (int)FAVIM_CAPSULE_SIZE, (int)(size), (int)FAVIM_VIRUS_CAPSULE_CHUNK_SIZE);
//    uint8_t size = 
//    neededBits = size;
    DEBUG("FAVIM VIRUS: Clearing need bits for caspule %i.%i. (%i)\n", (int)s->cap_currentMsg->id, (int)s->cap_currentMsg->capsule.version, (int)neededBits);
    printCapsuleNeedBitmask();
    // If neededBit isn't a multiple of 8, do some magic to clear some bits within a byte
    if (neededBits & 7) {
      uint8_t mask = (0x80 >> ((neededBits & 7) - 1));
      mask -= 1;
      mask = ~mask;
      s->cap_needBitmask[neededBits / 8] &= mask;
      neededBits -= neededBits & 7;
      neededBits += 8;
    }
    // Clear whole bytes for the rest of the bitmask
    for (i = neededBits / 8; i < FAVIM_VIRUS_CAPSULE_BITMASK_SIZE; i++) {
      s->cap_needBitmask[i]  = 0;
    }
    printCapsuleNeedBitmask();
}

static void capsuleRemoveNeededBit(uint16_t bit) {
    uint8_t mask = (1 << (7 - (bit & 7)));
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    DEBUG("FAVIM VIRUS: clear mask for %i: %hhx -> %hhx\n", bit/8, mask, ~mask);
    mask = ~mask;
    s->cap_needBitmask[bit / 8] &= mask;
}

static void toCapsuleRequestState(FavimCapsuleId capsuleNum, FavimCapsuleVersion version) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    // TODO moet/mag dit
    // Stop the capsule timer, we mustn't respond anymore
    ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
    DEBUG("FAVIM VIRUS: Moving to capsule request state, capsule %i version %i.\n", (int)capsuleNum, (int)version);
    s->cap_state = FAVIM_VIRUS_REQUEST;
    if(s->cap_currentMsg == NULL){
    		// allocate memory for it
    		uint8_t bla = sizeof(FavimCapsuleMsg);
    		s->cap_currentMsg = (FavimCapsuleMsg*) my_malloc(bla,FAVIM_VIRUS_PID);
    		if(s->cap_currentMsg == NULL){
    			// TODO error handling
    			panic();
    			return;
    		}
    }
    s->cap_currentMsg->id = capsuleNum;
    s->cap_currentMsg->capsule.version = version;
    s->cap_currentMsg->capsule.data = NULL;
    nmemset(s->cap_needBitmask, 0xff, FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
    
    s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
    newCounter(&s->cap_versionTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID, FAVIM_VIRUS_CAPSULE_VERSION_TIMER);
}

static void toCapsuleResponseState(FavimCapsuleId capsuleNum) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->cap_state = FAVIM_VIRUS_RESPOND;
    s->cap_current = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, capsuleNum);
    s->cap_chunk.capsuleNum = capsuleNum;
    
    DEBUG("FAVIM VIRUS: Moving to capsule response state, capsule %i version %i.\n", (int)capsuleNum, (int)s->cap_current->version);
    newCounter(&s->cap_timer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, FAVIM_VIRUS_CAPSULE_TIMER);

    s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
    newCounter(&s->cap_versionTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID, FAVIM_VIRUS_CAPSULE_VERSION_TIMER);
}

static void toCapsuleMaintainState() {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->cap_state = FAVIM_VIRUS_MAINTAIN;
    ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
}
 
static int8_t capsuleGetRandomChunkIndex() {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    uint8_t idx;
    uint8_t count = 0;
    uint16_t size = (s->cap_current->dataSize + FAVIM_VIRUS_CAPSULE_CHUNK_SIZE - 1) / FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;

    DEBUG("FAVIM VIRUS: Get random chunk index from\n");
    printCapsuleNeedBitmask();
//    size -= (FAVIM_CAPSULE_SIZE - s->cap_current->dataSize) * sizeof(FavimOpCode);
//    size += (FAVIM_VIRUS_CAPSULE_CHUNK_SIZE - 1);
//    size /= FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
    for (idx = 0; idx < size; idx++) {
        if (capsuleBitEntry(s->cap_needBitmask, idx)) {
            count++;
        }
    }

    if (count == 0) {
      DEBUG("FAVIM VIRUS: Tried to select capsule chunk to send, but none were needed. Returning -1.\n");
      return -1;
    }

    count = (ker_rand() % count) + 1;

    for (idx = 0; idx < size; idx++) {
        if (capsuleBitEntry(s->cap_needBitmask, idx)) {
            count--;
        }
        if (count == 0) {
            DEBUG("FAVIM VIRUS: Select capsule chunk %i to send.\n", (int)idx);
            return (int8_t)idx;
        }
    }
    DEBUG("FAVIM VIRUS: Tried to select capsule chunk to send, but encountered an error due to miscount. Returning -1.\n");
    return -1;
  }

static void sendCapsulePacket() {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    if (s->sendBusy) {
        DEBUG("FAVIM VIRUS: Tried to send capsule, but send busy.\n");
        return;
    }
    else {
        int8_t idx = capsuleGetRandomChunkIndex();
        if (idx >= 0) {
            s->cap_chunk.piece = idx;
            s->cap_chunk.version = s->cap_current->version;
            s->cap_chunk.length = s->cap_current->dataSize;
            uint8_t* chunkPtr = (uint8_t*)s->cap_current->data; 
            chunkPtr += FAVIM_VIRUS_CAPSULE_CHUNK_SIZE * idx;
            size_t size = FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
            if(s->cap_current->dataSize - idx * FAVIM_VIRUS_CAPSULE_CHUNK_SIZE < FAVIM_VIRUS_CAPSULE_CHUNK_SIZE){
            		size = s->cap_current->dataSize - idx * FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
            }	
            memcopy(s->cap_chunk.chunk, chunkPtr, size);

            if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULECHUNKMSG, sizeof(FavimCapsuleChunkMsg), &s->cap_chunk, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
                DEBUG("FAVIM VIRUS: Sent chunk %i of capsule %i.%i.\n", (int)idx, (int)s->cap_chunk.capsuleNum, (int)s->cap_current->version);
                s->sendBusy++;
                capsuleRemoveNeededBit(idx);
            } else {
                DEBUG("FAVIM VIRUS: Sending chunk %i of capsule %i.%i FAILED.\n", (int)idx, (int)s->cap_chunk.capsuleNum, (int)s->cap_current->version);
            }
        } else {
            DEBUG("FAVIM VIRUS: Tried to send capsule, but nothing to send. Return to maintain state.\n");
            toCapsuleMaintainState();
        }
    }
}

static void sendCapsuleVersionPacket(uint16_t to) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	
//	ker_led(LED_GREEN_TOGGLE);
    if (s->sendBusy) {
      DEBUG("FAVIM VIRUS: Tried to send capsule version packet, but send busy.\n");
      return;
    }
    
    if (s->cap_state == FAVIM_VIRUS_MAINTAIN || s->cap_state == FAVIM_VIRUS_RESPOND) {
    		uint8_t nb_cap = SOS_CALL(s->get_nb_capsules, favim_capsule_manager_get_nb_capsules_proto);
    		uint8_t send_cap = 0;
    		uint8_t piece = 0;
    		FavimCapsuleIdVersionPair * versions = SOS_CALL(s->get_all_capsule_versions, favim_capsule_manager_get_all_versions_proto);
    		ker_change_own(versions, FAVIM_VIRUS_PID);
    		uint8_t nb_chunks = (nb_cap + FAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE - 1) / FAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE;
    		uint8_t chunk_size;
    		FavimCapsuleVersionChunkMsg * msg;
    		while(piece < nb_chunks){
	    		if((nb_cap-send_cap) > FAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE){
	    			chunk_size = FAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE;
	    		} else {
	    			chunk_size = (nb_cap-send_cap);
	    		}
			msg = (FavimCapsuleVersionChunkMsg*) my_malloc(sizeof(FavimCapsuleVersionChunkMsg)+chunk_size*sizeof(FavimCapsuleIdVersionPair),FAVIM_VIRUS_PID);
			if(msg == NULL){
				//TODO error handling
				panic();
				return;
			}
			msg->nb_chunks = nb_chunks;
			msg->piece = piece++;
			msg->chunk_length = chunk_size;
	    		int i;
			for(i=0; i < chunk_size; ++i){
				msg->versions[i] = versions[send_cap++];
			}
	        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULE_VERSIONMSG, sizeof(FavimCapsuleVersionChunkMsg)+chunk_size*sizeof(FavimCapsuleIdVersionPair), msg, SOS_MSG_RELIABLE | SOS_MSG_RELEASE, to) == SOS_OK) {
	            DEBUG("FAVIM VIRUS: Sent version vector chunk %i.\n", piece-1);
	            s->sendBusy++;
	        }
    		}
    		ker_free(versions);
    } else if (s->cap_state == FAVIM_VIRUS_REQUEST && to == BCAST_ADDRESS) {
        s->cap_statusMsg.capsuleNum = s->cap_currentMsg->id;
        s->cap_statusMsg.version = s->cap_currentMsg->capsule.version;
        memcopy(s->cap_statusMsg.bitmask, s->cap_needBitmask, FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
        DEBUG("FAVIM VIRUS: In state %i, sending bitmask:\n", (int)s->cap_state);
        printCapsuleNeedBitmask();
        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULESTATUSMSG, sizeof(FavimCapsuleStatusMsg), &s->cap_statusMsg, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
            DEBUG("FAVIM VIRUS: Sent capsule version bitmask.\n");
            s->sendBusy++;
        }
    } else {
        DEBUG("FAVIM VIRUS: In invalid state for sending version packet: %i\n", (int)s->cap_state);
        return;
    }
}
  
static void receiveCapsuleVersionVector(Message *msg) {
	DEBUG("Received version vector\n");
//	ker_led(LED_RED_TOGGLE);
	int i = 0;
    FavimCapsuleVersionChunkMsg* versionMsg = (FavimCapsuleVersionChunkMsg*)msg->data;
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
	// If we're requesting, check version of the capsule we are requesting
    if (s->cap_state == FAVIM_VIRUS_REQUEST) {
		FavimCapsuleVersion version = 0;
    		BOOL present = FALSE;
    		for(i=0; i < versionMsg->chunk_length; ++i){
    			if(versionMsg->versions[i].id == s->cap_currentMsg->id){
    				version = versionMsg->versions[i].version;
    				present = TRUE;
    				break;
    			}
    		}
    		if(!present){
    			DEBUG("FAVIM VIRUS: Requested capsule not in this version vector chunk\n");
    			return;
    		}
        if (version > s->cap_currentMsg->capsule.version) {
            toCapsuleRequestState(s->cap_currentMsg->id, version);
        } else if (version == s->cap_currentMsg->capsule.version) {
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            newCounter(&s->cap_versionTimer);
            ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_VERSION_TIMER_ID, FAVIM_VIRUS_CAPSULE_VERSION_TIMER);
        } else {
            DEBUG("FAVIM VIRUS: In request state for %i.%i already.\n", (int)s->cap_currentMsg->id, (int)s->cap_currentMsg->capsule.version);
        }
        return;
    }

	// otherwise investigate whole vector
    if(versionMsg->nb_chunks > 1 && s->cap_versionNeedBitmask == NULL){
    		uint8_t neededBits = versionMsg->nb_chunks;
    		uint8_t size = (neededBits+7)/8;
    		s->cap_versionNeedBitmask = (uint8_t*)my_malloc(size,s->pid);
    		if(s->cap_versionNeedBitmask == NULL){
    			//TODO error handling
    			panic();
    			return;
    		}
    		for(i=0; i < size; ++i){
    			s->cap_versionNeedBitmask[i] = 0xff;
    		}
    		// If needed, set some bits to zero in the last byte of the bitmask
    		if(neededBits & 7){
	      uint8_t mask = (0x80 >> ((neededBits & 7) - 1));
	      mask -= 1;
	      mask = ~mask;
	      s->cap_versionNeedBitmask[size-1] &= mask;
    		}
    }

	int8_t comparison = 0; // msg is: -1 older, 0 same, 1 newer
	uint8_t pos=0;	
	for(i=0; i < versionMsg->chunk_length; ++i){
		FavimCapsule * caps = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, versionMsg->versions[i].id);
		if(caps != NULL){
			FavimCapsuleVersion version = caps->version;
			if(versionMsg->versions[i].version > version){
				pos = i;
				comparison = 1;
				break; // if we found a newer version: immediately stop and go requesting it
			} else if (comparison==0 && versionMsg->versions[i].version < version){ //if we have found nothing special this far and now find an older version, update comparison and pos
				pos = i;
				comparison = -1;
			}
		}
	}
		
    
    // Requests are given precedence over responses: there's no
    // point updating someone to something that's out of date.
    if (comparison == 0) {
	    if(s->cap_versionNeedBitmask != NULL){
			// Remove bit of received chunk from cap_versionNeedBitmask
			uint8_t mask = (1 << (7 - (versionMsg->piece & 7)));
			DEBUG("FAVIM VIRUS: clear version mask for %i: %hhx -> %hhx\n", (versionMsg->piece)/8, mask, ~mask);
			mask = ~mask;
			s->cap_versionNeedBitmask[versionMsg->piece / 8] &= mask;
	    		
			//check bitmask, if all chunks heard identically, increment numHeard, otherwise return
			for(i=0; i < (versionMsg->nb_chunks+7)/8; ++i){
				if(s->cap_versionNeedBitmask[i] != 0){
					return;
				}
			}
			
			// If we got here: clean up version need bitmask
			ker_free(s->cap_versionNeedBitmask);
			s->cap_versionNeedBitmask=NULL;
	    }
        DEBUG("FAVIM VIRUS: Heard identical capsule version vector, suppressing.\n");
        s->cap_versionTimer.numHeard++;
    } else if (comparison == 1) {
        DEBUG("Heard newer capsule vector chunk, requesting.\n");
        if (s->cap_state != FAVIM_VIRUS_REQUEST) {
            toCapsuleRequestState(versionMsg->versions[pos].id, versionMsg->versions[i].version);
        }
    } else if (comparison == -1) {
        DEBUG("Heard older capsule vector chunk, updating.\n");
        if (s->cap_state == FAVIM_VIRUS_MAINTAIN) {
            nmemset(s->cap_needBitmask, 0xff, FAVIM_VIRUS_CAPSULE_BITMASK_SIZE);
            toCapsuleResponseState(versionMsg->versions[pos].id);
        }
    }
    return;
}

static void installCapsuleChunk(FavimCapsuleChunkMsg* chunk) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
//    FavimCapsule * cap = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, cap_chunk->capsuleNum);
	s->cap_currentMsg->capsule.dataSize = chunk->length;
	if(s->cap_currentMsg->capsule.data == NULL){
		s->cap_currentMsg->capsule.data = my_malloc(sizeof(FavimOpCode)*chunk->length,s->pid);
		if(s->cap_currentMsg->capsule.data == NULL){
			DEBUG("FAVIM VIRUS: unable to get memory to install capsule\n");
			panic();
			return;
		}
	}
    FavimOpCode* destPtr = (FavimOpCode*)s->cap_currentMsg->capsule.data;

    destPtr += chunk->piece * FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
    DEBUG("FAVIM VIRUS: Copying chunk %d of capsule %d version %d.\n", (int)chunk->piece, (int)chunk->capsuleNum, (int)chunk->version);
    size_t size = FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
    if(chunk->length - chunk->piece * FAVIM_VIRUS_CAPSULE_CHUNK_SIZE < FAVIM_VIRUS_CAPSULE_CHUNK_SIZE){
    		size = chunk->length - chunk->piece * FAVIM_VIRUS_CAPSULE_CHUNK_SIZE;
    }
    memcopy(destPtr, chunk->chunk, size);
    // Remove that bit entry
    if (chunk->piece == 0) { // If it's the chunk metadata, clear out bits
      capsuleClearInvalidBits(chunk->length);
    }
    capsuleRemoveNeededBit(chunk->piece);
    printCapsuleNeedBitmask();
    post_short(s->pid, s->pid, FAVIM_CMD_VIRUS_CAPSULE_CHECK_NEED_TASK, 0, 0, 0);
}


static void capsuleChunkReceive(FavimCapsuleChunkMsg* chunk) {
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);

    s->cap_timer.numHeard++;
    if (s->cap_state == FAVIM_VIRUS_REQUEST) {
		
        if (chunk->capsuleNum == s->cap_currentMsg->id &&
            chunk->version == s->cap_currentMsg->capsule.version &&
            capsuleBitEntry(s->cap_needBitmask, chunk->piece)) {
			
            installCapsuleChunk(chunk);
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            newCounter(&s->cap_versionTimer);
        } else if (chunk->capsuleNum == s->cap_currentMsg->id 
        				&& chunk->version > s->cap_currentMsg->capsule.version) {
            toCapsuleRequestState(chunk->capsuleNum, chunk->version);
			
            installCapsuleChunk(chunk);
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            newCounter(&s->cap_versionTimer);
        }
    } else if (s->cap_state == FAVIM_VIRUS_MAINTAIN) { //TODO This can't happen while in response state????
    		FavimCapsule * cap = SOS_CALL(s->get_capsule, favim_capsule_manager_get_capsule_proto, chunk->capsuleNum);
    		if(cap == NULL){
    			//We don't have this context ? -> ignore
    			return;
    		}
        if (chunk->version > cap->version) {
            toCapsuleRequestState(chunk->capsuleNum, chunk->version);
            installCapsuleChunk(chunk);
            s->cap_versionTimer.interval = FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN;
            newCounter(&s->cap_versionTimer);
        }
    } 
}

// VMs
static void toVMMaintainState()
{
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->vm_state = FAVIM_VIRUS_MAINTAIN;
    ker_timer_stop(s->pid, FAVIM_VIRUS_VM_TIMER_ID);
}

static void toVMResponseState(FavimVMId id)
{
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    s->vm_state = FAVIM_VIRUS_RESPOND;
    s->vm_current_msg.desc.id = id;
    
    DEBUG("FAVIM VIRUS: Moving to vm response state, vm %i.\n", (int)id);
    newCounter(&s->vm_timer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_VM_TIMER_ID, FAVIM_VIRUS_VM_TIMER);
}

static void toVMRequestState(FavimVMId id, FavimVMVersion version)
{
    virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    
    // Stop the vm timer, we mustn't respond anymore
    ker_timer_stop(s->pid, FAVIM_VIRUS_VM_TIMER_ID);
    DEBUG("FAVIM VIRUS: Moving to vm request state, vm %i version %i.\n", (int)id, (int)version);
    s->vm_state = FAVIM_VIRUS_REQUEST;
    s->vm_current_msg.desc.id = id;
    s->vm_current_msg.version = version;

    s->vm_versionTimer.interval = FAVIM_VIRUS_VM_VERSION_TAU_MIN;
    newCounter(&s->vm_versionTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_VM_VERSION_TIMER_ID, FAVIM_VIRUS_VM_VERSION_TIMER);
}

static void sendVMPacket()
{
	virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
    if (s->sendBusy) {
        DEBUG("FAVIM VIRUS: Tried to send capsule, but send busy.\n");
        return;
    } else {
		DEBUG("FAVIM VIRUS: Sending vm packet, vm %i\n", (int)s->vm_current_msg.desc.id);
		
		s->vm_current_msg.version = SOS_CALL(s->get_vm_version, favim_vm_man_get_vm_version_proto, s->vm_current_msg.desc.id);
		
		if(!SOS_CALL(s->vm_present, favim_vm_man_vm_present_proto, s->vm_current_msg.desc.id)){
			s->vm_current_msg.present = FALSE;
		} else {
			s->vm_current_msg.present = TRUE;
			SOS_CALL(s->get_vm_description, favim_vm_get_description_proto, s->vm_current_msg.desc.id, &(s->vm_current_msg.desc));
			if(s->vm_current_msg.desc.id == FAVIM_VM_INVALID){
				//For some reason the vm is not present???
				//TODO error handling
				return;
			}
		}
		
		if(post_net(s->pid,s->pid,FAVIM_MSG_VIRUS_VMMSG,sizeof(FavimVMMsg),&(s->vm_current_msg),SOS_MSG_RELIABLE,BCAST_ADDRESS)==SOS_OK){
			DEBUG("Sent vm done\n");
			s->sendBusy++;
		} else {
            DEBUG("FAVIM VIRUS: Sending vm FAILED.\n");
        }
    }
}

static void sendVMVersionPacket(uint16_t to)
{
	int i;
	virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	FavimVMVersionMsg * msg = my_malloc(sizeof(FavimVMVersionMsg),s->pid);
	if(msg == NULL){
		// TODO error handling
		panic();
		return;
	}
	if(s->sendBusy){
		DEBUG("Tried to sent VM version packet, but send busy!\n");
		ker_free(msg);
		return;
	}
	
	for(i=0; i<FAVIM_VM_COUNT; ++i){
		//msg->versions[i].id = i;
		msg->versions[i]/*.version*/ = SOS_CALL(s->get_vm_version, favim_vm_man_get_vm_version_proto, i);
	}
	
	if(post_net(s->pid,s->pid,FAVIM_MSG_VIRUS_VM_VERSIONMSG,sizeof(FavimVMVersionMsg), msg, SOS_MSG_RELIABLE|SOS_MSG_RELEASE, to) == SOS_OK) {
		DEBUG("FAVIM VIRUS: sent vm version packet\n");
//		ker_led(LED_GREEN_TOGGLE);
		s->sendBusy++;
	} else {
		DEBUG("FAVIM VIRUS: sending vm version packet failed\n");
	}
}

static void receiveVMVersionVector(FavimVMVersionMsg *msg)
{
	virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	if(s->vm_state == FAVIM_VIRUS_REQUEST){
		if(msg->versions[s->vm_current_msg.desc.id]/*.version*/ >= s->vm_current_msg.version){
			toVMRequestState(s->vm_current_msg.desc.id,msg->versions[s->vm_current_msg.desc.id]/*.version*/);
		} else {
			//We are already requesting this one with a newer version
			DEBUG("FAVIM VIRUS VM: already in request state for a newer version\n");
		}
		return;
	}
	int comparison=0;
	int i;
	for(i=0; i<FAVIM_VM_COUNT; ++i){
		FavimVMVersion vers = SOS_CALL(s->get_vm_version, favim_vm_man_get_vm_version_proto, i /*msg->versions[i].id*/);
		if(vers < msg->versions[i]/*.version*/){
			comparison = 1;
			break;
		} else if(vers > msg->versions[i]/*.version*/) {
			comparison = -1;
			break;
		}
	}
	if(comparison == 0){
		//Heard identical vector
		s->vm_versionTimer.numHeard++;
		DEBUG("FAVIM VIRUS: heard identical vm version vector: suppressing\n");
	} else if (comparison == 1){
		//Heard newer vector
		DEBUG("FAVIM VIRUS: heard newer vm version vector: requesting\n");
		toVMRequestState(i /*msg->versions[i].id*/,msg->versions[i]/*.version*/);
	} else if (comparison == -1){
		//Heard older vector
		toVMResponseState(i /*msg->versions[i].id*/);
	}
}

static void receiveVM(FavimVMMsg *msg)
{
	virus_state_t *s = (virus_state_t *) ker_get_module_state(FAVIM_VIRUS_PID);
	
	if(msg->desc.id >= FAVIM_VM_COUNT){ //  < 0 not necessary (unsigned)
		//TODO error handling
		ker_free(msg);
		return;
	}
	
	FavimVMVersion vers = SOS_CALL(s->get_vm_version, favim_vm_man_get_vm_version_proto, msg->desc.id);

	if(s->vm_state == FAVIM_VIRUS_REQUEST && s->vm_current_msg.desc.id == msg->desc.id && s->vm_current_msg.version == msg->version){
		post_long(FAVIM_VM_MANAGER_PID, s->pid, FAVIM_CMD_VM_MAN_VMMSG_ARRIVED,sizeof(FavimVMMsg),msg,SOS_MSG_RELEASE);
		//We were requesting this one, return to maintain state
		toVMMaintainState();
	} else if(s->vm_state == FAVIM_VIRUS_RESPOND && s->vm_current_msg.desc.id == msg->desc.id && vers == msg->version){
		// We are responding for this one
		s->vm_timer.numHeard++;
		ker_free(msg);
	} else { // maintain state
		if ( vers < msg->version){
			//we are out of date, install it
			post_long(FAVIM_VM_MANAGER_PID, s->pid, FAVIM_CMD_VM_MAN_VMMSG_ARRIVED,sizeof(FavimVMMsg),msg,SOS_MSG_RELEASE);
			s->vm_versionTimer.interval = FAVIM_VIRUS_VM_VERSION_TAU_MIN;
			newCounter(&(s->vm_versionTimer));
		} else if (vers > msg->version){
			//other one is out of date, start responding
			toVMResponseState(msg->desc.id);
			ker_free(msg);
		}
	}
}

#ifndef _MODULE_
mod_header_ptr virus_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif


