/**
 * @file virus.h
 * 
 * FAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 *
 */

#ifndef _FAVIM_VIRUS_H_
#define _FAVIM_VIRUS_H_

typedef struct FavimTrickleTimer FavimTrickleTimer;
typedef struct FavimCapsuleChunkMsg FavimCapsuleChunkMsg;
typedef struct FavimCapsuleStatusMsg FavimCapsuleStatusMsg;
typedef struct FavimCapsuleVersionChunkMsg FavimCapsuleVersionChunkMsg;
typedef struct FavimVMVersionMsg FavimVMVersionMsg;
typedef struct FavimVMMsg FavimVMMsg;
//typedef struct FavimVMIdVersionPair FavimVMIdVersionPair;

#include "../favim.h"
#include "../capsulestore/capsule.h"
#include "../capsulestore/capsule_manager.h"
#include "../vmstore/vm.h"
#include "../vmstore/vm_manager.h"

typedef enum
{
  /* These first two constants define the granularity at which t values
     are calculated (in ms). Version vectors and capsules have separate
     timers, as version timers decay (lengthen) while capsules timers
     are constant, as they are not a continuous process.*/
  FAVIM_VIRUS_CAPSULE_VERSION_TIMER = 100,           // The units of time (ms)
  FAVIM_VIRUS_CAPSULE_TIMER = 100,           // The units of time (ms)
  FAVIM_VIRUS_VM_VERSION_TIMER = 100,
  FAVIM_VIRUS_VM_TIMER = 100,

  /* These constants define how many times a capsule is transmitted,
     the timer interval for Trickle suppression, and the redundancy constant
     k. Due to inherent loss, having a repeat > 1 is preferrable, although
     it should be small. It's better to broadcast the data twice rather
     than require another metadata announcement to trigger another
     transmission. It's not clear whether REDUNDANCY should be > or = to
     REPEAT. In either case, both constants should be small (e.g, 2-4). */
  
//  FAVIM_VIRUS_CAPSULE_REPEAT = 2,            // How many times to repeat a capsule
  FAVIM_VIRUS_CAPSULE_TAU = 10,              // Capsules have a fixed tau
  FAVIM_VIRUS_CAPSULE_REDUNDANCY = 2,        // Capsule redundancy (suppression pt.)

  /* These constants define the minimum and maximum tau values for
     version vector exchange, as well as the version vector redundancy
     constant k. Note that the tau values are in terms of multiples
     of the TIMER value above (e.g., a MIN of 10 and a TIMER of 100
     means a MIN of 1000 ms, or one second). */
  FAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN = 10,          // Version scaling tau minimum
  FAVIM_VIRUS_CAPSULE_VERSION_TAU_MAX = 600,         // Version scaling tau maximum
  FAVIM_VIRUS_CAPSULE_VERSION_REDUNDANCY = 1,        // Version redundancy (suppression pt.)
  
  /* Constants vm version timer */
  FAVIM_VIRUS_VM_VERSION_TAU_MIN = 10,
  FAVIM_VIRUS_VM_VERSION_TAU_MAX = 600,
  FAVIM_VIRUS_VM_VERSION_REDUNDANCY = 1,
  
  /* Constants vm timer */
//  FAVIM_VIRUS_VM_REPEAT = 2,
  FAVIM_VIRUS_VM_TAU = 10,
  FAVIM_VIRUS_VM_REDUNDANCY = 2,
  
  /* These constants are all for sending data larger than a single
     packet; they define the size of a program chunk, bitmasks, etc.*/ //TODO
  FAVIM_VIRUS_CAPSULE_CHUNK_SIZE = 16, // Send 16 bytes at once
  FAVIM_VIRUS_CAPSULE_BITMASK_ENTRIES = (FAVIM_CAPSULE_MAX_SIZE + FAVIM_VIRUS_CAPSULE_CHUNK_SIZE - 1)/FAVIM_VIRUS_CAPSULE_CHUNK_SIZE,
  FAVIM_VIRUS_CAPSULE_BITMASK_SIZE = (FAVIM_VIRUS_CAPSULE_BITMASK_ENTRIES + 7) / 8,
  
  /* Constants for the sending of dynamic version packets */
  // possibly not all <capsuleid,capsuleversion> fit in one packet
  FAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE = 120,
} FavimVirusConstants;

struct FavimTrickleTimer {
  uint16_t elapsed;      // Current time (in ticks)
  uint16_t threshold;    // Time to consider transmitting (in ticks) (t)
  uint16_t interval;     // Size of current interval (in ticks)      (tau)
  uint16_t numHeard;     // Number of messages heard                 (c)
} __attribute__((packed));

struct FavimCapsuleChunkMsg {
  FavimCapsuleVersion version;
  FavimCapsuleId capsuleNum;
  FavimCapsuleLength length;
  uint8_t piece;
  FavimOpCode chunk[FAVIM_VIRUS_CAPSULE_CHUNK_SIZE];
}  __attribute__((packed));

struct FavimCapsuleVersionChunkMsg {
	uint8_t nb_chunks;
	uint8_t piece;
	uint8_t chunk_length;
	FavimCapsuleIdVersionPair versions[];
}  __attribute__((packed));

struct FavimCapsuleStatusMsg {
  FavimCapsuleVersion version;
  uint8_t capsuleNum;
  uint8_t bitmask[FAVIM_VIRUS_CAPSULE_BITMASK_SIZE];
}  __attribute__((packed));

struct FavimVMVersionMsg { // For now we know this can't be larger than one packet, when this no longers holds, do something like FavimCapsuleVersionChunkMsg
	FavimVMVersion versions[FAVIM_VM_COUNT];
} __attribute__((packed));

struct FavimVMMsg {
	FavimVMDescription desc;
	FavimVMVersion version;
	BOOL present;
} __attribute__((packed));

//struct FavimVMIdVersionPair
//{
//	FavimVMId id;
//	FavimVMVersion version;
//} __attribute__((packed));

typedef enum {
    FAVIM_VIRUS_MAINTAIN,    //!< Everything seems to be in order: default state
    FAVIM_VIRUS_REQUEST,     //!< Somebody has something newer: receive capsules
    FAVIM_VIRUS_RESPOND,     //!< Somebody has something older: send capsules
} FavimVirusState;

//! The state of the virus module
typedef struct {
	func_cb_ptr get_capsule;
	func_cb_ptr get_nb_capsules;
	func_cb_ptr get_all_capsule_versions;
	func_cb_ptr get_vm_description;
	func_cb_ptr vm_present;
	func_cb_ptr get_vm_version;
    sos_pid_t pid;
    uint8_t sendBusy;
    //Capsules
    FavimTrickleTimer cap_versionTimer;
    FavimTrickleTimer cap_timer;
    FavimVirusState cap_state;
	FavimCapsule * cap_current; //!< Which capsule are we sending
	FavimCapsuleMsg * cap_currentMsg; //!< We are receiving a capsule, save it in here
    uint8_t cap_needBitmask[FAVIM_VIRUS_CAPSULE_BITMASK_SIZE];
	uint8_t* cap_versionNeedBitmask;
//	int8_t cap_comparison;
    FavimCapsuleChunkMsg cap_chunk;
    FavimCapsuleStatusMsg cap_statusMsg;
    //VMs
    FavimTrickleTimer vm_versionTimer;
    FavimTrickleTimer vm_timer;
	FavimVirusState vm_state;
	FavimVMMsg vm_current_msg;
} __attribute__((packed)) virus_state_t;

#endif
