/**
 * @file virus.c
 * @brief The Virus module of FAViM
 * @author Wouter Horr&eacute;
 * 
 * FAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 * 
 * TODO
 * This code is an adaptation of the code of ... (license?)
 * 
 */
 
#include <module.h>

#include <virus/virus.h>
#include <random.h>
#include <led.h>
#include <sos_timer.h>

typedef enum {
    FAVIM_VIRUS_MAINTAIN,    //!< Everything seems to be in order: default state
    FAVIM_VIRUS_REQUEST,     //!< Somebody has something newer: receive capsules
    FAVIM_VIRUS_RESPOND,     //!< Somebody has something older: send capsules
} FavimVirusState;

//! The state of the virus module
typedef struct {
    sos_pid_t pid;
    FavimCapsule* capsules[MATE_CAPSULE_NUM];
    FavimCapsuleVersion versions[MATE_CAPSULE_NUM];
    FavimTrickleTimer versionTimer;
    FavimTrickleTimer capsuleTimer;
    FavimVirusState state;
    FavimCapsuleId currentCapsule;  //!< If we're exchanging1 a capsule, which one
    FavimCapsuleVersion currentVersion; //!< If we need a capsule, what version
    uint8_t needBitmask[FAVIM_VIRUS_BITMASK_SIZE];
    bool sendBusy;
    bool capsuleBusy;
    FavimVersionMsg versionMsg;
    FavimCapsuleChunkMsg chunk;
    FavimCapsuleStatusMsg statusMsg;
} virus_state_t;

// Forward declarations
static int8_t module(void *state, Message *e);
static void sendCapsulePacket() ;
static void sendVersionPacket(uint16_t to) ;
static void installChunk(FavimCapsuleChunkMsg* chunk) ;
static void chunkReceive(FavimCapsuleChunkMsg* chunk) ;
static void receiveVector(Message *msg) ;
static void sendCapsulePacket() ;
static void toMaintainState() ;
static void toResponseState(uint8_t capsuleNum) ;
static int8_t registerCapsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule) ;
static void toRequestState(uint8_t capsuleNum, FavimCapsuleVersion version) ;
static void printNeedBitmask() ;
static uint8_t bitEntry(uint8_t* mask, uint8_t which) ;
static void newCounter(FavimTrickleTimer* timer) ;
static void decayNeedField() ;
static void clearInvalidBits() ;
static void removeNeededBit(uint16_t bit) ;
static int8_t getRandomChunkIndex(uint8_t capsule) ;

//! The module header of the virus module
static mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = FAVIM_VIRUS_PID,
	.state_size     = sizeof(virus_state_t),
    .num_sub_func	= 0,
    .num_prov_func	= 1,
    .num_timers		= 2,
    .module_handler	= module,
    .funct = {
        {registerCapsule, "cya2", FAVIM_VIRUS_PID, FAVIM_VIRUS_REGISTERCAPSULE_FID},
    },
};

//! The main message handling function
static int8_t module(void *state, Message *msg)
{
    virus_state_t *s = (virus_state_t *) state;

    switch (msg->type){
        case MSG_INIT: {
            s->pid = msg->did;
            s->state = FAVIM_VIRUS_MAINTAIN;
            s->currentCapsule = MATE_CAPSULE_INVALID;
			s->chunk.version = 0;
            nmemset(s->needBitmask, 0, FAVIM_VIRUS_BITMASK_SIZE);
			nmemset(s->versions, 0, MATE_CAPSULE_NUM*sizeof(FavimCapsuleVersion));
            // Initialize (but do not start) the timers
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            s->capsuleTimer.interval = FAVIM_VIRUS_CAPSULE_TAU;
            newCounter(&s->versionTimer);
            newCounter(&s->capsuleTimer);    

            DEBUG("MVirus: Initialized:\n\tchunk size=%i; bitmask size=%i\n", (int)FAVIM_VIRUS_CHUNK_SIZE, (int)FAVIM_VIRUS_BITMASK_SIZE);
            DEBUG("\tversion constants: quantum=%i;tau_l=%i;tau_h=%i;k=%i\n", (int)FAVIM_VIRUS_FAVIM_VIRUS_VERSION_TIMER_ID, (int)FAVIM_VIRUS_VERSION_TAU_MIN, (int)FAVIM_VIRUS_VERSION_TAU_MAX, (int)FAVIM_VIRUS_VERSION_REDUNDANCY);
            DEBUG("\tcapsule constants: quantum=%i;tau=%i;repeat=%i;k=%i.\n", (int)FAVIM_VIRUS_CAPSULE_TIMER, (int)FAVIM_VIRUS_CAPSULE_TAU, (int)FAVIM_VIRUS_CAPSULE_REPEAT, (int)FAVIM_VIRUS_CAPSULE_REDUNDANCY);

            s->sendBusy = FALSE;
            s->capsuleBusy = FALSE;
            
            ker_timer_init(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, TIMER_REPEAT);
            ker_timer_init(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, TIMER_REPEAT);
            ker_timer_start(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
			DEBUG("MVIRUS: ENTER HERE-------------------------Initialized\n");
            
            break;  
        }
        case MSG_FINAL: {
            ker_timer_stop(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID);
            ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
            break;
        }
        case FAVIM_CMD_VIRUS_VERSION_TIMER_TASK: {
            s->versionTimer.elapsed++;
            if (s->versionTimer.elapsed == s->versionTimer.threshold) {
                if (s->versionTimer.numHeard < FAVIM_VIRUS_VERSION_REDUNDANCY) {
                    sendVersionPacket(BCAST_ADDRESS);
                }
            } else if (s->versionTimer.elapsed >= s->versionTimer.interval) {
                DEBUG("MVirus: Version tau elapsed, picking new t.\n");
                s->versionTimer.interval *= 2;
                if (s->versionTimer.interval > FAVIM_VIRUS_VERSION_TAU_MAX) {
                    s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MAX;
                }
                newCounter(&s->versionTimer);
            }
            
             break; 
        }
        case FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK: {
            s->capsuleTimer.elapsed++;
            if (s->capsuleTimer.elapsed == s->capsuleTimer.threshold) {
                if (s->capsuleTimer.numHeard < FAVIM_VIRUS_CAPSULE_REDUNDANCY) {
                    sendCapsulePacket();
                    s->capsuleTimer.numHeard++;
                }
            } else if (s->capsuleTimer.elapsed >= s->capsuleTimer.interval) {
                newCounter(&s->capsuleTimer);
                decayNeedField();
            }
            
            break;
        }
        case FAVIM_CMD_VIRUS_CHECK_NEED_TASK: {
            uint8_t i;
            for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
                if (s->needBitmask[i] != 0) {
                    return -EINVAL;
                }
            }
            // TODO
            // post_short(M_HANDLER_STORE, s->pid, VIRUS_CAPSULE_INSTALLED, s->currentCapsule, 0, 0);	 
            s->versions[s->currentCapsule] = s->currentVersion;
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
            toMaintainState();
            // TODO
            // post_short(MATE_ENGINE_M, FAVIM_VIRUS_PID, RESUME, 0, 0, 0);
            
            break;
        }
//		case MSG_FROM_USER: 
//		{
//			unsigned char a = (unsigned char)*(msg->data);
//			if (a == '1') {
//				DEBUG("Sending timer0 context\n");
//				s->chunk.version = 1;
//				s->chunk.capsuleNum = MATE_CONTEXT_TIMER0;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 5;
//				s->chunk.chunk[1] = 0;		//5 byte data length
//				s->chunk.chunk[2] = 0x01;	//requires basic lib
//				s->chunk.chunk[3] = 0x01;	//PUSH 31
//				s->chunk.chunk[4] = 0x1F;	//toggling all leds		
//				s->chunk.chunk[5] = 0x2;	//LED
//				s->chunk.chunk[6] = 0x0;	//HALT
//			}
//			else if (a == '2') { 
//				s->chunk.version++;
//				s->chunk.capsuleNum = MATE_CONTEXT_REBOOT;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 5;
//				s->chunk.chunk[1] = 0;
//				s->chunk.chunk[2] = 0x01;	//requires only basic library
//				s->chunk.chunk[3] = 0x01;	//PUSH
//				s->chunk.chunk[4] = 0x0A;	//10
//				s->chunk.chunk[5] = 0x0D;	//SETTIMER0 REPEAT
//				s->chunk.chunk[6] = 0x00;	//HALT
//			}
//			else { 
//				s->chunk.version++;
//				s->chunk.capsuleNum = MATE_CONTEXT_TIMER0;
//				s->chunk.piece = 0;
//				s->chunk.chunk[0] = 0x1D;
//				s->chunk.chunk[1] = 0x0;
//				s->chunk.chunk[2] = 0x3;	//requires only basic library
//				s->chunk.chunk[3] = 0x1;	//push
//				s->chunk.chunk[4] = 0x5;	//7
//				s->chunk.chunk[5] = 0x4;	//led
//				s->chunk.chunk[6] = 0x22;	//halt
//				s->chunk.chunk[7] = 0x21;	//halt
//				s->chunk.chunk[8] = 0x8;	//halt
//				s->chunk.chunk[9] = 0x4;	//halt
//				s->chunk.chunk[10] = 0xE;	//halt
//				s->chunk.chunk[11] = 0x1C;	//halt
//				s->chunk.chunk[12] = 0x1;	//halt
//				s->chunk.chunk[13] = 0xF;	//halt
//				s->chunk.chunk[14] = 0xF;	//halt
//				s->chunk.chunk[15] = 0x5;	//halt
//				s->chunk.chunk[16] = 0x6;	//halt
//				s->chunk.chunk[17] = 0x5;	//halt
//				s->chunk.chunk[18] = 0x22;	//halt
//				s->chunk.chunk[19] = 0x9;	//halt
//				s->chunk.chunk[20] = 0x11;	//halt
//				s->chunk.chunk[21] = 0x16;	//halt
//				s->chunk.chunk[22] = 0x11;	//halt
//				s->chunk.chunk[23] = 0x14;	//halt
//				s->chunk.chunk[24] = 0x14;	//halt
//				s->chunk.chunk[25] = 0x14;	//halt
//				s->chunk.chunk[26] = 0x14;	//halt
//				s->chunk.chunk[27] = 0x10;	//halt
//				s->chunk.chunk[28] = 0x8D;	//halt
//				s->chunk.chunk[29] = 0x23;	//halt
//				s->chunk.chunk[30] = 0x0;	//halt
//			}
//			post_long(s->pid, s->pid, FAVIM_MSG_VIRUS_CAPSULECHUNKMSG, sizeof(FavimCapsuleChunkMsg), &(s->chunk), 0);
//			break;
//		}
        case FAVIM_MSG_VIRUS_CAPSULECHUNKMSG: {
            FavimCapsuleChunkMsg* chunk = (FavimCapsuleChunkMsg*)msg->data;
            DEBUG("MVirus: Received chunk broadcast.\n");
	    		ker_led(LED_GREEN_TOGGLE);
            chunkReceive(chunk);
            break;
        }
        case FAVIM_MSG_VIRUS_CAPSULESTATUSMSG: {
            FavimCapsuleStatusMsg* statusMsg = (FavimCapsuleStatusMsg*)msg->data;
            FavimCapsuleVersion version = statusMsg->version;
            FavimCapsuleId which = statusMsg->capsuleNum;
            DEBUG("MVirus: Received status message for %d.%d\n", (int)which, (int)version);
            // If I'm requesting this capsule already...
            if ((s->state == FAVIM_VIRUS_REQUEST) &&
                (s->currentCapsule == which)) {
                // and this version, increment my counter.
                if (s->currentVersion == version) {
                    s->versionTimer.numHeard++;
                }
                // and an older version, request the new version.
                else if (s->currentVersion < version) {
                    toRequestState(which, version);
                }
            }
            // If I'm idle or responding, and need this capsule,
            // start requesting it.
            else if (((s->state == FAVIM_VIRUS_MAINTAIN) || (s->state == FAVIM_VIRUS_RESPOND)) &&
                (s->versions[which] < version)) {
                toRequestState(which, version);
            }
            // If I'm idle and have this capsule, respond with parts of it.
            else if (s->state == FAVIM_VIRUS_MAINTAIN &&
                (s->versions[which] == version)) {
                memcopy(s->needBitmask, statusMsg->bitmask, FAVIM_VIRUS_BITMASK_SIZE);
                toResponseState(which);
            }
    
            // If I'm responding, then add these to the parts I know are needed.
            if ((s->state == FAVIM_VIRUS_RESPOND) &&
                (s->currentCapsule == which) &&
                (s->currentCapsule == version)) {
                int i;
                for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
                    s->needBitmask[i] |= statusMsg->bitmask[i];
                }
            }
            
            break;
        }
        case FAVIM_MSG_VIRUS_VERSIONMSG: {
            receiveVector(msg);
            break;
        }
        case FAVIM_MSG_VIRUS_VERSIONREQUESTMSG: {
            sendVersionPacket(UART_ADDRESS);
            break;
        }
        case MSG_PKT_SENDDONE: {
	    s->sendBusy = FALSE;
            break;
        }
        case MSG_TIMER_TIMEOUT: {
            MsgParam *t = (MsgParam *) msg->data;
            
            if (t->byte == FAVIM_VIRUS_VERSION_TIMER_ID) 
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_VERSION_TIMER_TASK, 0, 0, 0);
            else if (t->byte == FAVIM_VIRUS_CAPSULE_TIMER_ID)
                post_short(FAVIM_VIRUS_PID, FAVIM_VIRUS_PID, FAVIM_CMD_VIRUS_CAPSULE_TIMER_TASK, 0, 0, 0);
                
            break;
        }
        default:
            break;
    }
    
    return SOS_OK;
}

static int8_t registerCapsule(func_cb_ptr proto, FavimCapsuleId id, FavimCapsule* capsule) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    
    DEBUG("MVirus: Registering capsule %i as %p\n", (int)id, capsule);
    s->capsules[id] = capsule;
    return SOS_OK;
}

static void printNeedBitmask() {
#ifdef PC_PLATFORM
    int i;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    DEBUG("\tneedBitmask: ");
    for (i = 0; i < FAVIM_VIRUS_BITMASK_ENTRIES; i++) {
      DEBUG_SHORT("%s", (s->needBitmask[i/8] & (0x80 >> (i % 8)))? "1":"0");
    }
    DEBUG_SHORT("\n");
#endif
}
  
static uint8_t bitEntry(uint8_t* mask, uint8_t which) {
    return (mask[which >> 3] & (1 << (7 - (which & 7))));
}
  
  /* Select a new threshold, in the range [interval/2,interval], and
     clear out temporary state. */
static void newCounter(FavimTrickleTimer* timer) {
    timer->elapsed = 0;
    timer->threshold = timer->interval / 2;
    timer->threshold += ker_rand() % (timer->interval / 2);
    timer->numHeard = 0;
}

static void decayNeedField() {
    int i;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    DEBUG("MVirus: Decaying need field.\n");
    printNeedBitmask();
    for (i = 0; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
      s->needBitmask[i] &= (uint8_t)(ker_rand());
    }
    printNeedBitmask();
}

static void clearInvalidBits() {
    int i;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    uint8_t neededBits;
    uint16_t size = sizeof(FavimCapsule) - FAVIM_CAPSULE_SIZE + s->capsules[s->currentCapsule]->dataSize;
    DEBUG("MVirus: Total size of capsule is %d-%d+%d = %d (%d).\n", (int)sizeof(FavimCapsule), (int)FAVIM_CAPSULE_SIZE, (int)s->capsules[s->currentCapsule]->dataSize, (int)(size), (int)FAVIM_VIRUS_CHUNK_SIZE);
    size += (FAVIM_VIRUS_CHUNK_SIZE - 1);
    size /= FAVIM_VIRUS_CHUNK_SIZE;
    neededBits = size;
    DEBUG("MVirus: Clearing need bits for caspule %i.%i. (%i)\n", (int)s->currentCapsule, (int)s->currentVersion, (int)size);
    printNeedBitmask();
    if (neededBits & 7) {
      uint8_t mask = (0x80 >> ((neededBits & 7) - 1));
      mask -= 1;
      mask = ~mask;
      s->needBitmask[neededBits / 8] &= mask;
      neededBits -= neededBits & 7;
      neededBits += 8;
    }
    for (i = neededBits / 8; i < FAVIM_VIRUS_BITMASK_SIZE; i++) {
      s->needBitmask[i]  = 0;
    }
    printNeedBitmask();
}

static void removeNeededBit(uint16_t bit) {
    uint8_t mask = (1 << (7 - (bit & 7)));
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    DEBUG("MVirus: clear mask for %i: %hhx -> %hhx\n", bit/8, mask, ~mask);
    mask = ~mask;
    s->needBitmask[bit / 8] &= mask;
}

static void toRequestState(uint8_t capsuleNum, FavimCapsuleVersion version) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    
    DEBUG("MVirus: Moving to request state, capsule %i version %i.\n", (int)capsuleNum, (int)version);
    s->state = FAVIM_VIRUS_REQUEST;
    s->currentCapsule = capsuleNum;
    s->currentVersion = version;
    nmemset(s->needBitmask, 0xff, FAVIM_VIRUS_BITMASK_SIZE);
    // TODO
    //post_short(MATE_ENGINE_M, FAVIM_VIRUS_PID, HALT, 0, 0, 0);
    
    s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
    newCounter(&s->versionTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
}

static void toResponseState(uint8_t capsuleNum) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    
    DEBUG("MVirus: Moving to response state, capsule %i version %i.\n", (int)capsuleNum, (int)s->versions[capsuleNum]);
    s->state = FAVIM_VIRUS_RESPOND;
    s->currentCapsule = capsuleNum;
    
    newCounter(&s->capsuleTimer);
    ker_timer_restart(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID, FAVIM_VIRUS_CAPSULE_TIMER);
}

static void toMaintainState() {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    
    s->state = FAVIM_VIRUS_MAINTAIN;
    ker_timer_stop(s->pid, FAVIM_VIRUS_CAPSULE_TIMER_ID);
}
 
static int8_t getRandomChunkIndex(uint8_t capsule) {
    uint8_t idx;
    uint8_t count = 0;
    uint16_t size = sizeof(FavimCapsule);
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;

    DEBUG("MVirus: Get random chunk index from\n");
    printNeedBitmask();
    size -= (FAVIM_CAPSULE_SIZE - s->capsules[capsule]->dataSize);
    size += (FAVIM_VIRUS_CHUNK_SIZE - 1);
    size /= FAVIM_VIRUS_CHUNK_SIZE;
    for (idx = 0; idx < size; idx++) {
        if (bitEntry(s->needBitmask, idx)) {
            count++;
        }
    }

    if (count == 0) {
      DEBUG("MVirus: Tried to select capsule chunk to send, but none were needed. Returning -1.\n");
      return -1;
    }

    count = (ker_rand() % count) + 1;

    for (idx = 0; idx < size; idx++) {
        if (bitEntry(s->needBitmask, idx)) {
            count--;
        }
        if (count == 0) {
            DEBUG("MVirus: Select capsule chunk %i to send.\n", (int)idx);
            return (int8_t)idx;
        }
    }
    DEBUG("MVirus: Tried to select capsule chunk to send, but encountered an error due to miscount. Returning -1.\n");
    return -1;
  }

static void sendCapsulePacket() {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    
    if (s->sendBusy) {
        DEBUG("MVirus: Tried to send capsule, but send busy.\n");
        return;
    }
    else {
        int8_t idx = getRandomChunkIndex(s->currentCapsule);
        if (idx >= 0) {
            uint8_t* chunkPtr = (uint8_t*)s->capsules[s->currentCapsule];
            chunkPtr += FAVIM_VIRUS_CHUNK_SIZE * idx;
            memcopy(s->chunk.chunk, chunkPtr, FAVIM_VIRUS_CHUNK_SIZE);

            s->chunk.capsuleNum = s->currentCapsule;
            s->chunk.piece = idx;
            s->chunk.version = s->versions[s->currentCapsule];
            if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULECHUNKMSG, sizeof(FavimCapsuleChunkMsg), &s->chunk, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
                DEBUG("MVirus: Sent chunk %i of capsule %i.%i.\n", (int)idx, (int)s->currentCapsule, (int)s->currentVersion);
                s->sendBusy = TRUE;
                removeNeededBit(idx);
            } else {
                DEBUG("MVirus: Sending chunk %i of capsule %i.%i FAILED.\n", (int)idx, (int)s->currentCapsule, (int)s->currentVersion);
            }
        } else {
            DEBUG("MVirus: Tried to send capsule, but nothing to send. Return to maintain state.\n");
            toMaintainState();
        }
    }
}

static void sendVersionPacket(uint16_t to) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;

    if (s->sendBusy) {
      DEBUG("MVirus: Tried to send version packet, but send busy.\n");
      return;
    }
    
    if (s->state == FAVIM_VIRUS_MAINTAIN || s->state == FAVIM_VIRUS_RESPOND) {
        int i;
        DEBUG("MVirus: In state %i, sending version vector:\n  ", (int)s->state);
        for (i = 0; i < MATE_CAPSULE_NUM; i++) {
            DEBUG_SHORT("[%i]", (int)s->versions[i]);
            s->versionMsg.versions[i] = s->versions[i];
        }
        DEBUG_SHORT("\n");
        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_VERSIONMSG, sizeof(FavimVersionMsg), &s->versionMsg, SOS_MSG_RELIABLE, to) == SOS_OK) {
            DEBUG("MVirus: Sent version vector.\n");
            s->sendBusy = TRUE;
        }
    } else if (s->state == FAVIM_VIRUS_REQUEST && to == BCAST_ADDRESS) {
        s->statusMsg.capsuleNum = s->currentCapsule;
        s->statusMsg.version = s->currentVersion;
        memcopy(s->statusMsg.bitmask, s->needBitmask, FAVIM_VIRUS_BITMASK_SIZE);
        DEBUG("MVirus: In state %i, sending bitmask:\n", (int)s->state);
        printNeedBitmask();
        if (post_net(FAVIM_VIRUS_PID, s->pid, FAVIM_MSG_VIRUS_CAPSULESTATUSMSG, sizeof(FavimCapsuleStatusMsg), &s->statusMsg, SOS_MSG_RELIABLE, BCAST_ADDRESS) == SOS_OK) {
            DEBUG("MVirus: Sent version bitmask.\n");
            s->sendBusy = TRUE;
        }
    } else {
        DEBUG("MVirus: In invalid state for sending version packet: %i\n", (int)s->state);
        return;
    }
}
  
static void receiveVector(Message *msg) {
    uint8_t capsuleNum;
    int8_t comparison = 0; // msg is: -1 older, 0 same, 1 newer
    FavimVersionMsg* versionMsg = (FavimVersionMsg*)msg->data;
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;

    if (s->state == FAVIM_VIRUS_REQUEST) {
        if (versionMsg->versions[s->currentCapsule] > s->currentVersion) {
            toRequestState(s->currentCapsule, versionMsg->versions[s->currentCapsule]);
        } else if (versionMsg->versions[s->currentCapsule] == s->currentVersion) {
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
            ker_timer_restart(s->pid, FAVIM_VIRUS_VERSION_TIMER_ID, FAVIM_VIRUS_VERSION_TIMER);
        } else {
            DEBUG("MVirus: In request state for %i.%i already.\n", (int)s->currentCapsule, (int)s->currentVersion);
        }
        return;
    }
    
    for (capsuleNum = 0; capsuleNum < MATE_CAPSULE_NUM; capsuleNum++) {
        if (s->capsules[capsuleNum] != NULL) {
            if(versionMsg->versions[capsuleNum] > s->versions[capsuleNum]) {
                comparison = 1;
                break;
            } else if (versionMsg->versions[capsuleNum] < s->versions[capsuleNum]) {
                comparison = -1;
                break;
            }
        }
    }
    
    // Requests are given precedence over responses: there's no
    // point updating someone to something that's out of date.
    if (comparison == 0) {
        DEBUG("MVirus: Heard identical vector, suppressing.\n");
        s->versionTimer.numHeard++;
    } else if (comparison == 1) {
        DEBUG("Heard newer vector, requesting.\n");
        if (s->state != FAVIM_VIRUS_REQUEST) {
            toRequestState(capsuleNum, versionMsg->versions[capsuleNum]);
        }
    } else if (comparison == -1) {
        DEBUG("Heard older vector, updating.\n");
        if (s->state == FAVIM_VIRUS_MAINTAIN) {
            nmemset(s->needBitmask, 0xff, FAVIM_VIRUS_BITMASK_SIZE);
            toResponseState(capsuleNum);
        }
    }
    return;
}

static void installChunk(FavimCapsuleChunkMsg* chunk) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;
    uint8_t* destPtr = (uint8_t*)s->capsules[chunk->capsuleNum];

    destPtr += chunk->piece * FAVIM_VIRUS_CHUNK_SIZE;
    DEBUG("MVirus: Copying chunk %d of capsule %d version %d.\n", (int)chunk->piece, (int)chunk->capsuleNum, (int)chunk->version);
    memcopy(destPtr, chunk->chunk, FAVIM_VIRUS_CHUNK_SIZE);
    // Remove that bit entry
    if (chunk->piece == 0) { // If it's the chunk metadata, clear out bits
      clearInvalidBits();
    }
    removeNeededBit(chunk->piece);
    printNeedBitmask();
    post_short(s->pid, s->pid, FAVIM_CMD_VIRUS_CHECK_NEED_TASK, 0, 0, 0);
}


static void chunkReceive(FavimCapsuleChunkMsg* chunk) {
    sos_module_t *m = ker_get_module(FAVIM_VIRUS_PID);
    virus_state_t *s = (virus_state_t *)m->handler_state;

    s->capsuleTimer.numHeard++;
    if (s->state == FAVIM_VIRUS_REQUEST) {
		
        if (chunk->capsuleNum == s->currentCapsule &&
            chunk->version == s->currentVersion &&
            bitEntry(s->needBitmask, chunk->piece)) {
			
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        } else if (chunk->capsuleNum == s->currentCapsule && chunk->version > s->currentVersion) {
            toRequestState(chunk->capsuleNum, chunk->version);
			
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        }
    } else if (s->state == FAVIM_VIRUS_MAINTAIN) {
        if (chunk->version > s->versions[chunk->capsuleNum]) {
            toRequestState(chunk->capsuleNum, chunk->version);
            installChunk(chunk);
            s->versionTimer.interval = FAVIM_VIRUS_VERSION_TAU_MIN;
            newCounter(&s->versionTimer);
        }
    } 
}


#ifndef _MODULE_
mod_header_ptr virus_get_header()
{
    return sos_get_header_address(mod_header);
}
#endif

