/**
 * @file virus.h
 * 
 * FAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 *
 */

#ifndef _FAVIM_VIRUS_H_
#define _FAVIM_VIRUS_H_

typedef struct FavimTrickleTimer FavimTrickleTimer;
typedef struct FavimCapsuleChunkMsg FavimCapsuleChunkMsg;
//typedef struct FavimCapsuleMsg FavimCapsuleMsg;
typedef struct FavimCapsuleStatusMsg FavimCapsuleStatusMsg;
typedef struct FavimVersionMsg FavimVersionMsg;

#include "../favim.h"
#include "../capsulestore/capsule.h"
#include "../capsulestore/capsule_manager.h"

//*************************************************
// TODO TEMP
//*************************************************

// # capsules moet dynamisch of er moet een maximum op

typedef enum {
  //MATE_CAPSULE_TRIGGER	 = 0, // unique("MateCapsuleID"),
  MATE_CAPSULE_TIMER0	 = 0, // unique("MateCapsuleID"),
  //MATE_CAPSULE_TIMER1	 = 2, // unique("MateCapsuleID"),
  MATE_CAPSULE_PUSHBUTTON	 = 1, // unique("MateCapsuleID"),
  MATE_CAPSULE_REBOOT	 = 2, // unique("MateCapsuleID"),
  MATE_CAPSULE_CLIFF	 = 3, // unique("MateCapsuleID"),
  MATE_CAPSULE_NUM	 = 4, //9, // unique("MateCapsuleID"),
  MATE_CAPSULE_INVALID = 255
} MateCapsuleID;

//*************************************************
// END TEMP
//*************************************************

enum
{
	FAVIM_VIRUS_REGISTERCAPSULE_FID = 0,
};

typedef int8_t (*favim_virus_registerCapsule_proto)(func_cb_ptr, FavimCapsuleId, FavimCapsule);


typedef enum
{
  /* These first two constants define the granularity at which t values
     are calculated (in ms). Version vectors and capsules have separate
     timers, as version timers decay (lengthen) while capsules timers
     are constant, as they are not a continuous process.*/
  FAVIM_VIRUS_VERSION_TIMER = 100,           // The units of time (ms)
  FAVIM_VIRUS_CAPSULE_TIMER = 100,           // The units of time (ms)

  /* These constants define how many times a capsule is transmitted,
     the timer interval for Trickle suppression, and the redundancy constant
     k. Due to inherent loss, having a repeat > 1 is preferrable, although
     it should be small. It's better to broadcast the data twice rather
     than require another metadata announcement to trigger another
     transmission. It's not clear whether REDUNDANCY should be > or = to
     REPEAT. In either case, both constants should be small (e.g, 2-4). */
  
  FAVIM_VIRUS_CAPSULE_REPEAT = 2,            // How many times to repeat a capsule
  FAVIM_VIRUS_CAPSULE_TAU = 10,              // Capsules have a fixed tau
  FAVIM_VIRUS_CAPSULE_REDUNDANCY = 2,        // Capsule redundancy (suppression pt.)

  /* These constants define the minimum and maximum tau values for
     version vector exchange, as well as the version vector redundancy
     constant k. Note that the tau values are in terms of multiples
     of the TIMER value above (e.g., a MIN of 10 and a TIMER of 100
     means a MIN of 1000 ms, or one second). */
  FAVIM_VIRUS_VERSION_TAU_MIN = 10,          // Version scaling tau minimum
  FAVIM_VIRUS_VERSION_TAU_MAX = 600,         // Version scaling tau maximum
  FAVIM_VIRUS_VERSION_REDUNDANCY = 1,        // Version redundancy (suppression pt.)
  
  /* These constants are all for sending data larger than a single
     packet; they define the size of a program chunk, bitmasks, etc.*/
  FAVIM_VIRUS_CAPSULE_HEADER_SIZE = 2,
  FAVIM_VIRUS_CHUNK_SIZE = FAVIM_CAPSULE_SIZE + FAVIM_VIRUS_CAPSULE_HEADER_SIZE,
  FAVIM_VIRUS_BITMASK_ENTRIES = ((FAVIM_CAPSULE_SIZE + FAVIM_VIRUS_CHUNK_SIZE - 1) / FAVIM_VIRUS_CHUNK_SIZE),
  FAVIM_VIRUS_BITMASK_SIZE = (FAVIM_VIRUS_BITMASK_ENTRIES + 7) / 8,
} FavimVirusConstants;

struct FavimTrickleTimer {
  uint16_t elapsed;      // Current time (in ticks)
  uint16_t threshold;    // Time to consider transmitting (in ticks) (t)
  uint16_t interval;     // Size of current interval (in ticks)      (tau)
  uint16_t numHeard;     // Number of messages heard                 (c)
} __attribute__((packed));

struct FavimCapsuleChunkMsg {
  FavimCapsuleVersion version;
  uint8_t capsuleNum;
  uint8_t piece;
  uint8_t chunk[FAVIM_VIRUS_CHUNK_SIZE];
}  __attribute__((packed));

struct FavimVersionMsg {
  FavimCapsuleVersion versions[MATE_CAPSULE_NUM];
}  __attribute__((packed));

struct FavimCapsuleStatusMsg {
  FavimCapsuleVersion version;
  uint8_t capsuleNum;
  uint8_t bitmask[FAVIM_VIRUS_BITMASK_SIZE];
}  __attribute__((packed));

//struct FavimCapsuleMsg {
//  FavimCapsule capsule;
//}  __attribute__((packed));


#endif
