/*
 * Copyright (c) 2006, Wouter Horré
 * All rights reserved
 * 
 * See LICENSE for details
 */
/**
 * @file virus.h
 * 
 * DAViM Virus uses the Trickle algorithm for code propagation and maintenance.
 * A full description and evaluation of the algorithm can be found in
 *
 * Philip Levis, Neil Patel, David Culler, and Scott Shenker.
 * "Trickle: A Self-Regulating Algorithm for Code Propagation and Maintenance
 * in Wireless Sensor Networks." In Proceedings of the First USENIX/ACM
 * Symposium on Networked Systems Design and Implementation (NSDI 2004).
 *
 * A copy of the paper can be downloaded from Phil Levis' web site:
 *        http://www.cs.berkeley.edu/~pal/
 *
 * A brief description of the algorithm can be found in the comments
 * at the head of MVirus.nc in the original implementation of the algorithm
 * included with Mat&eacute; (http://www.cs.berkeley.edu/~pal/mate-web/index.html).
 *
 */

#ifndef _DAVIM_VIRUS_H_
#define _DAVIM_VIRUS_H_

typedef struct DavimTrickleTimer DavimTrickleTimer;
typedef struct DavimCapsuleChunkMsg DavimCapsuleChunkMsg;
typedef struct DavimCapsuleStatusMsg DavimCapsuleStatusMsg;
typedef struct DavimCapsuleVersionChunkMsg DavimCapsuleVersionChunkMsg;
typedef struct DavimVMVersionMsg DavimVMVersionMsg;
typedef struct DavimVMMsg DavimVMMsg;
//typedef struct DavimVMIdVersionPair DavimVMIdVersionPair;

#include "../davim.h"
#include "../capsulestore/capsule.h"
#include "../capsulestore/capsule_manager.h"
#include "../vmstore/vm.h"
#include "../vmstore/vm_manager.h"

typedef enum
{
  /* These first two constants define the granularity at which t values
     are calculated (in ms). Version vectors and capsules have separate
     timers, as version timers decay (lengthen) while capsules timers
     are constant, as they are not a continuous process.*/
  DAVIM_VIRUS_CAPSULE_VERSION_TIMER = 100,           // The units of time (ms)
  DAVIM_VIRUS_CAPSULE_TIMER = 100,           // The units of time (ms)
  DAVIM_VIRUS_VM_VERSION_TIMER = 100,
  DAVIM_VIRUS_VM_TIMER = 100,

  /* These constants define how many times a capsule is transmitted,
     the timer interval for Trickle suppression, and the redundancy constant
     k. Due to inherent loss, having a repeat > 1 is preferrable, although
     it should be small. It's better to broadcast the data twice rather
     than require another metadata announcement to trigger another
     transmission. It's not clear whether REDUNDANCY should be > or = to
     REPEAT. In either case, both constants should be small (e.g, 2-4). */
  
//  DAVIM_VIRUS_CAPSULE_REPEAT = 2,            // How many times to repeat a capsule
  DAVIM_VIRUS_CAPSULE_TAU = 10,              // Capsules have a fixed tau
  DAVIM_VIRUS_CAPSULE_REDUNDANCY = 2,        // Capsule redundancy (suppression pt.)

  /* These constants define the minimum and maximum tau values for
     version vector exchange, as well as the version vector redundancy
     constant k. Note that the tau values are in terms of multiples
     of the TIMER value above (e.g., a MIN of 10 and a TIMER of 100
     means a MIN of 1000 ms, or one second). */
  DAVIM_VIRUS_CAPSULE_VERSION_TAU_MIN = 10,          // Version scaling tau minimum
  DAVIM_VIRUS_CAPSULE_VERSION_TAU_MAX = 600,         // Version scaling tau maximum
  DAVIM_VIRUS_CAPSULE_VERSION_REDUNDANCY = 1,        // Version redundancy (suppression pt.)
  
  /* Constants vm version timer */
  DAVIM_VIRUS_VM_VERSION_TAU_MIN = 10,
  DAVIM_VIRUS_VM_VERSION_TAU_MAX = 600,
  DAVIM_VIRUS_VM_VERSION_REDUNDANCY = 1,
  
  /* Constants vm timer */
//  DAVIM_VIRUS_VM_REPEAT = 2,
  DAVIM_VIRUS_VM_TAU = 10,
  DAVIM_VIRUS_VM_REDUNDANCY = 2,
  
  /* These constants are all for sending data larger than a single
     packet; they define the size of a program chunk, bitmasks, etc.*/ //TODO
  DAVIM_VIRUS_CAPSULE_CHUNK_SIZE = 16, // Send 16 bytes at once
  DAVIM_VIRUS_CAPSULE_BITMASK_ENTRIES = (DAVIM_CAPSULE_MAX_SIZE + DAVIM_VIRUS_CAPSULE_CHUNK_SIZE - 1)/DAVIM_VIRUS_CAPSULE_CHUNK_SIZE,
  DAVIM_VIRUS_CAPSULE_BITMASK_SIZE = (DAVIM_VIRUS_CAPSULE_BITMASK_ENTRIES + 7) / 8,
  
  /* Constants for the sending of dynamic version packets */
  // possibly not all <capsuleid,capsuleversion> fit in one packet
  DAVIM_VIRUS_CAPSULE_VERSION_CHUNK_SIZE = 120,
} DavimVirusConstants;

struct DavimTrickleTimer {
  uint16_t elapsed;      // Current time (in ticks)
  uint16_t threshold;    // Time to consider transmitting (in ticks) (t)
  uint16_t interval;     // Size of current interval (in ticks)      (tau)
  uint16_t numHeard;     // Number of messages heard                 (c)
} __attribute__((packed));

struct DavimCapsuleChunkMsg {
  DavimCapsuleVersion version;
  DavimCapsuleId capsuleNum;
  DavimCapsuleLength length;
  uint8_t piece;
  DavimOpCode chunk[DAVIM_VIRUS_CAPSULE_CHUNK_SIZE];
}  __attribute__((packed));

struct DavimCapsuleVersionChunkMsg {
	uint8_t nb_chunks;
	uint8_t piece;
	uint8_t chunk_length;
	DavimCapsuleIdVersionPair versions[];
}  __attribute__((packed));

struct DavimCapsuleStatusMsg {
  DavimCapsuleVersion version;
  uint8_t capsuleNum;
  uint8_t bitmask[DAVIM_VIRUS_CAPSULE_BITMASK_SIZE];
}  __attribute__((packed));

struct DavimVMVersionMsg { // For now we know this can't be larger than one packet, when this no longers holds, do something like DavimCapsuleVersionChunkMsg
	DavimVMVersion versions[DAVIM_VM_COUNT];
} __attribute__((packed));

struct DavimVMMsg {
	DavimVMDescription desc;
	DavimVMVersion version;
	BOOL present;
} __attribute__((packed));

//struct DavimVMIdVersionPair
//{
//	DavimVMId id;
//	DavimVMVersion version;
//} __attribute__((packed));

typedef enum {
    DAVIM_VIRUS_MAINTAIN,    //!< Everything seems to be in order: default state
    DAVIM_VIRUS_REQUEST,     //!< Somebody has something newer: receive capsules
    DAVIM_VIRUS_RESPOND,     //!< Somebody has something older: send capsules
} DavimVirusState;

//! The state of the virus module
typedef struct {
	func_cb_ptr get_capsule;
	func_cb_ptr get_nb_capsules;
	func_cb_ptr get_all_capsule_versions;
	func_cb_ptr get_vm_description;
	func_cb_ptr vm_present;
	func_cb_ptr get_vm_version;
    sos_pid_t pid;
    uint8_t sendBusy;
    //Capsules
    DavimTrickleTimer cap_versionTimer;
    DavimTrickleTimer cap_timer;
    DavimVirusState cap_state;
	DavimCapsule * cap_current; //!< Which capsule are we sending
	DavimCapsuleMsg * cap_currentMsg; //!< We are receiving a capsule, save it in here
    uint8_t cap_needBitmask[DAVIM_VIRUS_CAPSULE_BITMASK_SIZE];
	uint8_t* cap_versionNeedBitmask;
//	int8_t cap_comparison;
    DavimCapsuleChunkMsg cap_chunk;
    DavimCapsuleStatusMsg cap_statusMsg;
    //VMs
    DavimTrickleTimer vm_versionTimer;
    DavimTrickleTimer vm_timer;
	DavimVirusState vm_state;
	DavimVMMsg vm_current_msg;
} __attribute__((packed)) virus_state_t;

#endif
